/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "ROM.h"
#include "CPU.h"
#include "RSP.h"
#include "PIF.h"		// CController
#include "R4300.h"
#include "ROMBuffer.h"
#include "RomSettings.h"
#include "Compatibility.h"

#include "Daedalus_CRC.h"
#include "DaedStream.h"
#include "DaedIO.h"

#include "Interface\MainWindow.h"
#include "Interface\RomDB.h"

#include "Debug\DBGConsole.h"
#include "Debug\Debug.h"
#include "OSHLE\Patch.h"			// Patch_ApplyPatches
#include "OSHLE\ultra_r4300.h"
#include "OSHLE\ultra_os.h"		// System type

#include "Utility\ROMFile.h"

#include "Plugins\AudioPlugin.h"
#include "Plugins\GraphicsPlugin.h"


#undef min

//*****************************************************************************
// This isn't really the most apropriate place. Need to check with
// the graphics plugin really
//*****************************************************************************
u32 g_dwNumFrames = 0;

//*****************************************************************************
// TODO: Move these to ultra_ramrom.h
//*****************************************************************************
static const u32 RAMROM_FONTDATA_SIZE		= 1152;
static const u32 RAMROM_CLOCKRATE_MASK		= 0xfffffff0;

static const u32 RAMROM_CLOCKRATE_OFFSET	= 0x4;
static const u32 RAMROM_BOOTADDR_OFFSET		= 0x8;
static const u32 RAMROM_RELEASE_OFFSET		= 0xc;
static const u32 RAMROM_BOOTSTRAP_OFFSET	= 0x40;
static const u32 RAMROM_FONTDATA_OFFSET		= 0xb70;
static const u32 RAMROM_GAME_OFFSET			= 0x1000;

DAEDALUS_STATIC_ASSERT( sizeof(ROMHeader) == RAMROM_BOOTSTRAP_OFFSET );

//*****************************************************************************
//
//*****************************************************************************
static void ROM_GetRomNameFromHeader( char * p_rom_name, const ROMHeader & header );
static void ROM_CheckSumCic6101();
static void ROM_CheckSumCic6102();
static void ROM_CheckSumCic6103();
static void ROM_CheckSumCic6104();
static void ROM_CheckSumCic6105();
static void ROM_CheckSumCic6106();

RomInfo g_ROM;

static bool ROM_LoadCartToMemory( const char * filename, bool allow_dynamic_loading );
static void DumpROMInfo( const ROMHeader & header );


//*****************************************************************************
// CRC / CIC values
//*****************************************************************************
#define CIC_6101_BOOT_CRC 0xb4086651		// CIC-6101
#define CIC_6102_BOOT_CRC 0xb9c47dc8		// CIC-6102 or CIC-7101
#define CIC_6103_BOOT_CRC 0xedce5ad9		// CIC-6103 or CIC-7103
#define CIC_6104_BOOT_CRC 0xb3d6a525		// CIC-6104?   CIC-7102
#define CIC_6105_BOOT_CRC 0xb53c588a		// CIC-6105 or CIC-7105
#define CIC_6106_BOOT_CRC 0x06d8ed9c		// CIC-6106 or CIC-7106

#define CIC_6101_BOOT_CIC 0x00023f3f
#define CIC_6102_BOOT_CIC 0x00063f3f
#define CIC_6103_BOOT_CIC 0x0002783f
#define CIC_6104_BOOT_CIC 0x00023f3f		// Incorrect?
#define CIC_6105_BOOT_CIC 0x0002913f
#define CIC_6106_BOOT_CIC 0x0002853f		// Same as FZero???

struct SCicInfo
{
	const CHAR * Name;			// Name of the CIC chip
	u32		BootCRC;			// CRC of the bootcode
	u32		CICValues;			// CIC security values
	void (* CheckSumProc)();	// Function to calculate/fix checksum
};

#define CIC_CHIP_ENTRY( c )		{ "CIC-"#c, CIC_##c##_BOOT_CRC, CIC_##c##_BOOT_CIC, ROM_CheckSumCic##c }

static const SCicInfo gCICInfo[] = 
{
	CIC_CHIP_ENTRY( 6101 ),
	CIC_CHIP_ENTRY( 6102 ),
	CIC_CHIP_ENTRY( 6103 ),
	CIC_CHIP_ENTRY( 6104 ),
	CIC_CHIP_ENTRY( 6105 ),
	CIC_CHIP_ENTRY( 6106 ),
};
DAEDALUS_STATIC_ASSERT( ARRAYSIZE( gCICInfo ) == NUM_CIC_CHIPS );

//*****************************************************************************
//
//*****************************************************************************
void DumpROMInfo( const ROMHeader & header )
{	
	// The "Header" is actually something to do with the PI_DOM_*_OFS values...
	DBGConsole_Msg(0, "Header:			0x%02x%02x%02x%02x", header.x1, header.x2, header.x3, header.x4);
	DBGConsole_Msg(0, "Clockrate:       0x%08x", header.ClockRate);
	DBGConsole_Msg(0, "BootAddr:		0x%08x", SwapEndian(header.BootAddress));
	DBGConsole_Msg(0, "Release:         0x%08x", header.Release);
	DBGConsole_Msg(0, "CRC1:            0x%08x", header.CRC1);
	DBGConsole_Msg(0, "CRC2:            0x%08x", header.CRC2);
	DBGConsole_Msg(0, "Unknown0:        0x%08x", header.Unknown0);
	DBGConsole_Msg(0, "Unknown1:        0x%08x", header.Unknown1);
	DBGConsole_Msg(0, "ImageName:       '%s'",   header.Name);
	DBGConsole_Msg(0, "Unknown2:        0x%08x", header.Unknown2);
	DBGConsole_Msg(0, "Unknown3:        0x%04x", header.Unknown3);
	DBGConsole_Msg(0, "Unknown4:        0x%02x", header.Unknown4);
	DBGConsole_Msg(0, "Manufacturer:    0x%02x", header.Manufacturer);
	DBGConsole_Msg(0, "CartID:          0x%04x", header.CartID);
	DBGConsole_Msg(0, "CountryID:       0x%02x - '%c'", header.CountryID, (char)header.CountryID);
	DBGConsole_Msg(0, "Unknown5:        0x%02x", header.Unknown5);
}

//*****************************************************************************
//
//*****************************************************************************
static void ROM_RunPIFBoot( ECicType cic_chip )
{
	CPU_SetPC(0xBFC00000);

	// Set up CIC value in 0xbfc007e4
	*(u32*)((u8*)g_pMemoryBuffers[MEM_PIF_RAM] + 0x7C0 + 0x24) = gCICInfo[ cic_chip ].CICValues;
}

//*****************************************************************************
//
//*****************************************************************************
static void ROM_SimulatePIFBoot( ECicType cic_chip, u32 nSystemType )
{
	u32		cic( gCICInfo[ cic_chip ].CICValues );

	// g_qwCPR[0][C0_SR]		= 0x34000000;	//*SR_FR |*/ SR_ERL | SR_CU2|SR_CU1|SR_CU0;
	R4300_SetSR(0x34000000);
	g_qwCPR[0][C0_CONFIG]._u64	= 0x0006E463;	// 0x00066463;	

	// These are the effects of the PIF Boot Rom:
	gGPR[REG_r0]._u64 = 0x0000000000000000LL;	// 00
	gGPR[REG_at]._u64 = 0x0000000000000000LL;	// 01
	gGPR[REG_v0]._u64 = 0xFFFFFFFFD1731BE9LL;	// 02 Required by CIC-6105, ignored by other boot codes
	gGPR[REG_v1]._u64 = 0xFFFFFFFFD1731BE9LL;	// 03
	gGPR[REG_a0]._u64 = 0x0000000000001BE9LL;	// 04
	gGPR[REG_a1]._u64 = 0xFFFFFFFFF45231E5LL;	// 05
	gGPR[REG_a2]._u64 = 0xFFFFFFFFA4001F0CLL;	// 06
	gGPR[REG_a3]._u64 = 0xFFFFFFFFA4001F08LL;	// 07
	gGPR[REG_t0]._u64 = 0x00000000000000C0LL;	// 08 0x70?
	gGPR[REG_t1]._u64 = 0x0000000000000000LL;	// 09
	gGPR[REG_t2]._u64 = 0x0000000000000040LL;	// 10
	gGPR[REG_t3]._u64 = 0xFFFFFFFFA4000040LL;	// 11 Required by CIC-6105, ignored by other boot codes
	gGPR[REG_t4]._u64 = 0xFFFFFFFFD1330BC3LL;	// 12
	gGPR[REG_t5]._u64 = 0xFFFFFFFFD1330BC3LL;	// 13
	gGPR[REG_t6]._u64 = 0x0000000025613A26LL;	// 14
	gGPR[REG_t7]._u64 = 0x000000002EA04317LL;	// 15
	gGPR[REG_s0]._u64 = 0x0000000000000000LL;	// 16
	gGPR[REG_s1]._u64 = 0x0000000000000000LL;	// 17
	gGPR[REG_s2]._u64 = 0x0000000000000000LL;	// 18
	gGPR[REG_s3]._u64 = 0x0000000000000000LL;	// 19
	gGPR[REG_s4]._u64 = nSystemType;			// 20
	gGPR[REG_s5]._u64 = 0x0000000000000000LL;	// 21
    gGPR[REG_s6]._u64 = ( cic & 0x0000FF00 ) >> 8;	// 22
    gGPR[REG_s7]._u64 = 0x0000000000000006LL;	// 23
	gGPR[REG_t8]._u64 = 0x0000000000000000LL;	// 24
	gGPR[REG_t9]._u64 = 0xFFFFFFFFD73F2993LL;	// 25
	gGPR[REG_k0]._u64 = 0x0000000000000000LL;	// 26
	gGPR[REG_k1]._u64 = 0x0000000000000000LL;	// 27
	gGPR[REG_gp]._u64 = 0x0000000000000000LL;	// 28
	gGPR[REG_sp]._u64 = 0xFFFFFFFFA4001FF0LL;	// 29
	gGPR[REG_s8]._u64 = 0x0000000000000000LL;	// 30
	gGPR[REG_ra]._u64 = 0xFFFFFFFFA4001554LL;	// 31  0xFFFFFFFFA4000040;????

	// Copy low 1000 bytes to MEM_SP_MEM
	RomBuffer::GetRomBytesRaw( (u8*)g_pMemoryBuffers[MEM_SP_MEM] + RAMROM_BOOTSTRAP_OFFSET, 
		   RAMROM_BOOTSTRAP_OFFSET,
		   RAMROM_GAME_OFFSET - RAMROM_BOOTSTRAP_OFFSET );

//				pal						               ntsc
		/*	gGPR[REG_a1]._u64 = 0xFFFFFFFFDECAAAD1;		gGPR[REG_a1]._u64 = 0x000000005493FB9A;
			gGPR[REG_t6]._u64 = 0x000000000CF85C13;		gGPR[REG_t6]._u64 = 0xFFFFFFFFC2C20384;
			gGPR[REG_t8]._u64 = 0x0000000000000002;		gGPR[REG_t8]._u64 = 0x0000000000000003;
			gGPR[REG_s4]._u64 = 0x0000000000000000;		gGPR[REG_s4]._u64 = 0x0000000000000001;
			gGPR[REG_s7]._u64 = 0x0000000000000006;		gGPR[REG_s7]._u64 = 0x0000000000000000;
			gGPR[REG_ra]._u64 = 0xFFFFFFFFA4001554;		gGPR[REG_ra]._u64 = 0xFFFFFFFFA4001550;*/

	// Need to copy crap to SP_IMEM for CIC-6105 boot.
	u8 * pIMemBase = (u8*)g_pMemoryBuffers[ MEM_SP_MEM ] + 0x1000;
	
	u32 data[] = 
	{
		0x3c0dbfc0,
		0x8da807fc,		// 0xbda807fc for pal?
		0x25ad07c0,
		0x31080080,
		0x5500fffc,
		0x3c0dbfc0,
		0x8da80024,
		0x3c0bb000
	};

	memcpy( pIMemBase, data, sizeof(data) );

	// Also need to set up PI_BSD_DOM1 regs etc!
	CPU_SetPC(0xA4000040);
}

//*****************************************************************************
// Get the name of the correct pit for the region
//*****************************************************************************
static const CHAR * ROM_GetPIFName( u32 tv_type )
{
	switch ( tv_type )
	{
		case OS_TV_PAL:		return "pal.pif";
		case OS_TV_NTSC:	return "ntsc.pif";
		case OS_TV_MPAL:	return "mpal.pif";
	}

	DAEDALUS_ASSERTMSG( "Unknown TV type" );
	return NULL;
}

//*****************************************************************************
// Load pif bootcode if it is available
//*****************************************************************************
static bool ROM_LoadPIF( u32 tv_type )
{
	const CHAR * pif_name;
	CHAR filename[ MAX_PATH ];
	FILE * fp;
	u32 op;

	pif_name = ROM_GetPIFName( tv_type );
	if ( !pif_name )
	{
		return false;
	}

	IO::Path::Combine( filename, g_szDaedalusExeDir, "Boot\\" );
	IO::Path::Append( filename, pif_name );

	fp = fopen( filename, "rb" );
	if ( fp )
	{
		u32 * pPifRom = (u32*)g_pMemoryBuffers[MEM_PIF_RAM];
		u32 i = 0;
		while (fread(&op, 4, 1, fp) == 1 && i < MemoryRegionSizes[MEM_PIF_RAM])
		{
			*pPifRom = SwapEndian( op );
			pPifRom++;
			i+=4;
		}

		fclose(fp);

		return true;
	}
	else
	{
		DBGConsole_Msg( 0, "Couldn't find '[W%s]', simulating boot", filename );
		return false;
	}
}

//*****************************************************************************
// Find out the CIC type
//*****************************************************************************
static ECicType ROM_GenerateCICType( const u8 * p_rom_base )
{
	u32		crc( daedalus_crc32(0, p_rom_base + RAMROM_BOOTSTRAP_OFFSET, RAMROM_GAME_OFFSET - RAMROM_BOOTSTRAP_OFFSET) );

	for ( u32 i = 0; i < NUM_CIC_CHIPS; i++ )
	{
		if ( crc == gCICInfo[ i ].BootCRC )
		{
			return static_cast< ECicType >( i );
		}
	}

	return CIC_UNKNOWN;
}

//*****************************************************************************
//
//*****************************************************************************
void ROM_ReBoot()
{
	// Do this here - probably not necessary as plugin loaded/unloaded with audio thread
	if (g_pAiPlugin != NULL)
		g_pAiPlugin->RomClosed();

	if (g_ROM.settings.ExpansionPakEnabled)
		Memory_Reset(MEMORY_8_MEG);
	else
		Memory_Reset(MEMORY_4_MEG);
	
	// Reset here
	CPU_Reset(0xbfc00000);
	RSP_Reset();
	
	g_dwNumFrames = 0;

#ifdef DAEDALUS_ENABLE_OS_HOOKS
	Patch_Reset();
#endif
	/*hr = */CController::Get()->OnRomOpen( g_ROM.settings.SaveType );

	//
	// Find out the CIC type and initialise various systems based on the CIC type
	//
	u8			rom_base[ RAMROM_GAME_OFFSET ];
	RomBuffer::GetRomBytesRaw( rom_base, 0, RAMROM_GAME_OFFSET );

	ECicType	cic_chip( ROM_GenerateCICType( rom_base ) );

	if (cic_chip == CIC_UNKNOWN)
	{
		//DAEDALUS_ASSERTMSG( DSPrintf( "Unknown CIC CRC: 0x%08x\nAssuming CIC-6102", crc ) );
		//DBGConsole_Msg(0, "[MUnknown CIC CRC: 0x%08x]", crc );
		DBGConsole_Msg(0, "[MUnknown CIC]" );
	}
	else
	{
		gCICInfo[ cic_chip ].CheckSumProc();
		DBGConsole_Msg(0, DSPrintf( "[MRom uses %s]", gCICInfo[ cic_chip ].Name ) );
	}

	// XXXX Update this rom's boot info

	g_ROM.nTvType = ROM_GetTvTypeFromID( g_ROM.rh.CountryID );

	if ( g_ROM.nTvType == OS_TV_NTSC )
	{
		CPU_SetTvFrequency( 50.0f );
	}
	else
	{
		CPU_SetTvFrequency( 60.0f );
	}

	if ( ROM_LoadPIF( g_ROM.nTvType ) )
	{
		ROM_RunPIFBoot( cic_chip );
	}
	else
	{
		ROM_SimulatePIFBoot( cic_chip, g_ROM.nTvType );
	}
	
	//CPU_AddBreakPoint( 0xA4000040 );
	//CPU_AddBreakPoint( 0xbfc00000 );

	DBGConsole_UpdateDisplay();
}

//*****************************************************************************
//
//*****************************************************************************
void ROM_Unload()
{
	RomBuffer::Finalise();
	Memory_Cleanup();
	CController::Get()->OnRomClose();
}



//*****************************************************************************
// Copy across text, null terminate, and strip spaces
//*****************************************************************************
void ROM_GetRomNameFromHeader( char * p_rom_name, const ROMHeader & header )
{
	memcpy( p_rom_name, header.Name, 20 );
	p_rom_name[20] = '\0';

	char * p( p_rom_name + (strlen(p_rom_name) -1) );		// -1 to skip null
	while (p >= p_rom_name && *p == ' ')
	{
		*p = 0;
		p--;
	}
}

//*****************************************************************************
//
//*****************************************************************************
bool ROM_LoadFile(const char * filename)
{
	RomID		rom_id;
	u32			rom_size;
	ECicType	boot_type;

	if(ROM_GetRomID( filename, &rom_id, &rom_size, &boot_type ))
	{
		RomSettings	settings;

		if ( !CRomSettingsDB::Get()->GetSettings( rom_id, &settings ) )
		{
			ROM_ResetSettings( &settings );

			// Don't add a new entry...
		}

		return ROM_LoadFile( filename, rom_id, settings );
	}
	else
	{
		return false;
	}
}

//*****************************************************************************
//
//*****************************************************************************
bool ROM_LoadFile(const char * filename, const RomID & rom_id, const RomSettings & settings )
{
	ROM_Unload();

	DBGConsole_Msg(0, "Reading rom image: [C%s]", filename);

	//
	//	This is a workaround for goldeneye. At the moment it won't
	//	work with dynamically loaded roms (i.e. the special mapping
	//	routines assume that the memory is contiguous)
	//
	bool	allow_dynamic_loading( true );

	if( rom_id == ROM_ID_GE_USA || rom_id == ROM_ID_GE_EUR )
	{
		DBGConsole_Msg( 0, "[MCompatibility:] Disabling dynamic loading of rom." );
		allow_dynamic_loading = false;
	}
	
	if ( !ROM_LoadCartToMemory( filename, allow_dynamic_loading ) )
	{
		DBGConsole_Msg( 0, "ROM_LoadCartToMemory(%s) Failed", filename );
		return false;
	}

	strncpy(g_ROM.szFileName, filename, MAX_PATH);
	g_ROM.szFileName[MAX_PATH-1] = '\0';
	
	// Get information about the rom header
	RomBuffer::GetRomBytesRaw( &g_ROM.rh, 0, sizeof(ROMHeader) );
	ROMFile::ByteSwap_3210( &g_ROM.rh, sizeof(ROMHeader) );
	
#ifdef _DEBUG
	RomID	header_id( g_ROM.rh.CRC1, g_ROM.rh.CRC2, g_ROM.rh.CountryID );
	DAEDALUS_ASSERT( header_id == rom_id, "Why is the rom id incorrect?" );
#endif

	// Copy across various bits
	ROM_GetRomNameFromHeader( g_ROM.settings.szGameName, g_ROM.rh );
	g_ROM.mRomID = rom_id;
	g_ROM.settings = settings;

	DumpROMInfo( g_ROM.rh );

	//
	//
	//
	g_bApplyPatches		= g_ROM.settings.PatchesEnabled;
	gCheckTextureHashFrequency = ROM_GetTexureHashFrequencyAsFrames( g_ROM.settings.CheckTextureHashFrequency );
	g_bSpeedSync		= g_ROM.settings.SpeedSyncEnabled;
	g_bUseDynarec		= g_ROM.settings.DynarecEnabled;

	DBGConsole_Msg(0, "[G%s]", g_ROM.settings.szGameName);
	DBGConsole_Msg(0, "This game has been certified as [G%s] (%s)", g_ROM.settings.szComment, g_ROM.settings.szInfo);
	DBGConsole_Msg(0, "ApplyPatches: [G%s]", g_bApplyPatches ? "on" : "off");
	DBGConsole_Msg(0, "Check Texture Hash Freq: [G%d]", gCheckTextureHashFrequency);
	DBGConsole_Msg(0, "SaveType: [G%s]", ROM_GetSaveTypeName( g_ROM.settings.SaveType ) );
	DBGConsole_Msg(0, "SpeedSync: [G%s]", g_bSpeedSync ? "on" : "off");
	DBGConsole_Msg(0, "DynaRec: [G%s]", g_bUseDynarec ? "on" : "off");

	//Patch_ApplyPatches();

	ROM_ReBoot();

	if (g_DaedalusConfig.RunAutomatically)
	{
		CHAR szReason[300+1];
		// Ignore failure
		CPU_StartThread(szReason, 300);
	}

	return true;
}

//*****************************************************************************
//
//*****************************************************************************
bool ROM_GetRomName( const char * filename, TCHAR * game_name )
{
	ROMFile * p_rom_file( ROMFile::Create( filename ) );
	if(p_rom_file == NULL)
	{
		return false;
	}

	COutputStringStream messages;

	if( !p_rom_file->Open( messages ) )
	{
		delete p_rom_file;
		return false;
	}

	// Only read in the header
	const u32	bytes_to_read( sizeof(ROMHeader) );
	u8 *		p_bytes;
	u32			buffer_size;
	if(!p_rom_file->LoadDataChunk( bytes_to_read, &p_bytes, &buffer_size, messages ))
	{
		// Lots of files don't have any info - don't worry about it
		delete p_rom_file;
		return false;
	}

	//
	//	Swap into native format
	//
	ROMFile::ByteSwap_3210( p_bytes, buffer_size );

	//
	// Get the address of the rom header
	// Setup the rom id and size
	//
	const ROMHeader * prh( reinterpret_cast<const ROMHeader *>( p_bytes ) );
	ROM_GetRomNameFromHeader( game_name, *prh );

	delete [] p_bytes;
	delete p_rom_file;
	return true;
}

//*****************************************************************************
// Initialise the RomInfo structure
//*****************************************************************************
bool ROM_GetRomID( const char * filename, RomID * p_id, u32 * p_rom_size, ECicType * p_boot_type )
{
	//
	// First of all, check if we have these details cached in the rom database
	//
	if ( CRomDB::Get()->QueryRom( filename, p_id, p_rom_size, p_boot_type ) )
	{
		return true;
	}

	ROMFile * p_rom_file( ROMFile::Create( filename ) );
	if(p_rom_file == NULL)
	{
		return false;
	}

	COutputStringStream messages;

	if( !p_rom_file->Open( messages ) )
	{
		delete p_rom_file;
		return false;
	}

	//
	// They weren't there - so we need to find this info out for ourselves
	// Only read in the header + bootcode
	//
	u32			bytes_to_read( RAMROM_GAME_OFFSET );
	u32			buffer_size;
	u8 *		p_bytes;
	if(!p_rom_file->LoadDataChunk( bytes_to_read, &p_bytes, &buffer_size, messages ))
	{
		// Lots of files don't have any info - don't worry about it
		delete p_rom_file;
		return false;
	}

	//
	//	Swap into native format
	//
	ROMFile::ByteSwap_3210( p_bytes, buffer_size );

	//
	// Get the address of the rom header
	// Setup the rom id and size
	//
	const ROMHeader * prh( reinterpret_cast<const ROMHeader *>( p_bytes ) );
	*p_id = RomID( prh->CRC1, prh->CRC2, prh->CountryID );
	*p_rom_size = p_rom_file->GetRomSize();
	if (*p_rom_size >= RAMROM_GAME_OFFSET)
	{
		*p_boot_type = ROM_GenerateCICType( p_bytes );
	}
	else
	{
		*p_boot_type = CIC_UNKNOWN;
	}

	//
	// Store this information in the rom database
	//
	CRomDB::Get()->AddRom( filename, *p_id, *p_rom_size, *p_boot_type );

	delete [] p_bytes;
	delete p_rom_file;
	return true;
}

//*****************************************************************************
// Read a cartridge into memory. this routine byteswaps if necessary
//*****************************************************************************
bool ROM_LoadCartToMemory( const char * filename, bool allow_dynamic_loading )
{
	COutputStringStream messages;
	if(!RomBuffer::Initialise( filename, allow_dynamic_loading, messages ))
	{
#ifdef DAEDALUS_W32
		CMainWindow::Get()->MessageBox( messages.c_str(), g_szDaedalusName, MB_ICONERROR | MB_OK );
#endif
		return false;
	}

	return true;
}

//*****************************************************************************
//
//*****************************************************************************
void ROM_CheckSumCic6101()
{
	CDebugConsole::Get()->Msg( 0, "[MUnable to check CRC for CIC-6101]" );	
}

//*****************************************************************************
//
//*****************************************************************************
void ROM_CheckSumCic6102()
{
	u32 dwAddress;
	u32 a1;
	u32 t7;
	u32 v1 = 0;
	u32 t0 = 0;
	u32 v0 = 0xF8CA4DDC; //(CIC_6102_BOOT_CIC * 0x5d588b65) + 1;
	u32 a3 = 0xF8CA4DDC;
	u32 t2 = 0xF8CA4DDC;
	u32 t3 = 0xF8CA4DDC;
	u32 s0 = 0xF8CA4DDC;
	u32 a2 = 0xF8CA4DDC;
	u32 t4 = 0xF8CA4DDC;
	u32 t8, t6, a0;

	CDebugConsole::Get()->MsgOverwriteStart();

	for (dwAddress = 0; dwAddress < 0x00100000; dwAddress+=4)
	{
		if ((dwAddress % 0x2000) == 0)
		{
			CDebugConsole::Get()->MsgOverwrite(0, "Generating CRC [M%d / %d]",
				dwAddress, 0x00100000);
		}
		v0 = RomBuffer::ReadValueRaw< u32 >( dwAddress + RAMROM_GAME_OFFSET );
		v1 = a3 + v0;
		a1 = v1;
		if (v1 < a3) 
			t2++;
	
		v1 = v0 & 0x001f;
		t7 = 0x20 - v1;
		t8 = (v0 >> (t7&0x1f));
		t6 = (v0 << (v1&0x1f));
		a0 = t6 | t8;

		a3 = a1;
		t3 ^= v0;
		s0 += a0;
		if (a2 < v0)
			a2 ^= a3 ^ v0;
		else
			a2 ^= a0;

		t0 += 4;
		t7 = v0 ^ s0;
		t4 += t7;
	}
	CDebugConsole::Get()->MsgOverwrite(0, "Generating CRC [M%d / %d]",
		0x00100000, 0x00100000);

	CDebugConsole::Get()->MsgOverwriteEnd();

	a3 ^= t2 ^ t3;	// CRC1
	s0 ^= a2 ^ t4;	// CRC2

	if (a3 != RomBuffer::ReadValueRaw< u32 >( 0x10 ) ||
		s0 != RomBuffer::ReadValueRaw< u32 >( 0x14 ))
	{
		DBGConsole_Msg(0, "[MWarning, CRC values don't match, fixing %08x %08x]", a3, s0);

		RomBuffer::WriteValueRaw< u32 >( 0x10, a3 );
		RomBuffer::WriteValueRaw< u32 >( 0x14, s0 );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void ROM_CheckSumCic6103()
{
	CDebugConsole::Get()->Msg( 0, "[MUnable to check CRC for CIC-6103]" );	
}

//*****************************************************************************
//
//*****************************************************************************
void ROM_CheckSumCic6104()
{
	CDebugConsole::Get()->Msg( 0, "[MUnable to check CRC for CIC-6104]" );	
}


//*****************************************************************************
// Thanks Lemmy!
//*****************************************************************************
void ROM_CheckSumCic6105()
{
	u32 dwAddress;
	u32 dwAddress2;
	u32 t5 = 0x00000020;
	u32 a3 = 0xDF26F436;
	u32 t2 = 0xDF26F436;
	u32 t3 = 0xDF26F436;
	u32 s0 = 0xDF26F436;
	u32 a2 = 0xDF26F436;
	u32 t4 = 0xDF26F436;
	u32 v0, v1, a1, a0;

	dwAddress2 = 0;

	CDebugConsole::Get()->MsgOverwriteStart();

	for (dwAddress = 0; dwAddress < 0x00100000; dwAddress += 4)
	{
		if ((dwAddress % 0x2000) == 0)
			CDebugConsole::Get()->MsgOverwrite(0, "Generating CRC [M%d / %d]",
				dwAddress, 0x00100000);

		v0 = RomBuffer::ReadValueRaw< u32 >( dwAddress + RAMROM_GAME_OFFSET );
		v1 = a3 + v0;
		a1 = v1;
		
		if (v1 < a3)
			t2++;

		v1 = v0 & 0x1f;
		a0 = (v0 >> (t5-v1)) | (v0 << v1);
		a3 = a1;
		t3 = t3 ^ v0;
		s0 += a0;
		if (a2 < v0)
			a2 ^= a3 ^ v0;
		else
			a2 ^= a0;

		t4 += RomBuffer::ReadValueRaw< u32 >( dwAddress2 + 0x750 ) ^ v0;
		dwAddress2 = (dwAddress2 + 4) & 0xFF;

	}

	a3 ^= t2 ^ t3;
	s0 ^= a2 ^ t4;

	CDebugConsole::Get()->MsgOverwrite(0, "Generating CRC [M%d / %d]",
		0x00100000, 0x00100000);

	CDebugConsole::Get()->MsgOverwriteEnd();

	if (a3 != RomBuffer::ReadValueRaw< u32 >( 0x10 ) ||
		s0 != RomBuffer::ReadValueRaw< u32 >( 0x14 ))
	{
		DBGConsole_Msg(0, "[MWarning, CRC values don't match, fixing]");

		RomBuffer::WriteValueRaw< u32 >( 0x10, a3 );
		RomBuffer::WriteValueRaw< u32 >( 0x14, s0 );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void ROM_CheckSumCic6106()
{
	CDebugConsole::Get()->Msg( 0, "[MUnable to check CRC for CIC-6106]" );	
}

//*****************************************************************************
// Reset to default values
//*****************************************************************************
void	ROM_ResetSettings( RomSettings * p_settings )
{
	strcpy(p_settings->szGameName, "");
	strcpy(p_settings->szComment, "");
	strcpy(p_settings->szInfo, "");

	p_settings->PatchesEnabled = true;
	p_settings->CheckTextureHashFrequency = THF_EVERY_5;
	p_settings->SpeedSyncEnabled = true;
	p_settings->DynarecEnabled = true;
	p_settings->ExpansionPakEnabled = false;
	p_settings->SaveType = SAVE_TYPE_UNKNOWN;
	p_settings->dwRescanCount = 0;
}

//*****************************************************************************
// Association between a country id value, tv type and name
//*****************************************************************************
struct CountryIDInfo
{
	s8				CountryID;
	const char *	CountryName;
	u32				TvType;
};

static const CountryIDInfo g_CountryCodeInfo[] = 
{
	{  0,  "0",			OS_TV_NTSC },
	{ '7', "Beta",		OS_TV_NTSC },
	{ 'A', "NTSC",		OS_TV_NTSC },
	{ 'D', "Germany",	OS_TV_PAL },
	{ 'E', "USA",		OS_TV_NTSC },
	{ 'F', "France",	OS_TV_PAL },
	{ 'I', "Italy",		OS_TV_PAL },
	{ 'J', "Japan",		OS_TV_NTSC },
	{ 'P', "Europe",	OS_TV_PAL },
	{ 'S', "Spain",		OS_TV_PAL },
	{ 'U', "Australia", OS_TV_PAL },
	{ 'X', "PAL",		OS_TV_PAL },
	{ 'Y', "PAL",		OS_TV_PAL }
};



//*****************************************************************************
// Get a string representing the country name from an ID value
//*****************************************************************************
const char * ROM_GetCountryNameFromID( u8 country_id )
{
	for ( u32 i = 0; i < ARRAYSIZE( g_CountryCodeInfo ); i++)
	{
		if ( g_CountryCodeInfo[i].CountryID == country_id )
		{
			return g_CountryCodeInfo[i].CountryName;
		}
	}
	return "?";
}


//*****************************************************************************
//
//*****************************************************************************
u32 ROM_GetTvTypeFromID( u8 country_id )
{
	for ( u32 i = 0; ARRAYSIZE( g_CountryCodeInfo ); i++ )
	{
		if ( g_CountryCodeInfo[i].CountryID == country_id )
		{
			return g_CountryCodeInfo[i].TvType;
		}
	}
	return OS_TV_NTSC;
}


//*****************************************************************************
// Get the name of a save type from an ESaveType enum
//*****************************************************************************
const char * ROM_GetSaveTypeName( ESaveType save_type )
{
	switch ( save_type )
	{
		case SAVE_TYPE_UNKNOWN:		return "";
		case SAVE_TYPE_EEP4K:		return "Eeprom 4k";
		case SAVE_TYPE_EEP16K:		return "Eeprom 16k";
		case SAVE_TYPE_SRAM:		return "SRAM";
		case SAVE_TYPE_FLASH:		return "FlashRam";
	}

	return "?";
}

//*****************************************************************************
//
//*****************************************************************************
const char *	ROM_GetCicTypeName( ECicType cic_type )
{
	if ( cic_type >= 0 && cic_type < NUM_CIC_CHIPS )
	{
		return gCICInfo[ cic_type ].Name;
	}

	return "?";
}

//*****************************************************************************
//
//*****************************************************************************
static const u32 gTextureHashFreqeuncies[] =
{
	0,	//THF_DISABLED = 0,
	1,	//THF_EVERY_FRAME,
	2,	//THF_EVERY_2,
	5,	//THF_EVERY_5,
	10,	//THF_EVERY_10,
	30,	//THF_EVERY_30,
};

static const char * const gTextureHashFreqeuncyDescriptions[] =
{
	"Disabled",			//THF_DISABLED = 0,
	"Every Frame",		//THF_EVERY_FRAME,
	"Every 2 Frames",	//THF_EVERY_2,
	"Every 5 Frames",	//THF_EVERY_5,
	"Every 10 Frames",	//THF_EVERY_10,
	"Every 30 Frames",	//THF_EVERY_30,
};

//*****************************************************************************
//
//*****************************************************************************
u32	ROM_GetTexureHashFrequencyAsFrames( ETextureHashFrequency thf )
{
	if(thf >= 0 && thf < NUM_THF)
	{
		return gTextureHashFreqeuncies[ thf ];
	}

	return 0;
}

//*****************************************************************************
//
//*****************************************************************************
ETextureHashFrequency	ROM_GetTextureHashFrequencyFromFrames( u32 frames )
{
	for( u32 i = 0; i < NUM_THF; ++i )
	{
		if( frames <= gTextureHashFreqeuncies[ i ] )
		{
			return ETextureHashFrequency( i );
		}
	}

	return THF_EVERY_30;	// Return the maximum
}


//*****************************************************************************
//
//*****************************************************************************
const char * ROM_GetTextureHashFrequencyDescription( ETextureHashFrequency thf )
{
	if(thf >= 0 && thf < NUM_THF)
	{
		return gTextureHashFreqeuncyDescriptions[ thf ];
	}

	return "?";
}


