/*

  Copyright (C) 2001 StrmnNrmn

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "StdAfx.h"

#include "Core\ROM.h"

#include "RomDB.h"

#include <vector>

static const u64 ROMDB_MAGIC_NO	= 0x42444D5244454144LL; //DAEDRMDB		// 44 41 45 44 52 4D 44 42
static const u32 ROMDB_CURRENT_VERSION = 2;

//*****************************************************************************
// Class implementations
//*****************************************************************************
class IRomDB : public CRomDB
{
	public:
		IRomDB();

		//
		// CRomDB implementation
		//
		bool			OpenDB( const CHAR * filename );
		void			Reset();
		bool			Commit();

		void			AddRom( const CHAR * filename, const RomID & id, u32 rom_size, ECicType cic_type );
		bool			QueryRom( const CHAR * filename, RomID * p_id, u32 * p_rom_size, ECicType * p_cic_type ) const;

	private:
		s32				FindRom( const CHAR * filename ) const;

	private:
		struct RomDBEntry
		{
			RomDBEntry() {}
			RomDBEntry( const CHAR * filename, const RomID & id, u32 rom_size, ECicType cic_type )
			{
				memset( FileName, 0, sizeof( FileName ) );
				strcpy( FileName, filename );
				ID = id;
				RomSize = rom_size;
				CicType = cic_type;
			}

			CHAR		FileName[ MAX_PATH + 1 ];
			RomID		ID;
			u32			RomSize;
			ECicType	CicType;

		};

		CHAR							mRomDBFileName[ MAX_PATH + 1 ];
		std::vector< RomDBEntry >		mRomDB;

};

//*****************************************************************************
// Singleton creator
//*****************************************************************************
template<> bool	CSingleton< CRomDB >::Create()
{
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	mpInstance = new IRomDB();

	return true;
};

//*****************************************************************************
// Constructor
//*****************************************************************************
IRomDB::IRomDB()
{
	mRomDBFileName[ 0 ] = '\0';

	mRomDB.reserve( 100 );
}

//*****************************************************************************
//
//*****************************************************************************
bool	IRomDB::OpenDB( const CHAR * filename )
{
	u32 num_read;

	//
	// Remember the filename
	//
	strcpy( mRomDBFileName, filename );

	FILE * fh = fopen( filename, "rb" );

	if ( !fh )
		return false;
	
	//
	// Check the magic number
	//
	u64 magic;
	num_read = fread( &magic, sizeof( magic ), 1, fh );
	if ( num_read != 1 || magic != ROMDB_MAGIC_NO )
	{
		fclose( fh );
		return false;
	}

	//
	// Check the version number
	//
	u32 version;
	num_read = fread( &version, sizeof( version ), 1, fh );
	if ( num_read != 1 || version != ROMDB_CURRENT_VERSION )
	{
		fclose( fh );
		return false;
	}

	while ( true )
	{
		RomDBEntry entry;

		num_read = fread( &entry, sizeof( entry ), 1, fh );
		if ( num_read == 1 )
		{
			mRomDB.push_back( entry );
		}
		else
		{
			break;
		}
	}

	fclose( fh );

	return true;
}

//*****************************************************************************
//
//*****************************************************************************
void	IRomDB::Reset()
{
	mRomDB.clear();
}

//*****************************************************************************
//
//*****************************************************************************
bool	IRomDB::Commit()
{
	//
	// Check if we have a valid filename
	// 
	if ( strlen( mRomDBFileName ) <= 0 )
		return false;

	FILE * fh = fopen( mRomDBFileName, "wb" );

	if ( !fh )
		return false;

	//
	// Write the magic
	//
	fwrite( &ROMDB_MAGIC_NO, sizeof( ROMDB_MAGIC_NO ), 1, fh );

	//
	// Write the version
	//
	fwrite( &ROMDB_CURRENT_VERSION, sizeof( ROMDB_CURRENT_VERSION ), 1, fh );
	
	for ( u32 i = 0; i < mRomDB.size(); i++ )
	{
		fwrite( &mRomDB[ i ], sizeof( RomDBEntry ), 1, fh );
	}

	fclose( fh );

	return true;
}

//*****************************************************************************
//
//*****************************************************************************
s32		IRomDB::FindRom( const CHAR * filename ) const
{
	for ( u32 i = 0; i < mRomDB.size(); i++ )
	{
		if ( stricmp( mRomDB[ i ].FileName, filename ) == 0 )
		{
			//
			// Match (we may want to canonicalise the filename first?)
			//
			return i;
		}
	}

	//
	// No match
	//
	return -1;
}

//*****************************************************************************
//
//*****************************************************************************
void	IRomDB::AddRom( const CHAR * filename, const RomID & id, u32 rom_size, ECicType cic_type )
{
	RomDBEntry new_entry( filename, id, rom_size, cic_type );

	s32 i = FindRom( filename );

	if ( i < 0 )
	{
		// Add a new entry
		mRomDB.push_back( new_entry );
	}
	else
	{
		// Update the existing entry
		mRomDB[ i ] = new_entry;
	}
}

//*****************************************************************************
//
//*****************************************************************************
bool	IRomDB::QueryRom( const CHAR * filename, RomID * p_id, u32 * p_rom_size, ECicType * p_cic_type ) const
{
	s32 i = FindRom( filename );

	if ( i >= 0 )
	{
		*p_id = mRomDB[ i ].ID;
		*p_rom_size = mRomDB[ i ].RomSize;
		*p_cic_type = mRomDB[ i ].CicType;
		return true;
	}
	else
	{
		// No entry
		return false;
	}
}



