/*
Copyright (C) 2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "NativeTexture.h"

#include "VideoMemoryManager.h"
#include "DaedMathUtil.h"

#include <png.h>

#include <pspgu.h>
#include <pspdebug.h>
#include <pspkernel.h>



namespace
{

//
//	This is slower, but necessary because it deals with heights of < 8
//
void swizzle_slow(u8* out, const u8* in, unsigned int width, unsigned int height)
{
	unsigned int i,j;
	unsigned int rowblocks = (width / 16);

	for (j = 0; j < height; ++j)
	{
		for (i = 0; i < width; ++i)
		{
			unsigned int blockx = i / 16;
			unsigned int blocky = j / 8;

			unsigned int x = (i - blockx*16);
			unsigned int y = (j - blocky*8);
			unsigned int block_index = blockx + ((blocky) * rowblocks);
			unsigned int block_address = block_index * 16 * 8;

			out[block_address + x + y * 16] = in[i+j*width];
		}
	}
}

void swizzle_fast(u8* out, const u8* in, unsigned int width, unsigned int height)
{
	unsigned int blockx, blocky;
	unsigned int j;

	unsigned int width_blocks = (width / 16);
	unsigned int height_blocks = (height / 8);

	unsigned int src_pitch = (width-16)/4;
	unsigned int src_row = width * 8;

	const u8* ysrc = in;
	u32* dst = (u32*)out;

	for (blocky = 0; blocky < height_blocks; ++blocky)
	{
		const u8* xsrc = ysrc;
		for (blockx = 0; blockx < width_blocks; ++blockx)
		{
			const u32* src = (u32*)xsrc;
			for (j = 0; j < 8; ++j)
			{
				*(dst++) = *(src++);
				*(dst++) = *(src++);
				*(dst++) = *(src++);
				*(dst++) = *(src++);
				src += src_pitch;
			}
			xsrc += 16;
		}
		ysrc += src_row;
	}
}

bool swizzle(u8* out, const u8* in, unsigned int width, unsigned int height)
{
	if(height < 8)
	{
		//swizzle_slow( out, in, width, height );

		memcpy( out, in, width * height );
		return false;
	}
	else
	{
		swizzle_fast( out, in, width, height );
		return true;
	}
}

//
//	Swizzle the texture, explicitly setting the texture colour to the specified value (alpha is left alone!)
//
void swizzle_fast_setcolour(u8* out, const u8* in, unsigned int width, unsigned int height, u32 mask, u32 colour)
{
	unsigned int blockx, blocky;
	unsigned int j;

	unsigned int width_blocks = (width / 16);
	unsigned int height_blocks = (height / 8);

	unsigned int src_pitch = (width-16)/4;
	unsigned int src_row = width * 8;

	const u8* ysrc = in;
	u32* dst = (u32*)out;

	const u32	MASK = mask;
	const u32	INV_MASK  = ~mask;

	// For speed, clear out the bits we're not setting
	u32 just_colour( colour & MASK );

	for (blocky = 0; blocky < height_blocks; ++blocky)
	{
		const u8* xsrc = ysrc;
		for (blockx = 0; blockx < width_blocks; ++blockx)
		{
			const u32* src = (u32*)xsrc;
			for (j = 0; j < 8; ++j)
			{
				*(dst++) = (*(src++) & INV_MASK) | just_colour;
				*(dst++) = (*(src++) & INV_MASK) | just_colour;
				*(dst++) = (*(src++) & INV_MASK) | just_colour;
				*(dst++) = (*(src++) & INV_MASK) | just_colour;
				src += src_pitch;
			}
			xsrc += 16;
		}
		ysrc += src_row;
	}
}
}

//*****************************************************************************
//
//*****************************************************************************
void	CNativeTexture::CorrectDimensions( u32 width, u32 height, u32 & adjusted_width, u32 & adjusted_height )
{
	static const u32 MIN_TEXTURE_WIDTH = 1;
	static const u32 MIN_TEXTURE_HEIGHT = 1;

	adjusted_width = Max( GetNextPowerOf2( width ), MIN_TEXTURE_WIDTH );
	adjusted_height = Max( GetNextPowerOf2( height ), MIN_TEXTURE_HEIGHT );
}

//*****************************************************************************
//
//*****************************************************************************
CNativeTexture *	CNativeTexture::Create( u32 width, u32 height )
{
	u32		adjusted_width;
	u32		adjusted_height;

	CorrectDimensions( width, height, adjusted_width, adjusted_height );

	return new CNativeTexture( adjusted_width, adjusted_height, 4 );
}

//*****************************************************************************
//
//*****************************************************************************
CNativeTexture *	CNativeTexture::CreateFromRaw( const char * p_filename, u32 width, u32 height )
{
	u32		adjusted_width;
	u32		adjusted_height;

	CorrectDimensions( width, height, adjusted_width, adjusted_height );

	//
	// Create a texture (do this first to help minimise memory fragmentation)
	//
	CNativeTexture *	p_texture( new CNativeTexture( adjusted_width, adjusted_height, 4 ) );

	//
	//	Read in the data (taking care that it's width/height is different from what we expect)
	//
	FILE * fh( fopen( p_filename, "rb" ) );
	if(fh != NULL)
	{
		u32 *	p_data( new u32[ adjusted_width * adjusted_height ] );
		for(u32 y = 0; y < height; ++y)
		{
			u32 *	p_row( &p_data[ y * adjusted_width ] );

			fread( p_row, width, sizeof(u32), fh );

			// Rest of the row is garbage
		}

		fclose( fh );
		p_texture->SetData( p_data, adjusted_width, adjusted_height );

		delete [] p_data;

	}
	else
	{
		p_texture->Release();
		p_texture = NULL;
	}
	
	return p_texture;
}

//*****************************************************************************
//
//*****************************************************************************
CNativeTexture *	CNativeTexture::CreateFromAlpha( const char * p_filename, u32 width, u32 height )
{
	CNativeTexture *	p_texture( new CNativeTexture( width, height, 4 ) );
	if(p_texture != NULL)
	{
		FILE * fh( fopen( p_filename, "rb" ) );
		if(fh != NULL)
		{
			u32 *	p_data( p_texture->mpData );
			u8 *	p_row( new u8[ width ] );

			for(u32 y = 0; y < height; ++y)
			{
				fread( p_row, width, 1, fh );
				for(u32 x = 0; x < width; ++x)
				{
					*(p_data++) = ((u32)p_row[x] << 24) | 0x00ffffff;
				}
			}

			delete [] p_row;
			fclose( fh );
		}

		p_texture->mIsSwizzled = false;
	}

	return p_texture;
}


//*****************************************************************************
//
//*****************************************************************************
CNativeTexture::CNativeTexture( u32 w, u32 h, u8 channels ) : mWidth( w ), mHeight( h ), mIsSwizzled( true ), mPixelSize( channels )
{
	u32 * pdata( (u32*)CVideoMemoryManager::Get()->Alloc( mWidth * mHeight * sizeof(u32) ) );
	if(pdata != NULL)
	{
		mpData = pdata;
		mIsVidMem = true;
	}
	else
	{
		mpData = new u32[ mWidth * mHeight ];
		mIsVidMem = false;
	}
}

//*****************************************************************************
//
//*****************************************************************************
CNativeTexture::CNativeTexture( u32 w, u32 h ) : mWidth( w ), mHeight( h ), mIsSwizzled( true ), mPixelSize( 4 )
{
	u32 * pdata( (u32*)CVideoMemoryManager::Get()->Alloc( mWidth * mHeight * mPixelSize ) );
	if(pdata != NULL)
	{
		mpData = pdata;
		mIsVidMem = true;
	}
	else
	{
		mpData = new u32[ mWidth * mHeight ];
		mIsVidMem = false;
	}

}

//*****************************************************************************
//
//*****************************************************************************
CNativeTexture::~CNativeTexture()
{
	if(mIsVidMem)
	{
		CVideoMemoryManager::Get()->Free( mpData );
	}
	else
	{
		delete [] mpData;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CNativeTexture::InstallTexture()
{
	sceGuEnable(GU_TEXTURE_2D);
	sceGuTexMode(GU_PSM_8888,0,0,mIsSwizzled ? 1 : 0);		// maxmips/a2/swizzle = 0
	sceGuTexImage(0,mWidth,mHeight,mWidth,mpData);
}

//*****************************************************************************
//	Thanks 71M/Shazz
//*****************************************************************************
bool	CNativeTexture::LoadPng( const char * p_filename )
{
	const size_t	signature_size = 8;
	u8	signature[ signature_size ];

	FILE * fh( fopen( p_filename,"rb" ) );
	if(fh == NULL)
	{
		return false;
	}

	if (fread( signature, sizeof(u8), signature_size, fh ) != signature_size)
	{
		fclose(fh);
		return false;
	}

	if ( !png_check_sig( signature, signature_size ) )
	{
		return false;
	}

	png_struct * p_png_struct( png_create_read_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL ) );
	if ( p_png_struct == NULL)
	{
		return false;
	}

	png_info * p_png_info( png_create_info_struct( p_png_struct ) );
	if ( p_png_info == NULL )
	{
		png_destroy_read_struct( &p_png_struct, NULL, NULL );
		return false;
	}

	if ( setjmp( p_png_struct->jmpbuf ) != 0 )
	{
		png_destroy_read_struct( &p_png_struct, NULL, NULL );
		return false;
	}

	png_init_io( p_png_struct, fh );
	png_set_sig_bytes( p_png_struct, signature_size );
	png_read_png( p_png_struct, p_png_info, PNG_TRANSFORM_STRIP_16 | PNG_TRANSFORM_PACKING | PNG_TRANSFORM_EXPAND | PNG_TRANSFORM_BGR, NULL );

	png_uint_32 width = p_png_info->width;
	png_uint_32 height = p_png_info->height;
	int color_type = p_png_info->color_type;

	if(mWidth < width || mHeight < height)
	{
		return false;
	}

	png_byte **p_row_table = p_png_info->row_pointers;
	unsigned char r=0, g=0, b=0, a=0;

	u8 * p_dest = (u8 *) new u8[mWidth*height*4];		// Just need to make sure width is correct
	u8 * bitmap_ptr = p_dest;

	for ( u32 y = 0; y < height; ++y )
	{
		const png_byte * pRow = p_row_table[ y ];

		for ( u32 x = 0; x < width; ++x )
		{
			switch ( color_type )
			{
			case PNG_COLOR_TYPE_GRAY:
				r = g = b = *pRow++;
				if ( r == 0 && g == 0 && b == 0 )	a = 0x00;
				else								a = 0xff;
				break;
			case PNG_COLOR_TYPE_GRAY_ALPHA:
				r = g = b = *pRow++;
				if ( r == 0 && g == 0 && b == 0 )	a = 0x00;
				else								a = 0xff;
				pRow++;
				break;
			case PNG_COLOR_TYPE_RGB:
				b = *pRow++;
				g = *pRow++;
				r = *pRow++;
				if ( r == 0 && g == 0 && b == 0 )	a = 0x00;
				else								a = 0xff;
				break;
			case PNG_COLOR_TYPE_RGB_ALPHA:
				b = *pRow++;
				g = *pRow++;
				r = *pRow++;
				a = *pRow++;
				break;
			}

			p_dest[ ( x << 2 ) + 0 ] = r;
			p_dest[ ( x << 2 ) + 1 ] = g;
			p_dest[ ( x << 2 ) + 2 ] = b;
			p_dest[ ( x << 2 ) + 3 ] = a;
		}

		p_dest += ( mWidth * 4 );
	}

	SetData( bitmap_ptr, mWidth, height );
	delete [] bitmap_ptr;

	png_destroy_read_struct( &p_png_struct, &p_png_info, NULL );

	fclose(fh);

	return true;
}

//*****************************************************************************
//
//*****************************************************************************
void	CNativeTexture::SetData( void * p_data, u32 w, u32 h )
{
	DAEDALUS_ASSERT_Q( w == mWidth );
	DAEDALUS_ASSERT_Q( h == mHeight );

	mIsSwizzled = swizzle( reinterpret_cast< u8 * >( mpData ), reinterpret_cast< u8 * >( p_data ), mWidth*mPixelSize, mHeight );

	// We need to flush out the data from the cache if this isn't in vid mem
	if( !mIsVidMem )
	{
		sceKernelDcacheWritebackAll();
		sceGuTexFlush();
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CNativeTexture::SetDataForceColour( void * p_data, u32 w, u32 h, u32 mask, u32 colour )
{
	DAEDALUS_ASSERT_Q( w == mWidth );
	DAEDALUS_ASSERT_Q( h == mHeight );

	swizzle_fast_setcolour( reinterpret_cast< u8 * >( mpData ), reinterpret_cast< u8 * >( p_data ), mWidth*mPixelSize, mHeight, mask, colour );
	mIsSwizzled = true;

	// We need to flush out the data from the cache if this isn't in vid mem
	if( !mIsVidMem )
	{
		sceKernelDcacheWritebackAll();
		sceGuTexFlush();
	}
}

//*****************************************************************************
//
//*****************************************************************************
u32	CNativeTexture::GetVideoMemoryUsage() const
{
	if( mIsVidMem )
	{
		return mWidth * mHeight * mPixelSize;
	}

	return 0;
}

//*****************************************************************************
//
//*****************************************************************************
u32	CNativeTexture::GetSystemMemoryUsage() const
{
	if( !mIsVidMem )
	{
		return mWidth * mHeight * mPixelSize;
	}

	return 0;
}
