/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __DAEDALUS_D3DRENDER_H__
#define __DAEDALUS_D3DRENDER_H__

#include "FFXMaths.h"
#include "Vector2.h"
#include "Vector3.h"
#include "Vector4.h"
#include "Matrix4x4.h"
#include "DaedSingleton.h"
#include "DaedalusVtx.h"

#include <set>

class CTexture;

struct CDaedalusViewport
{
	u32			X;
	u32			Y;
	u32			Width;
	u32			Height;
    float		MinZ;
    float		MaxZ;
};

struct DaedalusLight
{
	v3 dir;		// Should be normalised
	c4 col;
	u32 initialised;
};

//*****************************************************************************
//
//*****************************************************************************
class PSPRenderer : public CSingleton< PSPRenderer >
{
protected:
	friend class CSingleton< PSPRenderer >;
	PSPRenderer();

public:
	~PSPRenderer();

	void				BeginScene();
	void				EndScene();

	bool				InvalidateDeviceObjects();
	bool				RestoreDeviceObjects();
	void				Reset();

	// Verious rendering states
	void				SetTextureEnable(bool bEnable)			{ m_bTextureEnable = bEnable; }
	void				SetPrimitiveColor(u32 colour)			{ m_dwPrimitiveColor = colour; }
	void				SetEnvColor(u32 colour)					{ m_dwEnvColor = colour; }
	void				SetLighting(bool bLighting)				{ m_bLightingEnable = bLighting; }
	void				ZBufferEnable(bool bZBuffer)			{ m_bZBuffer = bZBuffer; }

	void				SetFogMult( float fFogMult )			{ m_fFogMult = fFogMult; }
	void				SetFogOffset( float fFogOffset )		{ m_fFogOffset = fFogOffset; }
	void				SetFogEnable(bool bEnable)				{ m_bFogEnable = bEnable; }
	void				SetFogColor(u32 colour)					{ m_dwFogColor = colour; }

	void				SetCullMode( bool bCullFront, bool bCullBack );
	void				SetSmooth( bool bSmooth )				{ mSmooth = bSmooth; }
	void				SetSmoothShade( bool bSmoothShade )		{ mSmoothShade = bSmoothShade; }
	void				SetAlphaRef(u32 dwAlpha);

	// Texture stuff
	enum ETextureAddressMode
	{
		TAM_CLAMP = 0,
		TAM_MIRROR,
		TAM_WRAP,
		TAM_NUM_ADDRESS_MODES,
	};

	void				SetTexture( u32 index, CTexture* lpsTexture, s32  nTileLeft, s32 nTileTop, u32 dwTileWidth, u32 dwTileHeight);
	void				SetTextureGen(bool bTextureGen);
	void				SetTextureScale(float fScaleX, float fScaleY);
	void				SetAddressMode( u32 index, ETextureAddressMode mode_u, ETextureAddressMode mode_v );

	// Viewport stuff
	void				SetD3DViewport( u32 dwX, u32 dwY, u32 dwWidth, u32 dwHeight );
	void				SetN64Viewport( const v3 & scale, const v3 & trans );
	void				SetScissor( s32 x0, s32 y0, s32 x1, s32 y1 );

	// Matrix stuff
	enum EMatrixLoadStyle
	{
		MATRIX_LOAD = 0,
		MATRIX_MUL = 1
	};

	void				ResetMatrices();
	void				PrintActive();
	void				SetProjection(const Matrix4x4 & mat, bool bPush, EMatrixLoadStyle load_style);
	void				SetWorldView(const Matrix4x4 & mat, bool bPush, EMatrixLoadStyle load_style);
	void				PopProjection();
	void				PopWorldView();

	// Vertex stuff
	void				SetNewVertexInfo(u32 dwAddress, u32 dwV0, u32 dwNum);	// Assumes dwAddress has already been checked!	
	void				SetNewVertexInfoDKR(u32 dwAddress, u32 dwV0, u32 dwNum);	// Assumes dwAddress has already been checked!	
	void				SetVtxColor( u32 vert, u32 color );
	void				SetVtxTextureCoord( u32 vert, short tu, short tv );
	void				SetVtxXY( u32 vert, float x, float y );

	// TextRect stuff
	bool				TexRect( const v2 & xy0, const v2 & xy1, const v2 & uv0, const v2 & uv1 );
	bool				TexRectFlip( const v2 & xy0, const v2 & xy1, const v2 & uv0, const v2 & uv1 );
	bool				FillRect( const v2 & xy0, const v2 & xy1, u32 color );
		
	// Returns true if triangle visible and rendered, false otherwise
	bool				AddTri(u32 v0, u32 v1, u32 v2);

	bool				FlushTris();
	
	void				SetNumLights(u32 dwNumLights)			{ m_dwNumLights = dwNumLights; }
	void				SetLightCol(u32 dwLight, u32 colour);
	void				SetLightDirection(u32 l, float x, float y, float z);
	void				SetAmbientLight( float r, float g, float b);

	bool				TestVerts( u32 v0, u32 vn ) const;				// Returns true if bounding volume is visible, false if culled
	const DaedalusVtx &	GetTransformedVtx( u32 i ) const		{ return m_VtxTransformed[ i ]; }
	u32					GetVtxFlags( u32 i ) const				{ return m_dwVecFlags[ i ]; }

	// Rendering stats
	u32					GetNumTrisRendered() const				{ return m_dwNumTrisRendered; }
	u32					GetNumTrisClipped() const				{ return m_dwNumTrisClipped; }


	// Debugging
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	void					SetRecordCombinerStates( bool enable )	{ mRecordCombinerStates = enable; }					// Sets whether combiner states will be recorded for the subsequent frames
	const std::set<u64> &	GetRecordedCombinerStates() const		{ return mRecordedCombinerStates; }

	bool				IsCombinerStateUnhandled( u64 state ) const	{ return mUnhandledCombinderStates.find( state ) != mUnhandledCombinderStates.end(); }
	bool				IsCombinerStateDisabled( u64 state ) const	{ return mDisabledCombinerStates.find( state ) != mDisabledCombinerStates.end(); }
	void				DisableCombinerState( u64 state )			{ mDisabledCombinerStates.insert( state ); }
	void				EnableCombinerState( u64 state )			{ mDisabledCombinerStates.erase( state ); }

	void				ToggleDisableCombinerState( u64 state )		{ if( IsCombinerStateDisabled( state )) { EnableCombinerState(state); } else { DisableCombinerState( state ); } }

#endif

private:
	void				InitBlendMode();
	void				InitCycleCopy( );
	void				InitCycleFill( );
	void				InitCycle1( );
	void				InitCycle2( );
	void				BodgeBlendMode( u32 num_cycles );
	void				PrepareVerticesForBlendMode( DaedalusVtx * p_sw_vertices, u32 num_verts );


	void				SetTextureFactor(u32 colour);
	bool				TestTri(u32 v0, u32 v1, u32 v2) const;			// Returns true if ok, false if clipped
	u32					LightVert( const v3 & norm );

	bool				RenderTriangleList( const DaedalusVtx * p_verts, u32 num_verts, const u16 * p_indices, u32 num_indices, bool disable_zbuffer = false, bool force_point_filter = false );
	bool				RenderSprites( const DaedalusVtx * p_verts, u32 num_verts, bool disable_zbuffer, bool force_point_filter );

private:
	enum { MAX_VERTS = 80 };		// F3DLP.Rej supports up to 80 verts!

	CDaedalusViewport	m_VP;

	float				m_fScreenMultX;
	float				m_fScreenMultY;

	v3					mVpScale;
	v3					mVpTrans;

	bool				m_bTextureEnable;
	bool				m_bLightingEnable;
	
	c4					m_AmbientCol;

	s32					m_dwNumLights;
	u32					m_dwTextureFactor;

	bool				m_bZBuffer;

	bool				m_bCullFront;
	bool				m_bCullBack;
	
	u32					m_dwAlpha;

	bool				mSmooth;
	bool				mSmoothShade;

	float				m_fFogMult;
	float				m_fFogOffset;

	bool				m_bFogEnable;
	u32					m_dwFogColor;
	u32					m_dwPrimitiveColor;
	u32					m_dwEnvColor;

	// Texutring
	CTexture *			m_lpsTexture[2];
	DaedalusLight		m_Lights[8];

	float				m_fTileWidth[2];		// Float to avoid converts when processing verts
	float				m_fTileHeight[2];
	float				m_fTexWidth[2];
	float				m_fTexHeight[2];		// Float to avoid converts when processing verts

	// Reciprocals for fast mults
	float				m_fInvTileWidth[2];
	float				m_fInvTileHeight[2];
	float				m_fInvTexWidth[2];
	float				m_fInvTexHeight[2];

	float				m_fTexLeft[2];
	float				m_fTexTop[2];
	float				m_fTexScaleX;
	float				m_fTexScaleY;

	bool				m_bTextureGen;			// Generate texture coords?


	static const u32 MATRIX_STACK_SIZE = 60;

	Matrix4x4			m_mProjection[MATRIX_STACK_SIZE];
	Matrix4x4			m_mModelView[MATRIX_STACK_SIZE];
	u32					m_dwProjectionTop;
	u32					m_dwModelViewTop;
		
	u16					m_swIndexBuffer[500];
	u32					m_dwNumVertices;
	u32					m_dwNumIndices;

	// Processed vertices waiting for output...
	DaedalusVtx			m_VtxTransformed[MAX_VERTS];
	v4					mVtxProjected[MAX_VERTS];			// Transformed and projected vertices (suitable for clipping etc)
	u32					m_dwVecFlags[MAX_VERTS];			// Z_POS Z_NEG etc

	//
	// Stats
	//
	u32					m_dwNumTrisRendered;
	u32					m_dwNumTrisClipped;

	// Debugging
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	bool				mRecordCombinerStates;
	std::set< u64 >		mRecordedCombinerStates;
	std::set< u64 >		mDisabledCombinerStates;
#endif
	std::set< u64 >		mUnhandledCombinderStates;

};

#endif // __DAEDALUS_D3DRENDER_H__
