/*
 * PSP Software Development Kit - http://www.pspdev.org
 * -----------------------------------------------------------------------
 * Licensed under the BSD license, see LICENSE in PSPSDK root for details.
 *
 * Copyright (c) 2005 Jesper Svennevid
 */

#include "stdafx.h"

//namespace psp
//{
#include <psptypes.h>
#include <pspkernel.h>
#include <pspdisplay.h>
#include <pspdebug.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include <pspgu.h>
#include <pspgum.h>
#include <pspctrl.h>
#include <pspsdk.h>
#include <psprtc.h>
#include <psppower.h>


#include "Debug\DBGConsole.h"
#include "Debug\Debug.h"
#include "Interface\RomDB.h"
#include "Core\RomSettings.h"
#include "Core\Memory.h"
#include "Core\CPU.h"
#include "Core\PIF.h"
#include "Input\InputManager.h"
#include "Utility\Profiler.h"
#include "DaedTiming.h"
#include "DaedIO.h"

#include "SysPSP\UI\UIContext.h"
#include "SysPSP\UI\MainMenuScreen.h"
#include "PSPGraphics\VideoMemoryManager.h"
#include "PSPGraphics\GraphicsContext.h"
#include "PSPGraphics\DrawText.h"
#include "PSPGraphics\DaedalusGraphics.h"
#include "PSPGraphics\TextureCache.h"

#include <stdarg.h>

TCHAR		g_szDaedalusExeDir[MAX_PATH+1] = "";
TCHAR		g_szDaedalusName[] = "Daedalus";
TCHAR		g_szDaedalusIniFile[MAX_PATH+1] = "daedalus.ini";
TCHAR		g_szDaedalusLogFile[MAX_PATH+1] = "daedalus.txt";

#define FAILED(hr)	(s32(hr) < 0)

static int  WaitForCPUThread(void);
static void DisplayRomsAndChoose();
static int DoMain();

#define BOOTLOADER_NAME "Daedalus"

#if defined( DAEDALUS_KERNEL_MODE )

PSP_MODULE_INFO( BOOTLOADER_NAME, 0x1000, 1, 1 );
PSP_MAIN_THREAD_ATTR( 0 );

#else

PSP_MODULE_INFO( BOOTLOADER_NAME, 0, 1, 1 );
PSP_MAIN_THREAD_ATTR( THREAD_ATTR_USER );

#endif

extern bool PSPGraphics_ExecuteCommand( const CHAR * pszCommand, void ** ppResult );



void * __cdecl operator new( size_t count )
{
	void * p_mem( malloc( count ) );
	if(p_mem == NULL)
	{
		char msg[ 1024 ];
		sprintf( msg, "Out of memory (operator new(%d)", count );
		DAEDALUS_ASSERTMSG( msg );
	}

	return p_mem;
}


void *__cdecl operator new[]( size_t count )
{
	void * p_mem( malloc( count ) );
	if(p_mem == NULL)
	{
		char msg[ 1024 ];
		sprintf( msg, "Out of memory (operator new[](%d)", count );
		DAEDALUS_ASSERTMSG( msg );
	}

	return p_mem;
}

void __cdecl operator delete[]( void * p_mem )
{
	if( p_mem != NULL )
	{
		free( p_mem );
	}
}

void __cdecl operator delete( void * p_mem )
{
	if( p_mem != NULL )
	{
		free( p_mem );
	}
}

// XXXX 
#define PSP_LINE_SIZE 512
#define PSP_PIXEL_FORMAT 3

#if defined( DAEDALUS_KERNEL_MODE )
void MyExceptionHandler(PspDebugRegBlock *regs)
{
	pspDebugScreenPrintf("\n\n");
	pspDebugDumpException(regs);
}
/**
 * Function that is called from _init in kernelmode before the
 * main thread is started in usermode.
 */

__attribute__ ((constructor))
void loaderInit()
{
    pspKernelSetKernelPC();
	pspDebugScreenInit();
	pspDebugInstallKprintfHandler(NULL);
	pspDebugInstallErrorHandler(MyExceptionHandler);
	pspSdkInstallNoDeviceCheckPatch();
	pspSdkInstallNoPlainModuleCheckPatch();
}
#endif

/* Exit callback */
int exit_callback(int arg1, int arg2, void *common)
{
	sceKernelExitGame();
	return 0;
}

/* Callback thread */
int CallbackThread(SceSize args, void *argp)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);

	sceKernelSleepThreadCB();

	return 0;
}

/* Sets up the callback thread and returns its thread id */
int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}


extern "C"
{
int main(void)
{
	pspDebugScreenInit();

	pspDebugScreenPrintf( "Entering DoMain()\n" );
	DoMain();
	pspDebugScreenPrintf( "Exited DoMain()\n" );

	sceKernelExitGame();
	return 0;
}
}

int WaitForCPUThread(void)
{
	while ( true )
	{
		pspDebugScreenPrintf( "Sleeping!\n" );
		sceKernelSleepThread();
	}

	return 0;
}

extern "C" { void _DisableFPUExceptions(); } 

int DoMain()
{
	int nResult;
	BOOL bResult;

// Disable for profiling
//	srand(time(0));

	pspDebugScreenPrintf( "Disabling FPU Exceptions\n" );

	_DisableFPUExceptions();

	SetupCallbacks();

	pspDebugScreenPrintf( "Callbacks done\n" );

	if (scePowerSetClockFrequency(333, 333, 166) == 0)
	{
		pspDebugScreenPrintf( "CPU set to 333MHz\n" );
	}
	else
	{
		pspDebugScreenPrintf( "Could not set CPU to 333MHz\n" );
	}

	pspDebugScreenPrintf( "vram base address: %p\n", sceGeEdramGetAddr() );
	pspDebugScreenPrintf( "vram size: %d bytes\n", sceGeEdramGetSize() );

    //setup Pad
    sceCtrlSetSamplingCycle(0);
    sceCtrlSetSamplingMode(PSP_CTRL_MODE_ANALOG);
	
	//
	// Read configuration before doing anything major
	//
//	ReadConfiguration();

	pspDebugScreenPrintf( "In DoMain()\n" );

	//
	// In ReleaseWithLogging mode, open the logfile for writing
	//
	#ifdef DAEDALUS_LOG
	hr = Debug_InitLogging();
	if (FAILED(hr))
		return 1;
	#endif

	pspDebugScreenPrintf( "DebugConsole up\n" );

	if (!CDebugConsole::Create())
		return 1;

	DBGConsole_Msg( 0, "ROMDB create" );
	
	//
	// Create and load the RomDB
	//
	if (!CRomDB::Create())
		return 1;

	DBGConsole_Msg( 0, "ROMDB init" );

	char romdb_filename[ MAX_PATH + 1 ];
	IO::Path::Combine( romdb_filename, g_szDaedalusExeDir, "rom.db" );
	bResult = CRomDB::Get()->OpenDB( romdb_filename );
	if ( !bResult )
	{
		// That's ok - it probably didn't exist
	}

	// Init the savegame directory
	strcpy( g_DaedalusConfig.szSaveDir, "SaveGames/" );

	//
	// Create and load the CRomSettingsDB singleton
	//
	DBGConsole_Msg( 0, "CRomSettingsDB create" );

	if (!CRomSettingsDB::Create())
		return 1;

	DBGConsole_Msg( 0, "CRomSettingsDB init" );

	bResult = CRomSettingsDB::Get()->OpenIniFile( g_szDaedalusIniFile );
	if (!bResult)
	{
		//MessageBox(NULL, "Unable to read IniFile from disk", g_szDaedalusName, MB_OK);
		//return 1;
	}

	//
	// Init memory before Console (which tries to display
	// things that are refs to memory)
	//
	DBGConsole_Msg( 0, "Memory init" );

	if (!Memory_Init())
		return 1;

	//
	// Create the console if it's enabled. Don't care about failures
	//
	CDebugConsole::Get()->EnableConsole( g_DaedalusConfig.ShowDebug ? true : false );
//	DisplayDisclaimer();
//	DisplayConfig();

	//
	// Create the Profiler singleton
	//
#ifdef DAEDALUS_ENABLE_PROFILING
	if (!CProfiler::Create())
		return 1;
#endif

	//
	// Create the input manager BEFORE the window is created and initialised
	//
	DBGConsole_Msg( 0, "InputManager init" );

	if (!CInputManager::Create())
		return 1;

	if(!CInputManager::Get()->Initialise())
		return 1;

	DBGConsole_Msg( 0, "Controller init" );

	if (!CController::Create())
		return 1;

	//
	// Create the main window singleton
	//
//	hr = CMainWindow::Create();
//	if (FAILED(hr))
//		return 1;
	if(!CVideoMemoryManager::Create())
		return false;

	if(!CGraphicsContext::Create())
		return false;

	CDrawText::Initialise();
	
	DisplayRomsAndChoose();

	//
	// Create/Show the main window
	//
//	hr = CMainWindow::Get()->CreateWnd( 0 );
//	if (FAILED(hr))
//		return 1;

	DBGConsole_Msg( 0, "Running Main" );

	// Wrap the main processing in an exception handler
	// This should at least let us save the config/inifile
	// if something goes wrong.
	{
		nResult = WaitForCPUThread();
	}

	//
	// Write current config out to the registry
	//
//	WriteConfiguration();

	//
	// Unload any loaded rom and cleanup
	//
	ROM_Unload();

	//
	// Turn off the debug console
	// 
	CDebugConsole::Get()->EnableConsole( false );

	Memory_Fini();

	//
	// Destroy the singletons
	//
//	CMainWindow::Destroy();
	CController::Destroy();
	CInputManager::Destroy();
#ifdef DAEDALUS_ENABLE_PROFILING
	CProfiler::Destroy();
#endif

	#ifdef DAEDALUS_LOG
	Debug_FinishLogging();
	#endif

	//
	// Destroy the inifile.
	// Will write ini file if changed
	//
	CRomSettingsDB::Destroy();

	//
	// Destroy / Commit the rom database
	//
	CRomDB::Get()->Commit();
	CRomDB::Destroy();

	CDebugConsole::Destroy();

	return nResult;
}

///////////////////////////////////////////////////////////////////////////////////////



static bool gTimerInitialised( false );
static u64	gLastTime( 0 );
static f32	gFreqInv( 1.0f );


float	GetElapsedTime()
{
	u64			now;
	NTiming::GetPreciseTime( &now );

	if(!gTimerInitialised)
	{
		u64		freq;
		NTiming::GetPreciseFrequency( &freq );
		gLastTime = now;
		gFreqInv = 1.0f / f32( freq );
		gTimerInitialised = true;
	}

	u64			elapsed_ticks( now - gLastTime );
	gLastTime = now;

	return f32(elapsed_ticks) * gFreqInv;
}

void ResetElapsedTime()
{
	u64			now;
	NTiming::GetPreciseTime( &now );

	gLastTime = now;
}


bool	gDumpNextScreen = false;

bool	gMenuActive = false;


#ifdef DAEDALUS_KERNEL_MODE
bool	gProfileNextFrame = false;
bool	gWasProfilingFrame = false;
bool	gShowProfilerResults = false;
#endif

#ifdef DAEDALUS_DEBUG_DISPLAYLIST
static bool	gShowTextureCacheUsage = false;
extern bool gSoftwareClipping;
extern bool gDebugDisplayList;
#include "PSPGraphics\DLParser.h"
#endif

void HandleEndOfFrame()
{
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	if(gDebugDisplayList)
		return;
#endif

#ifdef DAEDALUS_KERNEL_MODE
	//
	//	Update the profiling stats
	//
	if(gWasProfilingFrame)
	{
		// Show results
		pspDebugProfilerDisable();
		gShowProfilerResults = true;
		gWasProfilingFrame = false;

		// Drop back into the menu
		gMenuActive = true;
	}
#endif

#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	if( gShowTextureCacheUsage )
	{
		pspDebugScreenSetXY(0, 0);
		CTextureCache::Get()->DisplayStats();
	}
#endif

	if( gDumpNextScreen )
	{
		CGraphicsContext::Get()->DumpScreenShot();
		gDumpNextScreen = false;
	}

	//
	//	Figure out how long the last frame took
	//
	float		elapsed_time( GetElapsedTime() );
	float		framerate( 0.0f );
	if(elapsed_time > 0)
	{
		framerate = 1.0f / elapsed_time;
	}

	// Temp dynarec stats
#ifdef _DEBUG
	extern u64 gTotalInstructionsEmulated;
	extern u64 gTotalInstructionsExecuted;
	extern u32 gTotalRegistersCached;
	extern u32 gTotalRegistersUncached;
	extern u32 gFragmentLookupSuccess;
	extern u32 gFragmentLookupFailure;

	u32		dynarec_ratio( 0 );

	if(gTotalInstructionsExecuted + gTotalInstructionsEmulated > 0)
	{
		float fRatio = float(gTotalInstructionsExecuted * 100.0f / float(gTotalInstructionsEmulated+gTotalInstructionsExecuted));

		dynarec_ratio = u32( fRatio );

		//gTotalInstructionsExecuted = 0;
		//gTotalInstructionsEmulated = 0;
	}

	u32		cached_regs_ratio( 0 );
	if(gTotalRegistersCached + gTotalRegistersUncached > 0)
	{
		float fRatio = float(gTotalRegistersCached * 100.0f / float(gTotalRegistersCached+gTotalRegistersUncached));

		cached_regs_ratio = u32( fRatio );
	}

	const char * const TERMINAL_SAVE_CURSOR			= "\033[s";
	const char * const TERMINAL_RESTORE_CURSOR		= "\033[u";
//	const char * const TERMINAL_TOP_LEFT			= "\033[2A\033[2K";
	const char * const TERMINAL_TOP_LEFT			= "\033[H\033[2K";

	printf( TERMINAL_SAVE_CURSOR );
	printf( TERMINAL_TOP_LEFT );

	printf( "Frame: %dms, DynaRec %d%%, Regs cached %d%%, Lookup success %d/%d", u32(elapsed_time * 1000.0f), dynarec_ratio, cached_regs_ratio, gFragmentLookupSuccess, gFragmentLookupFailure );

	printf( TERMINAL_RESTORE_CURSOR );
	fflush( stdout );

	gFragmentLookupSuccess = 0;
	gFragmentLookupFailure = 0;
#endif

	//
	//	Enter the debug menu as soon as select is newly pressed
	//
    SceCtrlData pad;

	static u32 oldButtons = 0;

	sceCtrlReadBufferPositive(&pad, 1);
	if(oldButtons != pad.Buttons)
	{
		if(pad.Buttons & PSP_CTRL_SELECT)
		{
			gMenuActive = true;
		}
	}

	// Allow display list rendering to be disabled to see time taken
	//if(pad.Buttons & PSP_CTRL_CROSS)
	//{
		//extern bool	gSkipNextDisplayList;
		//gSkipNextDisplayList = true;
	//}

	oldButtons = pad.Buttons;

#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	bool	show_clipping_mode( false );
	bool	debug_displaylist( false );
#endif

	if(gMenuActive)
	{
		bool menu_button_pressed( false );

		pspDebugScreenSetTextColor( 0xffffffff );

		CController::Get()->FlushSaveGames();

		CGraphicsContext::Get()->SetDebugScreenTarget( CGraphicsContext::TS_FRONTBUFFER );

		// Remain paused until the Select button is pressed again
		while(!menu_button_pressed)
		{
			sceDisplayWaitVblankStart();
			pspDebugScreenSetXY(0, 0);
			pspDebugScreenPrintf( "Framerate: %#.2f   \n", framerate );

#ifdef DAEDALUS_KERNEL_MODE
			if(gShowProfilerResults)
			{
				pspDebugProfilerPrint();
			}
			if(gProfileNextFrame)
			{
				pspDebugScreenPrintf( "Will profile next frame on resume\n" );
			}
#endif
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
			if( gShowTextureCacheUsage )
			{
				CTextureCache::Get()->DisplayStats();
			}

			if(show_clipping_mode)
			{
				pspDebugScreenPrintf( "Software clipping %s\n", gSoftwareClipping ? "enabled" : "disabled" );
			}
			if(debug_displaylist)
			{
				pspDebugScreenPrintf( "Will debug next displaylist on resume" );
			}
#endif
			if(gDumpNextScreen)
			{
				pspDebugScreenPrintf( "Will grab screenshot of next frame on resume\n" );
			}

			sceCtrlReadBufferPositive(&pad, 1);
			if(oldButtons != pad.Buttons)
			{
				if(pad.Buttons & PSP_CTRL_SELECT)
				{
					menu_button_pressed = true;
				}
#ifdef DAEDALUS_KERNEL_MODE
				if(pad.Buttons & PSP_CTRL_CIRCLE)
				{
					gProfileNextFrame = !gProfileNextFrame;
				}
#endif
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
				if(pad.Buttons & PSP_CTRL_TRIANGLE)
				{
					gShowTextureCacheUsage = !gShowTextureCacheUsage;
				}
				if(pad.Buttons & PSP_CTRL_LTRIGGER)
				{
					gSoftwareClipping = !gSoftwareClipping;
					show_clipping_mode = true;
				}
				if(pad.Buttons & PSP_CTRL_RTRIGGER)
				{
					debug_displaylist = true;
				}
#endif
#ifdef _DEBUG
				if(pad.Buttons & PSP_CTRL_CROSS)
				{
					CPU_DumpFragmentCache();
				}
				if(pad.Buttons & PSP_CTRL_CIRCLE)
				{
					CPU_ResetFragmentCache();
				}
#endif

				// Need to do this every frame
				if(pad.Buttons & PSP_CTRL_SQUARE)
				{
					gDumpNextScreen = !gDumpNextScreen;
				}
			}

			oldButtons = pad.Buttons;
		}

		gMenuActive = false;

		CGraphicsContext::Get()->SetDebugScreenTarget( CGraphicsContext::TS_BACKBUFFER );

	}

#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	if(debug_displaylist)
	{
		PSPGraphics_ExecuteCommand( DAEDALUS_GFX_DEBUGDL, NULL );
	}
#endif


	//
	//	Reset the elapsed time to avoid glitches when we restart
	//
	ResetElapsedTime();

	//
	//	Check whether to profile the next frame
	//
#ifdef DAEDALUS_KERNEL_MODE
	gShowProfilerResults = false;
	if(gProfileNextFrame)
	{
		gProfileNextFrame = false;
		gWasProfilingFrame = true;

		pspDebugProfilerClear();
		pspDebugProfilerEnable();
	}
#endif
}

void DisplayRomsAndChoose()
{
	CUIContext *	p_context( CUIContext::Create() );

	if(p_context != NULL)
	{
		CMainMenuScreen *	p_main_menu( CMainMenuScreen::Create() );
		p_main_menu->Run( p_context );
		delete p_main_menu;
	}

	delete p_context;
}

