/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

#include "PSPRenderer.h"
#include "Texture.h"
#include "NativeTexture.h"
#include "RDP.h"
#include "PolyClipper.h"
#include "PolyTesselator.h"
#include "GraphicsContext.h"
#include "PSPColour.h"

#include "DebugDisplayList.h"

#include "Debug\Dump.h"
#include "Debug\DBGConsole.h"

#include "Core\Memory.h"			// We access the memory buffers
#include "Core\ROM.h"

#include "DaedIO.h"

#include "Utility\Profiler.h"
#include "PngUtil.h"

#include <pspgu.h>
#include <pspgum.h>
#include <psputils.h>
#include <pspdebug.h>
#include <pspctrl.h>
#include <pspdisplay.h>

#include <vector>

float DECAL_Z_OFFSET = +3.1f;		// Found through trial an error for the PSP

#include "PushStructPack1.h"
struct FiddledVtx
{
	s16 y;
	s16 x;

	s16 flag;
	s16 z;

	s16 tv;
	s16 tu;

	union
	{
		struct
		{
			u8 rgba_a;
			u8 rgba_b;
			u8 rgba_g;
			u8 rgba_r;
		};
		struct
		{
			s8 norm_a;
			s8 norm_z;	// b
			s8 norm_y;	// g
			s8 norm_x;	// r
		};
	};
};

DAEDALUS_STATIC_ASSERT( sizeof(FiddledVtx) == 16 );


struct FiddledVtxDKR
{
	s16 y;
	s16 x;

	u8 a;
	u8 b;
	s16 z;

	u8 g;
	u8 r;
};

#include "PopStructPack.h"


// Bits for clipping
// +-+-+-
// xxyyzz
#define Z_NEG  0x01
#define Z_POS  0x02
#define Y_NEG  0x04
#define Y_POS  0x08
#define X_NEG  0x10
#define X_POS  0x20


//#define SOFTWARE_CLIPPING
#define USE_NATIVE_PROJECTION
//#define FORCE_FLAT_SHADING
//#define TEXRECT_USE_3D_TRANSFORM

#ifdef USE_NATIVE_PROJECTION
bool		gSoftwareClipping( false );
#else
bool		gSoftwareClipping( true );
#endif

bool		gTesselateLargeTriangles( true );

#define GL_TRUE                           1
#define GL_FALSE                          0

#undef min
#undef max

enum CycleType
{
	CYCLE_1CYCLE = 0,		// Please keep in this order - matches RDP 
	CYCLE_2CYCLE,
	CYCLE_COPY,
	CYCLE_FILL,
};


static u32 sViWidth = 320;
static u32 sViHeight = 240;

static const u32 gPlaceholderTextureWidth( 16 );
static const u32 gPlaceholderTextureHeight( 16 );
static const float gFillRectDepth( 0.0f );
static const float gTexRectDepth( 0.0f );

static const float gMaxZValue( 1.0f);

u32		__attribute__((aligned(16)))	gWhiteTexture[gPlaceholderTextureWidth * gPlaceholderTextureHeight ];
u32		__attribute__((aligned(16)))	gPlaceholderTexture[gPlaceholderTextureWidth * gPlaceholderTextureHeight ];
u32		__attribute__((aligned(16)))	gSelectedTexture[gPlaceholderTextureWidth * gPlaceholderTextureHeight ];

//
//	If this flag is set, we override the diffuse colour (or just alpha) in the vertex with the specified colour
//
bool	gbModulateDiffuseColour = false;
u32		gSetDiffuseMask = 0;
u32		gSetDiffuseColour = 0xffffffff;

u32		gModulateDiffuseMask = 0;
u32		gModulateDiffuseColour = 0xffffffff;


static const u32 COL32_MASK_A = 0xff000000;
static const u32 COL32_MASK_B = 0x00ff0000;
static const u32 COL32_MASK_G = 0x0000ff00;
static const u32 COL32_MASK_R = 0x000000ff;

static const u32 COL32_MASK_RGB = 0x00ffffff;
static const u32 COL32_MASK_RGBA = 0x00ffffff;


extern void HandleEndOfFrame();

const u32 COL32_BLACK( MAKE_COLOUR_U32( 0,0,0 ) );
const u32 COL32_WHITE( MAKE_COLOUR_U32( 255, 255, 255 ) );



DaedalusVtx		DaedalusVtx::Interpolate( const DaedalusVtx & rhs, float factor ) const
{
	DaedalusVtx		out;

	out.pos = pos + (rhs.pos - pos) * factor;
	out.color = InterpolateColour( color, rhs.color, factor );
	out.t0 = t0 + (rhs.t0 - t0) * factor;

	return out;
}

//*****************************************************************************
// Creator function for singleton
//*****************************************************************************
template<> bool CSingleton< PSPRenderer >::Create()
{
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	mpInstance = new PSPRenderer();
	if (!mpInstance)
	{
		return false;
	}
	
	return true;
}

enum MuxVal
{
	MUX_0 = 0,
	MUX_1,
	MUX_COMBINED,
	MUX_TEXEL0,
	MUX_TEXEL1,
	MUX_PRIM,
	MUX_SHADE,
	MUX_ENV,
	MUX_COMBALPHA,
	MUX_T0_ALPHA,
	MUX_T1_ALPHA,
	MUX_PRIM_ALPHA,
	MUX_SHADE_ALPHA,
	MUX_ENV_ALPHA,
	MUX_LODFRAC,
	MUX_PRIMLODFRAC,
	MUX_UNK,
	MUX_PRIM_MOD_ENV,
	MUX_PRIM_ENV_MEAN
};

//*****************************************************************************
//
//*****************************************************************************
PSPRenderer::PSPRenderer() :
	m_fScreenMultX(2.0f),
	m_fScreenMultY(2.0f),

	m_bTextureEnable(false),
	m_bLightingEnable(false),

	m_dwNumLights(0),
	m_dwTextureFactor(MAKE_COLOUR_U32(255,255,255,255)),
	m_bZBuffer(false),
		
	m_bCullFront(false),
	m_bCullBack(true),

	m_dwAlpha(0xFF),

	mSmooth( true ),
	mSmoothShade( true ),
	
	m_fFogMult(0.0f),			// * 0 + 0 = 0
	m_fFogOffset(0.0f),
	m_bFogEnable(false),
	m_dwFogColor(0x00FFFFFF),

	m_bTextureGen(false),

	m_dwProjectionTop(0),
	m_dwModelViewTop(0),

	m_dwNumVertices(0),
	m_dwNumIndices(0),

	m_dwNumTrisRendered( 0 ),
	m_dwNumTrisClipped( 0 )
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
,	mRecordCombinerStates( false )
#endif
{
	for ( u32 t = 0; t < 2; t++ )
	{
		m_lpsTexture[t] = NULL;
		m_fTileWidth[t] = 32.0f;
		m_fTileHeight[t] = 32.0f;
		m_fTexWidth[t] = 32.0f;
		m_fTexHeight[t] = 32.0f;
		m_fInvTileWidth[t] = 1/32.0f;
		m_fInvTileHeight[t] = 1/32.0f;
		m_fInvTexWidth[t] = 1/32.0f;
		m_fInvTexHeight[t] = 1/32.0f;
	}

	memset( m_Lights, 0, sizeof(m_Lights) );

	m_fTexScaleX = 1.0f;
	m_fTexScaleY = 1.0f;

	m_AmbientCol.r = m_AmbientCol.g = m_AmbientCol.b = m_AmbientCol.a = 1.0f;

	gRDPMux._u64 = 0;

	memset( gWhiteTexture, 0xff, sizeof(gWhiteTexture) );

	u32	texel_idx( 0 );
	const u32	COL_MAGENTA( MAKE_COLOUR_U32( 255, 0, 255 ) );
	const u32	COL_GREEN( MAKE_COLOUR_U32( 0, 255, 0 ) );
	const u32	COL_BLACK( MAKE_COLOUR_U32( 0,0,0 ) );
	for(u32 y = 0; y < gPlaceholderTextureHeight; ++y)
	{
		for(u32 x = 0; x < gPlaceholderTextureWidth; ++x)
		{
			gPlaceholderTexture[ texel_idx ] = ((x&1) == (y&1)) ? COL_MAGENTA : COL_BLACK;
			gSelectedTexture[ texel_idx ]    = ((x&1) == (y&1)) ? COL_GREEN   : COL_BLACK;

			texel_idx++;
		}
	}

	RestoreDeviceObjects();
}

//*****************************************************************************
//
//*****************************************************************************
PSPRenderer::~PSPRenderer()
{	
	InvalidateDeviceObjects();
}
	
//*****************************************************************************
//
//*****************************************************************************
bool PSPRenderer::InvalidateDeviceObjects()
{
	for ( u32 i = 0; i < 2; i++ )
	{
		if ( m_lpsTexture[i] != NULL )
		{
			m_lpsTexture[i]->Release();
			m_lpsTexture[i] = NULL;
		}
	}

	return true;
}


//*****************************************************************************
//
//*****************************************************************************
bool PSPRenderer::RestoreDeviceObjects()
{
	// Initialise the device to our default state

	// We do our own culling
	sceGuDisable(GU_CULL_FACE);

	// But clip our tris please
	sceGuEnable(GU_CLIP_PLANES);
	//sceGuDisable(GU_CLIP_PLANES);

	sceGuScissor(0,0, 480,272);
	sceGuEnable(GU_SCISSOR_TEST);

	// We do our own lighting
	sceGuDisable(GU_LIGHTING);

	sceGuAlphaFunc(GU_GEQUAL, 0x04, 0xff );
	sceGuEnable(GU_ALPHA_TEST);
	sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
	sceGuEnable(GU_BLEND);

	// Default is ZBuffer disabled
	sceGuDepthMask(GL_TRUE);	// GL_TRUE to disable z-writes
	sceGuDepthFunc(GU_GEQUAL);		// GEQUAL?
	sceGuDisable(GU_DEPTH_TEST);

	// Initialise all the renderstate to our defaults.
	sceGuShadeModel(GU_SMOOTH);

	sceGuTexEnvColor(m_dwTextureFactor);
	sceGuTexOffset(0.0f,0.0f);

	//sceGuFog(near,far,m_dwFogColor);
	// Texturing stuff
	sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);
	sceGuTexFilter(GU_LINEAR,GU_LINEAR);
	sceGuTexWrap(GU_REPEAT,GU_REPEAT);

	sceGumMatrixMode(GU_PROJECTION);
	sceGumLoadIdentity();
#ifdef USE_NATIVE_PROJECTION
	// Init here?
#else
	sceGumOrtho(0, 480, 272, 0, 0.0f, -1.0f);
#endif

	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();

	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();

	return true;		
}

//*****************************************************************************
//
//*****************************************************************************
// Reset for a new frame
void	PSPRenderer::Reset()
{
	ResetMatrices();

	m_dwNumVertices = 0;
	m_dwNumIndices = 0;

	m_dwNumTrisRendered = 0;
	m_dwNumTrisClipped = 0;

	u32 dwScaleX = Memory_VI_GetRegister( VI_X_SCALE_REG ) & 0xFFF;
	u32 dwScaleY = Memory_VI_GetRegister( VI_Y_SCALE_REG ) & 0xFFF;

	float fScaleX = (float)dwScaleX / (1<<10);
	float fScaleY = (float)dwScaleY / (1<<10);

	//DBGConsole_Msg(DEBUG_VI, "VI_X_SCALE_REG set to 0x%08x (%f)", dwValue, 1/fScale);
	//DBGConsole_Msg(DEBUG_VI, "VI_Y_SCALE_REG set to 0x%08x (%f)", dwValue, 1/fScale);

	u32 dwHStartReg = Memory_VI_GetRegister( VI_H_START_REG );
	u32 dwVStartReg = Memory_VI_GetRegister( VI_V_START_REG );

	u32	hstart = dwHStartReg >> 16;
	u32	hend = dwHStartReg & 0xffff;
	//DBGConsole_Msg( 0, "h start/end %x %x", hstart, hend );		// 128 725 - 597

	u32	vstart = dwVStartReg >> 16;
	u32	vend = dwVStartReg & 0xffff;
	//DBGConsole_Msg( 0, "v start/end %x %x", vstart, vend );		// 56 501 - 445

	sViWidth  = (u32)( (hend-hstart)    * fScaleX);
	sViHeight = (u32)(((vend-vstart)/2) * fScaleY);				// 

	if ( sViWidth == 0 )
	{
		sViWidth = 320;
	}
	if ( sViHeight == 0 )
	{
		sViHeight = 240;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::BeginScene()
{
	CGraphicsContext::Get()->BeginFrame();

	RestoreDeviceObjects();

	extern bool			gNeedZBufferUdate;
	gNeedZBufferUdate = true;

#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	mRecordedCombinerStates.clear();
#endif
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::EndScene()
{
	CGraphicsContext::Get()->EndFrame();

	HandleEndOfFrame();
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetD3DViewport( u32 dwX, u32 dwY, u32 dwWidth, u32 dwHeight )
{
	m_VP.X = dwX;
	m_VP.Y = dwY;
	m_VP.Width = dwWidth;
	m_VP.Height = dwHeight;
	m_VP.MinZ = 0.0f;
	m_VP.MaxZ = 1.0f;

//	g_pD3DDev->SetViewport( &m_VP );

	m_fScreenMultX = (float)m_VP.Width / (float)(sViWidth);
	m_fScreenMultY = (float)m_VP.Height / (float)(sViHeight);
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetN64Viewport( const v3 & scale, const v3 & trans )
{
	mVpScale = scale;
	mVpTrans = trans;

	mVpScale.x *= (float)m_fScreenMultX;
	mVpScale.y *= -(float)m_fScreenMultY;		// Apply the flip here to invert the y to the correct value
	mVpScale.z *= 1.0f / 256.0f;

	mVpTrans.x *= (float)m_fScreenMultX;
	mVpTrans.y *= (float)m_fScreenMultY;
	mVpTrans.z *= 1.0f / 256.0f;
}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::InitCycleCopy( )
{
	DL_PF("CycleType: Copy");

	gSetDiffuseMask = 0;
	gSetDiffuseColour = 0xffffffff;
	gModulateDiffuseMask = 0;
	gModulateDiffuseColour = 0xffffffff;

	sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
	if(m_lpsTexture[0] != NULL)
	{
		CNativeTexture *	p_texture( m_lpsTexture[0]->GetTexture() );
		if(p_texture != NULL)
		{
			p_texture->InstallTexture();
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::InitCycleFill()
{
	DL_PF("CycleType: Fill");

	gSetDiffuseMask = 0;
	gSetDiffuseColour = 0xffffffff;
	gModulateDiffuseMask = 0;
	gModulateDiffuseColour = 0xffffffff;

	// XXXX
	// Need to select diffuse colour here - ignore the texture!
	// Need to remove the texture 
	//sceGuDisable(GU_TEXTURE_2D);
	sceGuEnable(GU_TEXTURE_2D);
	sceGuTexMode(GU_PSM_8888,0,0,GL_TRUE);		// maxmips/a2/swizzle = 0
	sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gWhiteTexture);
	sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
}

static const char * sc_colcombtypes32[32] =
{
	"Combined    ", "Texel0      ",
	"Texel1      ", "Primitive   ", 
	"Shade       ", "Env         ",
	"1           ", "CombAlp     ",
	"Texel0_Alp  ", "Texel1_Alp  ",
	"Prim_Alpha  ", "Shade_Alpha ",
	"Env_Alpha   ", "LOD_Frac    ",
	"PrimLODFrac ", "K5          ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ",	"0           "
};
static const char *sc_colcombtypes16[16] =
{
	"Combined    ", "Texel0      ",
	"Texel1      ", "Primitive   ", 
	"Shade       ", "Env         ",
	"1           ", "CombAlp     ",
	"Texel0_Alp  ", "Texel1_Alp  ",
	"Prim_Alp    ", "Shade_Alpha ",
	"Env_Alpha   ", "LOD_Frac    ",
	"PrimLOD_Frac", "0           "
};
static const char *sc_colcombtypes8[8] =
{
	"Combined    ", "Texel0      ",
	"Texel1      ", "Primitive   ", 
	"Shade       ", "Env         ",
	"1           ", "0           ",
};



void PrintMux( FILE * fh, u64 mux )
{
	u32 mux0 = (u32)(mux>>32);
	u32 mux1 = (u32)(mux);
	
	u32 aRGB0  = (mux0>>20)&0x0F;	// c1 c1		// a0
	u32 bRGB0  = (mux1>>28)&0x0F;	// c1 c2		// b0
	u32 cRGB0  = (mux0>>15)&0x1F;	// c1 c3		// c0
	u32 dRGB0  = (mux1>>15)&0x07;	// c1 c4		// d0

	u32 aA0    = (mux0>>12)&0x07;	// c1 a1		// Aa0
	u32 bA0    = (mux1>>12)&0x07;	// c1 a2		// Ab0
	u32 cA0    = (mux0>>9 )&0x07;	// c1 a3		// Ac0
	u32 dA0    = (mux1>>9 )&0x07;	// c1 a4		// Ad0

	u32 aRGB1  = (mux0>>5 )&0x0F;	// c2 c1		// a1
	u32 bRGB1  = (mux1>>24)&0x0F;	// c2 c2		// b1
	u32 cRGB1  = (mux0    )&0x1F;	// c2 c3		// c1
	u32 dRGB1  = (mux1>>6 )&0x07;	// c2 c4		// d1
	
	u32 aA1    = (mux1>>21)&0x07;	// c2 a1		// Aa1
	u32 bA1    = (mux1>>3 )&0x07;	// c2 a2		// Ab1
	u32 cA1    = (mux1>>18)&0x07;	// c2 a3		// Ac1
	u32 dA1    = (mux1    )&0x07;	// c2 a4		// Ad1

	fprintf(fh, "\n\t\tcase 0x%08x%08xLL:\n", mux0, mux1);
	fprintf(fh, "\t\t//aRGB0: (%s - %s) * %s + %s\n", sc_colcombtypes16[aRGB0], sc_colcombtypes16[bRGB0], sc_colcombtypes32[cRGB0], sc_colcombtypes8[dRGB0]);		
	fprintf(fh, "\t\t//aA0  : (%s - %s) * %s + %s\n", sc_colcombtypes8[aA0], sc_colcombtypes8[bA0], sc_colcombtypes8[cA0], sc_colcombtypes8[dA0]);
	fprintf(fh, "\t\t//aRGB1: (%s - %s) * %s + %s\n", sc_colcombtypes16[aRGB1], sc_colcombtypes16[bRGB1], sc_colcombtypes32[cRGB1], sc_colcombtypes8[dRGB1]);		
	fprintf(fh, "\t\t//aA1  : (%s - %s) * %s + %s\n", sc_colcombtypes8[aA1],  sc_colcombtypes8[bA1], sc_colcombtypes8[cA1],  sc_colcombtypes8[dA1]);

}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::BodgeBlendMode( u32 num_cycles )
{
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	mRecordedCombinerStates.insert( gRDPMux._u64 );

	if(IsCombinerStateDisabled( gRDPMux._u64 ))
	{
		// Use the nasty placeholder texture
		sceGuEnable(GU_TEXTURE_2D);
		sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gSelectedTexture);
		sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
		sceGuTexMode(GU_PSM_8888,0,0,GL_TRUE);		// maxmips/a2/swizzle = 0
		return;
	}
#endif

/*
  * Available apply-modes are: (TFX)
  *   - GU_TFX_MODULATE -			Cv=Ct*Cf					TCC_RGB: Av=Af	TCC_RGBA: Av=At*Af
  *   - GU_TFX_DECAL -	TCC_RGB:	Cv=Ct,Av=Af									TCC_RGBA: Cv=Cf*(1-At)+Ct*At Av=Af
  *   - GU_TFX_BLEND -				Cv=(Cf*(1-Ct))+(Cc*Ct)		TCC_RGB: Av=Af	TCC_RGBA: Av=At*Af
  *   - GU_TFX_REPLACE -			Cv=Ct						TCC_RGB: Av=Af	TCC_RGBA: Av=At
  *   - GU_TFX_ADD -				Cv=Cf+Ct					TCC_RGB: Av=Af	TCC_RGBA: Av=At*Af
  *

  *   - GU_TFX_BLEND -
					// Blend from Cf -> Cc using texture as the blending factor
					Cv = Cf + Ct(Cc - Cf) 
					Av = Af						Av = At*Af
*/
	bool	found( true );
	bool	install_texture( false );

	gSetDiffuseMask = 0;
	gSetDiffuseColour = 0xffffffff;
	gModulateDiffuseMask = 0;
	gModulateDiffuseColour = 0xffffffff;

	// Local vars for now
	u32	gSetTextureMask = 0;
	u32 gSetTextureColour = 0xffffffff;

/*
		// BK
		// XXXX Review
		case 0x0062fe043f15f9ffLL:
            //aRGB0: (1            - Primitive   ) * Env          + Primitive   
            //aA0  : (0            - 0           ) * 0            + Shade       
            //aRGB1: (Combined     - 0           ) * Shade        + 0           
            //aA1  : (Combined     - 0           ) * Env          + 0          
			break;
*/

	switch(gRDPMux._u64)
	{

		case 0x00119623ff2fffffLL:
		//aRGB0: (Texel0       - 0           ) * Primitive    + 0           
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (Texel0       - 0           ) * Primitive    + 0           
		//aA1  : (Texel0       - 0           ) * Primitive    + 0  
			// Modulate the texture*primitive for RGBA
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;

			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// RgRts
		case 0x0011b223ff67ffffLL:
		//aRGB0: (Texel0       - 0           ) * Primitive    + 0           
		//aA0  : (Primitive    - 0           ) * Texel0       + 0           
		//aRGB1: (Texel0       - 0           ) * Primitive    + 0           
		//aA1  : (Primitive    - 0           ) * Texel0       + 0           
			// Modulate the texture*primitive for RGBA
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;

			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// AL
		case 0x0011fffffffff238LL:
		//aRGB0: (Texel0       - 0           ) * Primitive    + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// Modulate the texture*primitive for RGB, T0 for A
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_A;		// Set alpha to 1.0
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		case 0x001197fffffffe38LL:
		//aRGB0: (Texel0       - 0           ) * Primitive    + 0           
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// WR
		case 0x0011c223ff87ffffLL:
		//aRGB0: (Texel0       - 0           ) * Primitive    + 0           
		//aA0  : (Shade        - 0           ) * Texel0       + 0           
		//aRGB1: (Texel0       - 0           ) * Primitive    + 0           
		//aA1  : (Shade        - 0           ) * Texel0       + 0           
			// Modulate the texture*prim for RGB, text*shade for A
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGB;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;			

		case 0x0011fe23fffff3f9LL:
		//aRGB0: (Texel0       - 0           ) * Primitive    + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Texel0       - 0           ) * Primitive    + 0           
		//aA1  : (0            - 0           ) * 0            + Texel0  
			// Modulate the texture*primitive for RGB, texture for alpha
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = (m_dwPrimitiveColor | COL32_MASK_A);		// 1.0 for Alpha
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		case 0x0011fe23fffff7fbLL:
		//aRGB0: (Texel0       - 0           ) * Primitive    + 0           
		//aA0  : (0            - 0           ) * 0            + Primitive   
		//aRGB1: (Texel0       - 0           ) * Primitive    + 0           
		//aA1  : (0            - 0           ) * 0            + Primitive   
			// Modulate the texture*primitive for RGB, primitive for alpha
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
			break;

		// OoT
		case 0x0011fe23fffffdf8LL:
		//aRGB0: (Texel0       - 0           ) * Primitive    + 0           
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (Texel0       - 0           ) * Primitive    + 0           
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_A;		// 1.0 for alpha
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
			break;

		// OoT
		case 0x0011fffffffff638LL:
		//aRGB0: (Texel0       - 0           ) * Primitive    + 0           
		//aA0  : (0            - 0           ) * 0            + Primitive   
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined  
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);	// _RGB - select fragment alpha
			break;

		case 0x0011fffffffffc38LL:
		//aRGB0: (Texel0       - 0           ) * Primitive    + 0           
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_A;		// Set alpha to 1.0
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);	// _RGB - select fragment alpha
			break;

		// XWP
		case 0x00121624ff2fffffLL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (Texel0       - 0           ) * Shade        + 0           
		//aA1  : (Texel0       - 0           ) * Primitive    + 0           
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// SFR2049
		case 0x001217fffffffe38LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// OoT
		case 0x00121803ff0fffffLL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (Texel0       - 0           ) * Shade        + 0           
		//aRGB1: (Combined     - 0           ) * Primitive    + 0           
		//aA1  : (Combined     - 0           ) * Primitive    + 0   
			install_texture = true;
			if( num_cycles == 1 )
			{
				// Nothing to do
			}
			else
			{
				gModulateDiffuseMask = COL32_MASK_RGBA;
				gModulateDiffuseColour = m_dwPrimitiveColor;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// PM
		case 0x00121803ffffff78LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0
		//aA0  : (Texel0       - 0           ) * Shade        + 0
		//aRGB1: (Combined     - 0           ) * Primitive    + Env
		//aA1  : (0            - 0           ) * 0            + Combined
			install_texture = true;
			if( num_cycles == 1 )
			{
				// Ok as is
			}
			else
			{
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;

				// XXXX need to add Env
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// PM
		case 0x00121803fffffff8LL:
        //aRGB0: (Texel0       - 0           ) * Shade        + 0
        //aA0  : (Texel0       - 0           ) * Shade        + 0
        //aRGB1: (Combined     - 0           ) * Primitive    + 0
        //aA1  : (0            - 0           ) * 0            + Combined
			install_texture = true;
			if( num_cycles == 1 )
			{
				// Ok as is
			}
			else
			{
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		case 0x00121824ff33ffffLL:
		//RGB0: (Texel0       - 0           ) * Shade        + 0           
		//A0  : (Texel0       - 0           ) * Shade        + 0           
		//RGB1: (Texel0       - 0           ) * Shade        + 0           
		//A1  : (Texel0       - 0           ) * Shade        + 0 
			// Modulate the texture*shade for RGBA
			install_texture = true;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// AFA
		case 0x001219fffffffe38LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (Texel0       - 0           ) * Shade        + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// Modulate the texture*shade for RGBA
			install_texture = true;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		case 0x00121a24ff37ffffLL:
		//RGB0: (Texel0       - 0           ) * Shade        + 0           
		//A0  : (Texel0       - 0           ) * Env          + 0           
		//RGB1: (Texel0       - 0           ) * Shade        + 0           
		//A1  : (Texel0       - 0           ) * Env          + 0 
			// XXXX think this is ok now
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwEnvColor;

			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// THPS
		case 0x00121a44fffffe38LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (Texel0       - 0           ) * Env          + 0           
		//aRGB1: (Texel1       - 0           ) * Shade        + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if( num_cycles == 1 )
			{
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwEnvColor;

				sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			}
			else
			{
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwEnvColor;

				sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);

				// XXXX Need to add T1*Shade
			}
			break;

		// THPS
		case 0x00127244ffffec38LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 1           ) * Texel0       + 1           
		//aRGB1: (Texel1       - 0           ) * Shade        + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			// XXXX incorrect - needs 1 - T0 for alpha, +T1*Shade for 2nd cycle
			if( num_cycles == 1 )
			{
				// RGB = T0 * Shade
				// A   = 1-T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			else
			{
				// RGB = T0 * Shade + T1 * Shade
				// A   = 1-T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);			// RGBA - modulate fragment alpha
			break;
	

		// OoT drawbridge chains
		case 0x00127e03ff0ff3ffLL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Combined     - 0           ) * Primitive    + 0           
		//aA1  : (Combined     - 0           ) * Primitive    + 0           
			install_texture = true;
			if( num_cycles == 1 )
			{
				// RGB = T0 * Shade
				// A   = T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			else
			{
				// RGB = T0 * Shade * Primitive
				// A   = T0 * Primitive
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwPrimitiveColor;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);			// RGBA - modulate fragment alpha
			break;

		// Qst64
		case 0x00127e03fffff378LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Combined     - 0           ) * Primitive    + Env         
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if( num_cycles == 1 )
			{
				// RGB = T0 * Shade
				// A   = T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			else
			{
				// RGB = T0 * Shade * Primitive + Env
				// A   = T0
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			// XXXX Need to add in env
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);			// RGBA - modulate fragment alpha
			break;

		// OoT background - mountains?
		case 0x00127e03fffff3f8LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Combined     - 0           ) * Primitive    + 0           
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if( num_cycles == 1 )
			{
				// RGB = T0 * Shade
				// A   = T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			else
			{
				// RGB = T0 * Shade * Primitive
				// A   = T0
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);			// RGBA - use texture alpha (diffuse is set to 1)
			break;

		// OoT drawbridge
		case 0x00127e03fffff9fbLL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Shade       
		//aRGB1: (Combined     - 0           ) * Primitive    + 0           
		//aA1  : (0            - 0           ) * 0            + Primitive   
			install_texture = true;
			if( num_cycles == 1 )
			{
				// RGB = T0 * Shade
				// A   = Shade
			}
			else
			{
				// RGB = T0 * Shade * Primitive
				// A   = Primitive
				gModulateDiffuseMask = COL32_MASK_RGBA;
				gModulateDiffuseColour = m_dwPrimitiveColor;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);			// RGB - use fragment alpha
			break;

		// OoT characters
		case 0x00127e03fffffbf8LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (Combined     - 0           ) * Primitive    + 0           
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if( num_cycles == 1 )
			{
				// RGB = T0 * Shade
				// A   = Env
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwEnvColor;
			}
			else
			{
				// RGB = T0 * Shade * Primitive
				// A   = Env
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwEnvColor;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);			// RGB - use fragment alpha of 1.0
			break;

		// OoT horse
		case 0x00127e03fffffdf8LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (Combined     - 0           ) * Primitive    + 0           
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if( num_cycles == 1 )
			{
				// RGB = T0 * Shade
				// A   = 1
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;
			}
			else
			{
				// RGB = T0 * Shade * Primitive
				// A   = 1
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);			// RGB - use fragment alpha of 1.0
			break;

		// OoT
		case 0x00127e05fffffbf8LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (Combined     - 0           ) * Env          + 0           
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if(num_cycles == 1)
			{
				// RGB = T0 * Shade
				// A   = Env
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwEnvColor;
			}
			else
			{
				// RGB = T0 * Shade * Env
				// A   = Env
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwEnvColor;

				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwEnvColor;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
			break;

		// OoT
		case 0x00127e05fffffdf8LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (Combined     - 0           ) * Env          + 0           
		//aA1  : (0            - 0           ) * 0            + Combined  
			install_texture = true;
			if(num_cycles == 1)
			{
				// RGB = T0 * Shade
				// A   = 1
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;
			}
			else
			{
				// RGB = T0 * Shade * Env
				// A   = 1
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwEnvColor;

				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
			break;

		// OoT
		case 0x00127e60f5fffd78LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (Primitive    - Env         ) * Combined     + Env         
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;

			// XXXXXXXXXXXXXXXXXXXXXXXXXX would be good to find out what this is!

			if(num_cycles == 1)
			{
				// RGB = T0 * Shade
				// A   = 1
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;
				sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
			}
			else
			{
				// RGB = Blend(Env, Prim, T0 * Shade)
				// A   = 1
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwEnvColor;

				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;

				// XXXX blend is blending texture, ignoring shade

				sceGuTexEnvColor( m_dwPrimitiveColor );
				sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGB);
			}
			break;

		case 0x00127e24fffff3f9LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Texel0       - 0           ) * Shade        + 0           
		//aA1  : (0            - 0           ) * 0            + Texel0      
			// Modulate the texture*shade for RGB, select shade for alpha
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = MAKE_COLOUR_U32( 0,0,0, 255 );		// Override the diffuse alpha with 1.0

			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// XXXX need to come up with something more cunning for this :(
		case 0x00127e2433fdf8fcLL:
		//aRGB0: (Texel0       - Primitive   ) * Shade        + Primitive   
		//aA0  : (0            - 0           ) * 0            + Shade       
		//aRGB1: (Texel0       - Primitive   ) * Shade        + Primitive   
		//aA1  : (0            - 0           ) * 0            + Shade       
			// Just select texture RGB and shade A for now
			install_texture = true;
			sceGuTexEnvColor( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);
			break;

		// AM:SH
		case 0x00127e24fffff7fbLL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Primitive   
		//aRGB1: (Texel0       - 0           ) * Shade        + 0           
		//aA1  : (0            - 0           ) * 0            + Primitive   
			// Modulate the texture*shade for RGB, select prim for alpha
			install_texture = true;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);

			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwPrimitiveColor;
			break;

		case 0x00127e24fffff9fcLL:
		//RGB0: (Texel0       - 0           ) * Shade        + 0           
		//A0  : (0            - 0           ) * 0            + Shade       
		//RGB1: (Texel0       - 0           ) * Shade        + 0           
		//A1  : (0            - 0           ) * 0            + Shade 
			// Modulate the texture*shade for RGB, select shade for alpha
			install_texture = true;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
			break;

		case 0x00127e24fffffbfdLL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (Texel0       - 0           ) * Shade        + 0           
		//aA1  : (0            - 0           ) * 0            + Env         
			// Modulate the texture*shade for RGB, select env for alpha
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwEnvColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
			break;

		// SpM
		case 0x00127e44fffff238LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Texel1       - 0           ) * Shade        + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if(num_cycles == 1)
			{
				// RGB = T0 * Shade
				// A = T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			else
			{
				// RGB = T0 * Shade + T1 * Shade
				// A = T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0

				// XXXX Needs T1
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// OoT - characters
		case 0x00127e60ff17f3ffLL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Primitive    - 0           ) * Combined     + 0           
		//aA1  : (Combined     - 0           ) * Env          + 0  
			install_texture = true;
			if(num_cycles == 1)
			{
				// RGB = T0 * Shade
				// A = T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			else
			{
				// RGB = T0 * Shade * Primitive
				// A = T0 * Env
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwEnvColor;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// OoT?
		case 0x00127e60fffff3f8LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Primitive    - 0           ) * Combined     + 0           
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if(num_cycles == 1)
			{
				// RGB = T0 * Shade
				// A   = T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			else
			{
				// RGB = T0 * Shade * Primitive
				// A   = T0
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;

				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// OoT
		case 0x00127ea0ff17f3ffLL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Env          - 0           ) * Combined     + 0           
		//aA1  : (Combined     - 0           ) * Env          + 0        
			install_texture = true;
			if(num_cycles == 1)
			{
				// RGB = T0 * Shade
				// A   = T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			else
			{
				// RGB = T0 * Shade * Env
				// A   = T0 * Env
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwEnvColor;

				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwEnvColor;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// OoT?
		case 0x00127ea0fffff3f8LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Env          - 0           ) * Combined     + 0           
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if(num_cycles == 1)
			{
				// RGB = T0 * Shade
				// A   = T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			else
			{
				// RGB = T0 * Shade * Env
				// A   = T0
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwEnvColor;

				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;		// Set alpha to 1.0
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// Qst64
		case 0x00127ec5fffff238LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (1            - 0           ) * Env          + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined
			install_texture = true;
			if( num_cycles == 1 )
			{
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;
				sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			}
			else
			{
				// RGB	= T0 * Shade + Env
				// A	= T0
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;
				sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);

				// XXXX needs to add in Env to result
			}
			break;

		// AG
		case 0x00127ffffffff238LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// Modulate the texture*shade for RGB, select texture alpha
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = COL32_MASK_A;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// M64
		case 0x00127ffffffff838LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Shade       
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// Modulate the texture*shade for RGB, select shade for alpha
			install_texture = true;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
			break;

		// THPS
		case 0x00127ffffffffc38LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined   
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = COL32_MASK_A;					// Alpha of 1.0
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
			break;

		// BK
		case 0x001298043f15ffffLL:
		//aRGB0: (Texel0       - Primitive   ) * Env          + Primitive   
		//aA0  : (Texel0       - 0           ) * Shade        + 0           
		//aRGB1: (Combined     - 0           ) * Shade        + 0           
		//aA1  : (Combined     - 0           ) * Env          + 0      
			install_texture = true;
			// RGB = Blend( Prim, T0, Env ) * Shade
			// A   = T0 * Shade * Env

			// Assume this is only ever used in 2 cycle
			
			// Hoplessly wrong - do Prim * Shade * T0, Env * Shade * T0
			install_texture = true;
			//gModulateDiffuseMask = COL32_MASK_RGBA;
			//gModulateDiffuseColour = (m_dwPrimitiveColor & COL32_MASK_RGB) | (m_dwEnvColor & COL32_MASK_A);

			install_texture = true;
			if( num_cycles == 1 )
			{
				gSetDiffuseMask = COL32_MASK_RGB;
				gSetDiffuseColour = COL32_MASK_RGB;		// Set RGB to 1.0, i.e. select Texture
			}
			else
			{
				// Leave RGB shade untouched

				gModulateDiffuseMask = COL32_MASK_A;
				gModulateDiffuseColour = m_dwEnvColor;
			}

			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);

			break;

		// WR
		case 0x00129825ff33ffffLL:
		//aRGB0: (Texel0       - 0           ) * Env          + 0           
		//aA0  : (Texel0       - 0           ) * Shade        + 0           
		//aRGB1: (Texel0       - 0           ) * Env          + 0           
		//aA1  : (Texel0       - 0           ) * Shade        + 0           
			// Modulate the texture*env for RGB, text*shade for A
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGB;
			gSetDiffuseColour = m_dwEnvColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		case 0x00129a25ff37ffffLL:
		//aRGB0: (Texel0       - 0           ) * Env          + 0           
		//aA0  : (Texel0       - 0           ) * Env          + 0           
		//aRGB1: (Texel0       - 0           ) * Env          + 0           
		//aA1  : (Texel0       - 0           ) * Env          + 0           
			// Modulate the texture*env for RGBA
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwEnvColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// MKRT
		case 0x0012fe25fffff3f9LL:
		//aRGB0: (Texel0       - 0           ) * Env          + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Texel0       - 0           ) * Env          + 0           
		//aA1  : (0            - 0           ) * 0            + Texel0      
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwEnvColor | COL32_MASK_A;		// Need to set alpha to 1.0 as it's modulated with t0
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// OoT
		case 0x0012fe25fffffdfeLL:
		//aRGB0: (Texel0       - 0           ) * Env          + 0           
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (Texel0       - 0           ) * Env          + 0           
		//aA1  : (0            - 0           ) * 0            + 1           
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwEnvColor | COL32_MASK_A;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);				// _RGB - use 1.0 for alpha
			break;

		// THPS
		case 0x0012fffffffffa38LL:
		//aRGB0: (Texel0       - 0           ) * Env          + 0           
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwEnvColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);				// _RGB - use fragment alpha
			break;

		// M64 eyes decal.
		case 0x00147e2844fe7b3dLL:
		//aRGB0: (Texel0       - Shade       ) * Texel0_Alp   + Shade       
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (Texel0       - Shade       ) * Texel0_Alp   + Shade       
		//aA1  : (0            - 0           ) * 0            + Env         
			// RGB = Blend( Shade, T0, T0Alpha )  -- essentially DECAL mode on PSP
			// A   = Env
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwEnvColor;
			sceGuTexEnvColor( m_dwEnvColor );
			sceGuTexFunc(GU_TFX_DECAL,GU_TCC_RGBA);		
			break;

		// F0
		case 0x00147e045ffefbf8LL:
		//aRGB0: (Texel0       - Env         ) * Texel0_Alp   + Env         
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (Combined     - 0           ) * Shade        + 0           
		//aA1  : (0            - 0           ) * 0            + Combined 
			// RGB = Blend(E,T0,T0a)*Shade	~= Blend(SE, ST0, T0a)
			// A   = Env

			// XXXX needs to modulate again by shade
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwEnvColor;
			sceGuTexEnvColor( m_dwEnvColor );
			sceGuTexFunc(GU_TFX_DECAL,GU_TCC_RGBA);		
			break;

		// OoT intro
		case 0x00167e6035fcff7eLL:
        //aRGB0: (Texel0       - Primitive   ) * Env_Alpha    + Texel0      
        //aA0  : (0            - 0           ) * 0            + 0           
        //aRGB1: (Primitive    - Env         ) * Combined     + Env         
        //aA1  : (0            - 0           ) * 0            + 1           
			// XXXX incorrect, need env and prim, blend has wrong source
			install_texture = true;

			// Use the primitive for the r,g,b, override the alpha with 1.0
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = (MAKE_COLOUR_U32( 0,0,0, 255) & COL32_MASK_A) | (m_dwPrimitiveColor & ~COL32_MASK_A);

			sceGuTexEnvColor( m_dwEnvColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGB);
			break;

		// OoT intro
		case 0x00267e60350cf37fLL:
        //aRGB0: (Texel1       - Primitive   ) * Env_Alpha    + Texel0      
        //aA0  : (0            - 0           ) * 0            + Texel0      
        //aRGB1: (Primitive    - Env         ) * Combined     + Env         
        //aA1  : (Combined     - 0           ) * Primitive    + 0           
			// XXXX need 2nd texture
			install_texture = true;

			// XXXX need to blend using env

			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// AFA
		case 0x0026a0041f1093ffLL:
		//aRGB0: (Texel1       - Texel0      ) * LOD_Frac     + Texel0      
		//aA0  : (Texel1       - Texel0      ) * Combined     + Texel0      
		//aRGB1: (Combined     - 0           ) * Shade        + 0           
		//aA1  : (Combined     - 0           ) * Shade        + 0           
			// XXXX need 2nd texture
			install_texture = true;
			if( num_cycles == 1 )
				sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
			else
				sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);

			break;

		// THPS
		case 0x0026a0041ffc93f8LL:
		//aRGB0: (Texel1       - Texel0      ) * LOD_Frac     + Texel0      
		//aA0  : (Texel1       - Texel0      ) * Combined     + Texel0      
		//aRGB1: (Combined     - 0           ) * Shade        + 0           
		//aA1  : (0            - 0           ) * 0            + Combined    
			// XXXX need 2nd texture
			install_texture = true;
			if( num_cycles == 1 )
				sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
			else
			{
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;			// Set alpha to 1.0
				sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			}
			break;

		// AM:SH
		case 0x00309661442e7f3fLL:
		//aRGB0: (Primitive    - Shade       ) * Texel0       + Shade       
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (Primitive    - Shade       ) * Texel0       + Shade       
		//aA1  : (Texel0       - 0           ) * Primitive    + 0     
			install_texture = true;
			// RGB = Blend( Env, Primitive, T0 )
			// A   = T0 * Primitive
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexEnvColor( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(s,p,t) for RGB, Af*At for alpha
			break;

		// XWP
		case 0x00309661552efb7dLL:
		//aRGB0: (Primitive    - Env         ) * Texel0       + Env         
		//aA0  : (Texel0       - 0           ) * Primitive    + Env         
		//aRGB1: (Primitive    - Env         ) * Texel0       + Env         
		//aA1  : (Texel0       - 0           ) * Primitive    + Env         
			install_texture = true;
			// RGB = Blend( Env, Primitive, T0 )
			// A   = T0 * Primitive + Env
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = (m_dwEnvColor & COL32_MASK_RGB) | (m_dwPrimitiveColor & COL32_MASK_A);
			sceGuTexEnvColor( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(e,p,t) for RGB, Modulate(t,p) for alpha

			// XXXX need to add in Env for alpha
 			break;


		// MKRT ingame - smoke puffs, OoT logo text
		case 0x00309661552eff7fLL:
		//aRGB0: (Primitive    - Env         ) * Texel0       + Env         
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (Primitive    - Env         ) * Texel0       + Env         
		//aA1  : (Texel0       - 0           ) * Primitive    + 0           
			install_texture = true;
			// RGB = Blend( Env, Primitive, T0 )
			// A   = T0 * Primitive
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = (m_dwEnvColor & COL32_MASK_RGB) | (m_dwPrimitiveColor & COL32_MASK_A);
			sceGuTexEnvColor( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(e,p,t) for RGB, Modulate(t,p) for alpha
 			break;

		// OoT - particles
		case 0x003097ff5ffefe38LL:
		//aRGB0: (Primitive    - Env         ) * Texel0       + Env         
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			// RGB = Blend( Env, Primitive, T0 )
			// A   = T0 * Primitive
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = (m_dwEnvColor & COL32_MASK_RGB) | (m_dwPrimitiveColor & COL32_MASK_A);
			sceGuTexEnvColor( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(e,p,t) for RGB, Modulate(t,p) for alpha
			break;

		// OoT
		case 0x003097fffffffe38LL:
		//aRGB0: (Primitive    - 0           ) * Texel0       + 0           
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// SpM
		case 0x003099ff5ffefe38LL:
		//aRGB0: (Primitive    - Env         ) * Texel0       + Env         
		//aA0  : (Texel0       - 0           ) * Shade        + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined   
			// Blend the texture*shade for RGB, 
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGB;
			gSetDiffuseColour = m_dwEnvColor;
			sceGuTexEnvColor( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);
			break;

		case 0x0030b26144664924LL:
		//aRGB0: (Primitive    - Shade       ) * Texel0       + Shade       
		//aA0  : (Primitive    - Shade       ) * Texel0       + Shade       
		//aRGB1: (Primitive    - Shade       ) * Texel0       + Shade       
		//aA1  : (Primitive    - Shade       ) * Texel0       + Shade       
			// Modulate the texture*shade for RGBA
			install_texture = true;
			sceGuTexEnvColor( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGB);			// XXXX Argh - need to interpolate alpha too!? We're just selecting Shade for now
			break;

		// OoT intro - moon
		case 0x0030b2615566ff7fLL:
        //aRGB0: (Primitive    - Env         ) * Texel0       + Env         
        //aA0  : (Primitive    - 0           ) * Texel0       + 0           
        //aRGB1: (Primitive    - Env         ) * Texel0       + Env         
        //aA1  : (Primitive    - 0           ) * Texel0       + 0           
			// Colour is similar to the above, but we force diffuse to be env colour
			// RGB = Blend( Env, Prim, T0 )
			// A   = Prim * To
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = (m_dwEnvColor & COL32_MASK_RGB) | (m_dwPrimitiveColor & COL32_MASK_A);

			sceGuTexEnvColor( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend RGB, Modulate for alpha
			break;

		// PM
		case 0x0030b261ff2fffffLL:
        //aRGB0: (Primitive    - 0           ) * Texel0       + 0
        //aA0  : (Primitive    - 0           ) * Texel0       + 0
        //aRGB1: (Primitive    - 0           ) * Texel0       + 0
        //aA1  : (Texel0       - 0           ) * Primitive    + 0
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// SpM
		case 0x0030b3ff5ffeda38LL:
		//aRGB0: (Primitive    - Env         ) * Texel0       + Env         
		//aA0  : (Primitive    - Env         ) * Texel0       + Env         
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			// RGB = Blend( Env, Primitive, T0 )
			// A   = T0 * Primitive
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = (m_dwEnvColor & COL32_MASK_RGB) | (m_dwPrimitiveColor & COL32_MASK_A);
			sceGuTexEnvColor( m_dwPrimitiveColor );
			// XXXX Needs to BLEND alpha values!
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(e,p,t) for RGB, Modulate(t,p) for alpha


		// OoT rain splashes
		case 0x0030b3ff5ffefe38LL:
		//aRGB0: (Primitive    - Env         ) * Texel0       + Env         
		//aA0  : (Primitive    - 0           ) * Texel0       + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			// RGB = Blend( Env, Primitive, T0 )
			// A   = T0 * Primitive
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = (m_dwEnvColor & COL32_MASK_RGB) | (m_dwPrimitiveColor & COL32_MASK_A);
			sceGuTexEnvColor( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(e,p,t) for RGB, Modulate(t,p) for alpha
			break;

		// SpM
		case 0x0030b3fffffffe38LL:
		//aRGB0: (Primitive    - 0           ) * Texel0       + 0           
		//aA0  : (Primitive    - 0           ) * Texel0       + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined   
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// AM:SH
		case 0x0030fe6144fe7339LL:
		//aRGB0: (Primitive    - Shade       ) * Texel0       + Shade       
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Primitive    - Shade       ) * Texel0       + Shade       
		//aA1  : (0            - 0           ) * 0            + Texel0  
			install_texture = true;
			// RGB = Blend( Env, Primitive, T0 )
			// A   = T0 * Primitive
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = COL32_MASK_A;				// Set alpha to 1.0
			sceGuTexEnvColor( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(s,p,t) for RGB, At*1 for alpha
			break;

		// AM:SH
		case 0x0030fe6144fe773bLL:
		//aRGB0: (Primitive    - Shade       ) * Texel0       + Shade       
		//aA0  : (0            - 0           ) * 0            + Primitive   
		//aRGB1: (Primitive    - Shade       ) * Texel0       + Shade       
		//aA1  : (0            - 0           ) * 0            + Primitive   
			install_texture = true;
			// RGB = Blend( Env, Primitive, T0 )
			// A   = T0 * Primitive
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexEnvColor( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGB);			// _RGBA = blend(s,p,t) for RGB, Af for alpha
			break;

		// MKRT opening (flag)
		case 0x00327e64fffff9fcLL:
		//aRGB0: (Primitive    - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Shade       
		//aRGB1: (Primitive    - 0           ) * Shade        + 0           
		//aA1  : (0            - 0           ) * 0            + Shade    
			gModulateDiffuseMask = COL32_MASK_RGB;
			gModulateDiffuseColour = m_dwPrimitiveColor;
			break;

		// AG
		case 0x00327e64fffffbfdLL:
		//aRGB0: (Primitive    - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (Primitive    - 0           ) * Shade        + 0           
		//aA1  : (0            - 0           ) * 0            + Env         
			gModulateDiffuseMask = COL32_MASK_RGB;
			gModulateDiffuseColour = m_dwPrimitiveColor;

			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwEnvColor;
			break;

		// Qst64
		case 0x00327ec5fffffa38LL:
		//aRGB0: (Primitive    - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (1            - 0           ) * Env          + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			if( num_cycles == 1 )
			{
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;

				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwEnvColor;
			}
			else
			{
				// RGB	= Prim * Shade + Env
				// A	= Env
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;

				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwEnvColor;

				// XXXX needs to add in Env
			}
			break;

		// WR
		case 0x00327ffffffff838LL:
		//aRGB0: (Primitive    - 0           ) * Shade        + 0           
		//aA0  : (0            - 0           ) * 0            + Shade       
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			gModulateDiffuseMask = COL32_MASK_RGB;
			gModulateDiffuseColour = m_dwPrimitiveColor;
			break;

		// TSph
		case 0x00323864ff73ffffLL:
		//aRGB0: (Primitive    - 0           ) * Shade        + 0           
		//aA0  : (Primitive    - 0           ) * Shade        + 0           
		//aRGB1: (Primitive    - 0           ) * Shade        + 0           
		//aA1  : (Primitive    - 0           ) * Shade        + 0           
			gModulateDiffuseMask = COL32_MASK_RGBA;
			gModulateDiffuseColour = m_dwPrimitiveColor;
			break;

		// Qst64
		case 0x0032ffff1ffff6fbLL:
        //aRGB0: (Primitive    - Texel0      ) * Env          + 0           
        //aA0  : (0            - 0           ) * 0            + Primitive   
        //aRGB1: (0            - 0           ) * 0            + Primitive   
        //aA1  : (0            - 0           ) * 0            + Primitive   
			found = false;
			break;

		case 0x0041c683ff8fffffLL:
		//aRGB0: (Shade        - 0           ) * Primitive    + 0           
		//aA0  : (Shade        - 0           ) * Primitive    + 0           
		//aRGB1: (Shade        - 0           ) * Primitive    + 0           
		//aA1  : (Shade        - 0           ) * Primitive    + 0           
			gModulateDiffuseMask = COL32_MASK_RGBA;
			gModulateDiffuseColour = m_dwPrimitiveColor;
			break;

		// OoT
		case 0x0041c7fffffffe38LL:
		//aRGB0: (Shade        - 0           ) * Primitive    + 0           
		//aA0  : (Shade        - 0           ) * Primitive    + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			gModulateDiffuseMask = COL32_MASK_RGBA;
			gModulateDiffuseColour = m_dwPrimitiveColor;
			break;

		case 0x0041fffffffffc38LL:
		//aRGB0: (Shade        - 0           ) * Primitive    + 0           
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
            gModulateDiffuseMask = COL32_MASK_RGB;
			gModulateDiffuseColour = m_dwPrimitiveColor;
			
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = COL32_MASK_A; // Alpha is 1.0
			break;

		case 0x0042ca85ff97ffffLL:
		//aRGB0: (Shade        - 0           ) * Env          + 0           
		//aA0  : (Shade        - 0           ) * Env          + 0           
		//aRGB1: (Shade        - 0           ) * Env          + 0           
		//aA1  : (Shade        - 0           ) * Env          + 0           
			gModulateDiffuseMask = COL32_MASK_RGBA;
			gModulateDiffuseColour = m_dwEnvColor;
			break;

		// AM:SH
		case 0x005096a1332dfeffLL:
		//aRGB0: (Env          - Primitive   ) * Texel0       + Primitive   
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (Env          - Primitive   ) * Texel0       + Primitive   
		//aA1  : (Texel0       - 0           ) * Primitive    + 0      
			install_texture = true;
			// RGB = Blend( Primitive, Env, T0 )
			// A   = T0 * Primitive
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexEnvColor( m_dwEnvColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(p,e,t) for RGB, Modulate(t,p) for alpha
			break;

		// THPS
		case 0x005094023f15ffffLL:
		//aRGB0: (Env          - Primitive   ) * Texel0       + Primitive   
		//aA0  : (Texel0       - 0           ) * Texel1       + 0           
		//aRGB1: (Combined     - 0           ) * Texel1       + 0           
		//aA1  : (Combined     - 0           ) * Env          + 0           
			install_texture = true;
			// XXXX Needs T1
			if( num_cycles == 1 )
			{
				// RGB = Blend( Primitive, Env, T0 )
				// A   = T0 * T1
				gSetDiffuseMask = COL32_MASK_RGBA;
				gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_A;
				sceGuTexEnvColor( m_dwEnvColor );
				sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(p,e,t) for RGB, T0*1 for alpha
			}
			else
			{
				// RGB = Blend( Primitive, Env, T0 )
				// A   = T0 * T1 * Env
				gSetDiffuseMask = COL32_MASK_RGBA;
				gSetDiffuseColour = (m_dwPrimitiveColor & COL32_MASK_RGB) | (m_dwEnvColor & COL32_MASK_A);
				sceGuTexEnvColor( m_dwEnvColor );
				sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(p,e,t) for RGB, T0*Env for alpha
			}
			break;

		// THPS
		case 0x0050ffff3ffdf238LL:
		//aRGB0: (Env          - Primitive   ) * Texel0       + Primitive   
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			// RGB = Blend( Primitive, Env, T0 )
			// A   = T0 * 1
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_A;
			sceGuTexEnvColor( m_dwEnvColor );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(p,e,t) for RGB, T0*1 for alpha
			break;

		// SpM
		case 0x0060ffff5ffef238LL:
		//aRGB0: (1            - Env         ) * Texel0       + Env         
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			// RGB = Blend( Env, 1, T0 )
			// A   = T0
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwEnvColor | COL32_MASK_A;		// 1.0 for alpha
			sceGuTexEnvColor( COL32_WHITE );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);

			break;

		// OoT, selected option
		case 0x006196c3ff2fffffLL:
		//aRGB0: (1            - 0           ) * Primitive    + 0           
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (1            - 0           ) * Primitive    + 0           
		//aA1  : (Texel0       - 0           ) * Primitive    + 0           
			// RGB = Prim
			// A   = Prim * T0
			install_texture = true;
			gSetTextureMask = COL32_MASK_RGB;
			gSetTextureColour = COL32_WHITE;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// AFA
		case 0x00627fff1ffcfc38LL:
		//aRGB0: (1            - Texel0      ) * Shade        + Texel0      
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// RGB = Blend( T0, 1, Shade )
			// A   = 1
			install_texture = true;

			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = COL32_MASK_A;

			// XXXX this is just selecting the texture colour, ignoring the blend
			// Could do this by manually shading texture...

			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);
			break;

		// BK
		case 0x0062fe043f15f9ffLL:
		//aRGB0: (1            - Primitive   ) * Env          + Primitive   
		//aA0  : (0            - 0           ) * 0            + Shade       
		//aRGB1: (Combined     - 0           ) * Shade        + 0           
		//aA1  : (Combined     - 0           ) * Env          + 0           
		{
			u32		blend( COL32_BLEND( m_dwPrimitiveColor, MAKE_COLOUR_U32(255,255,255), m_dwEnvColor ) );

			if( num_cycles == 1 )
			{
				gSetDiffuseMask = COL32_MASK_RGB;
				gSetDiffuseColour = blend;
			}
			else
			{
				gModulateDiffuseMask = COL32_MASK_RGBA;
				gModulateDiffuseColour = (blend & COL32_MASK_RGB) | (m_dwEnvColor & COL32_MASK_A);
			}
		}
			break;

		// XWP
		case 0x00f0b3e13365feffLL:
		//aRGB0: (0            - Primitive   ) * Texel0       + Primitive   
		//aA0  : (Primitive    - 0           ) * Texel0       + 0           
		//aRGB1: (0            - Primitive   ) * Texel0       + Primitive   
		//aA1  : (Primitive    - 0           ) * Texel0       + 0      
			install_texture = true;
			// RGB = Blend( Primitive, 0, T0 )
			// A   = T0 * Primitive
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexEnvColor( COL32_BLACK );
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);			// _RGBA = blend(p,e,t) for RGB, T0*Prim for alpha
			break;

		// XWP
		case 0x00ff97ffff2cfa7dLL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (Texel0       - 0           ) * Primitive    + Env         
		//aRGB1: (0            - 0           ) * 0            + Texel0      
		//aA1  : (Texel0       - 0           ) * Primitive    + Env         
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_RGB;	// Set RGB components to white

			// XXXX need to add in Env to alpha
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// PM
		case 0x00ff97ffff2cfe7fLL:
		//aRGB0: (0            - 0           ) * 0            + Texel0
		//aA0  : (Texel0       - 0           ) * Primitive    + 0
		//aRGB1: (0            - 0           ) * 0            + Texel0
		//aA1  : (Texel0       - 0           ) * Primitive    + 0
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_RGB;	// Set RGB components to white
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// OoT press start
		case 0x00ff97ffff2dfeffLL:
		//aRGB0: (0            - 0           ) * 0            + Primitive   
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (0            - 0           ) * 0            + Primitive   
		//aA1  : (Texel0       - 0           ) * Primitive    + 0           
			install_texture = true;
			gSetTextureMask = COL32_MASK_RGB;
			gSetTextureColour = COL32_WHITE;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// AL
		case 0x00ff97fffffcfe38LL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// RGB = T0
			// A   = T0 * Prim
			// XXXX placeholder implementation - think might be ok
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = COL32_MASK_RGB | m_dwPrimitiveColor;		// Set diffuse RGB to white
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// OoT?
		case 0x00ff97fffffdfe38LL:
		//aRGB0: (0            - 0           ) * 0            + Primitive   
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// RGB = Prim
			// A   = T0 * Prim
			// XXXX placeholder implementation - think might be ok
			install_texture = true;
			gSetTextureMask = COL32_MASK_RGB;
			gSetTextureColour = COL32_WHITE;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// BM
		case 0x00ff99ffff30fe7fLL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (Texel0       - 0           ) * Shade        + 0           
		//aRGB1: (0            - 0           ) * 0            + Texel0      
		//aA1  : (Texel0       - 0           ) * Shade        + 0           
			install_texture = true;
			gSetTextureMask = COL32_MASK_RGB;
			gSetTextureColour = COL32_WHITE;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;


		// MKRT selection reticule
		case 0x00ff99ffff327f3fLL:
        //aRGB0: (0            - 0           ) * 0            + Shade       
        //aA0  : (Texel0       - 0           ) * Shade        + 0           
        //aRGB1: (0            - 0           ) * 0            + Shade       
        //aA1  : (Texel0       - 0           ) * Shade        + 0           
			install_texture = true;
			gSetTextureMask = COL32_MASK_RGB;
			gSetTextureColour = COL32_WHITE;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// Qst64
		case 0x00ff99fffffffe38LL:
		//aRGB0: (0            - 0           ) * 0            + 0           
		//aA0  : (Texel0       - 0           ) * Shade        + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGB;
			gSetDiffuseColour = COL32_BLACK;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		case 0x00ff9bffff34fe7fLL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (Texel0       - 0           ) * Env          + 0           
		//aRGB1: (0            - 0           ) * 0            + Texel0      
		//aA1  : (Texel0       - 0           ) * Env          + 0           
			// Modulate the texture*env for A, select texture for RGB
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			// ABGR(env,255,255,255)
			gSetDiffuseColour = (m_dwEnvColor & COL32_MASK_A) | (MAKE_COLOUR_U32(255,255,255) & ~COL32_MASK_A);
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// WR
		case 0x00ff9bffff36ff7fLL:
		//aRGB0: (0            - 0           ) * 0            + Env         
		//aA0  : (Texel0       - 0           ) * Env          + 0           
		//aRGB1: (0            - 0           ) * 0            + Env         
		//aA1  : (Texel0       - 0           ) * Env          + 0           
			// Modulate the texture*env for A, select env for RGB
			install_texture = true;
			gSetTextureMask = COL32_MASK_RGB;
			gSetTextureColour = COL32_WHITE;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwEnvColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// Qst64
		case 0x00ff9bfffffcfe38LL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (Texel0       - 0           ) * Env          + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// RGB:	T0
			// A:	T0 * Env
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwEnvColor | COL32_MASK_RGB;			// Set RGB to white
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// Qst64
		case 0x00ff9bfffffefe38LL:
		//aRGB0: (0            - 0           ) * 0            + Env         
		//aA0  : (Texel0       - 0           ) * Env          + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// RGB:	Env
			// A:	T0 * Env
			install_texture = true;
			gSetTextureMask = COL32_MASK_RGB;
			gSetTextureColour = COL32_WHITE;							// Set texture to white
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwEnvColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// PM
		case 0x00ffb3ffff2fffffLL:
        //aRGB0: (0            - 0           ) * 0            + 0
        //aA0  : (Primitive    - 0           ) * Texel0       + 0
        //aRGB1: (0            - 0           ) * 0            + 0
        //aA1  : (Texel0       - 0           ) * Primitive    + 0
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor & COL32_MASK_A;	// Set RGB components to black
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// PM
		case 0x00ffb3ffff64fe7fLL:
		//aRGB0: (0            - 0           ) * 0            + Texel0
		//aA0  : (Primitive    - 0           ) * Texel0       + 0
		//aRGB1: (0            - 0           ) * 0            + Texel0
		//aA1  : (Primitive    - 0           ) * Texel0       + 0
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_RGB;	// Set RGB components to white
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// F0
		case 0x00ffb3ffff65feffLL:
		//aRGB0: (0            - 0           ) * 0            + Primitive   
		//aA0  : (Primitive    - 0           ) * Texel0       + 0           
		//aRGB1: (0            - 0           ) * 0            + Primitive   
		//aA1  : (Primitive    - 0           ) * Texel0       + 0           
			install_texture = true;
			gSetTextureMask = COL32_MASK_RGB;
			gSetTextureColour = COL32_WHITE;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// AM:SH
		case 0x00ffc7ffff8e7f3fLL:
		//aRGB0: (0            - 0           ) * 0            + Shade       
		//aA0  : (Shade        - 0           ) * Primitive    + 0           
		//aRGB1: (0            - 0           ) * 0            + Shade       
		//aA1  : (Shade        - 0           ) * Primitive    + 0           
			gModulateDiffuseMask = COL32_MASK_A;
			gModulateDiffuseColour = m_dwPrimitiveColor;
			break;

		// RgRts
		case 0x00fffe03ff0e79ffLL:
		//aRGB0: (0            - 0           ) * 0            + Shade       
		//aA0  : (0            - 0           ) * 0            + Shade       
		//aRGB1: (Combined     - 0           ) * Primitive    + 0           
		//aA1  : (Combined     - 0           ) * Primitive    + 0   
			break;
			if( num_cycles > 1 )
			{
				gModulateDiffuseMask = COL32_MASK_RGBA;
				gModulateDiffuseColour = m_dwPrimitiveColor;
			}
			break;

		// PM
		case 0x00fffe03fffe79f8LL:
        //aRGB0: (0            - 0           ) * 0            + Shade
        //aA0  : (0            - 0           ) * 0            + Shade
        //aRGB1: (Combined     - 0           ) * Primitive    + 0
        //aA1  : (0            - 0           ) * 0            + Combined
			if( num_cycles > 1 )
			{
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;
			}
			break;

		// F0
		case 0x00fffe04fffefbf8LL:
		//aRGB0: (0            - 0           ) * 0            + Env         
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (Combined     - 0           ) * Shade        + 0           
		//aA1  : (0            - 0           ) * 0            + Combined    
			gModulateDiffuseMask = COL32_MASK_RGB;
			gModulateDiffuseColour = m_dwEnvColor;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwEnvColor;
			break;

		// SpM
		case 0x00fffe44fffcf238LL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Texel1       - 0           ) * Shade        + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if( num_cycles == 1 )
			{
				sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
			}
			else
			{
				// RGB	= T0 + T1*Shade
				// A	= T0

				// XXXX Need to add in T1*Shade
				sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
			}
			break;

		// SpM
		case 0x00fffe44fffcfc38LL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (Texel1       - 0           ) * Shade        + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if( num_cycles == 1 )
			{
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;

				sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);			// RGB - Select diffuse alpha
			}
			else
			{
				// RGB	= T0 + T1*Shade
				// A	= T0

				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = COL32_MASK_A;

				// XXXX Need to add in T1*Shade
				sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);			// RGB - Select diffuse alpha
			}
			break;

		// SpM
		case 0x00fffe44fffdf238LL:
		//aRGB0: (0            - 0           ) * 0            + Primitive   
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (Texel1       - 0           ) * Shade        + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if( num_cycles == 1 )
			{
				gSetDiffuseMask = COL32_MASK_RGBA;
				gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_A;

				sceGuTexFunc(GU_TFX_MODULATE, GU_TCC_RGBA);
			}
			else
			{
				// RGB	= T0 + T1*Shade
				// A	= T0

				gSetDiffuseMask = COL32_MASK_RGBA;
				gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_A;

				// XXXX Need to add in T1*Shade
				sceGuTexFunc(GU_TFX_MODULATE, GU_TCC_RGBA);
			}
			break;

		// Qst64
		case 0x00fffec5fffcf238LL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (1            - 0           ) * Env          + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			if( num_cycles == 1 )
			{
				sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
			}
			else
			{
				// RGB	= T0 + Env
				// A	= T0
				gSetDiffuseMask = COL32_MASK_RGBA;
				gSetDiffuseColour = m_dwEnvColor | COL32_MASK_A;
				sceGuTexFunc(GU_TFX_ADD,GU_TCC_RGBA);			// ADD/RGBA modulates alphas
			}
			break;

		// Qst64
		case 0x00fffec5fffdfa38LL:
		//aRGB0: (0            - 0           ) * 0            + Primitive   
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (1            - 0           ) * Env          + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			if( num_cycles == 1 )
			{
				gSetDiffuseMask = COL32_MASK_RGBA;
				gSetDiffuseColour = (m_dwPrimitiveColor & COL32_MASK_RGB) | (m_dwEnvColor & COL32_MASK_A);
			}
			else
			{
				// RGB	= Prim+Env
				// A	= Env
				gSetDiffuseMask = COL32_MASK_RGBA;
				gSetDiffuseColour = (ADD_COLOUR_RGBA(m_dwPrimitiveColor,m_dwEnvColor) & COL32_MASK_RGB) | (m_dwEnvColor & COL32_MASK_A);
			}
			break;

		// M64
		case 0x00fffffffffcf238LL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// Select the texture
			install_texture = true;
			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
			break;

		case 0x00fffffffffcf279LL:
		//RGB0: (0            - 0           ) * 0            + Texel0      
		//A0  : (0            - 0           ) * 0            + Texel0      
		//RGB1: (0            - 0           ) * 0            + Texel0      
		//A1  : (0            - 0           ) * 0            + Texel0      
			// Select the texture
			install_texture = true;
			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
			break;

		// SpM
		case 0x00fffffffffcf438LL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (0            - 0           ) * 0            + Texel1      
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// Select the texture
			install_texture = true;
			// XXXX Needs T1
			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
			break;

		// PM
		case 0x00fffffffffcf67bLL:
        //aRGB0: (0            - 0           ) * 0            + Texel0
        //aA0  : (0            - 0           ) * 0            + Primitive
        //aRGB1: (0            - 0           ) * 0            + Texel0
        //aA1  : (0            - 0           ) * 0            + Primitive
			// Select the texture
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);	// _RGB - select alpha from fragment
			break;

		// AL
		case 0x00fffffffffcf83cLL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (0            - 0           ) * 0            + Shade       
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Shade       
			// Select the texture
			install_texture = true;
			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);
			break;

		case 0x00fffffffffcf87cLL:
        //aRGB0: (0            - 0           ) * 0            + Texel0      
        //aA0  : (0            - 0           ) * 0            + Shade       
        //aRGB1: (0            - 0           ) * 0            + Texel0      
        //aA1  : (0            - 0           ) * 0            + Shade       
			// Select the texture
			install_texture = true;
			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);
			break;

		case 0x00fffffffffcfa7dLL:
        //aRGB0: (0            - 0           ) * 0            + Texel0      
        //aA0  : (0            - 0           ) * 0            + Env         
        //aRGB1: (0            - 0           ) * 0            + Texel0      
        //aA1  : (0            - 0           ) * 0            + Env  
			install_texture = true;

			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwEnvColor;

			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);		// RGB from Tex, A from Fragment
			break;

		// SpM
		case 0x00fffffffffcfc38LL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = COL32_MASK_A;
			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);		// RGB from Tex, A from Fragment
			break;

		// SFx
		case 0x00fffffffffcfc7eLL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (0            - 0           ) * 0            + Texel0      
		//aA1  : (0            - 0           ) * 0            + 1           
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = COL32_MASK_A;
			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);		// RGB from Tex, A from Fragment
			break;


		// OoT
		case 0x00fffffffffcfe7fLL:
		//aRGB0: (0            - 0           ) * 0            + Texel0      
		//aA0  : (0            - 0           ) * 0            + 0           
		//aRGB1: (0            - 0           ) * 0            + Texel0      
		//aA1  : (0            - 0           ) * 0            + 0           
			install_texture = true;

			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = 0;		// Black, no alpha ?

			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);		// RGB from Tex, A from Fragment
			break;

		// MKRT ingame - banners
		case 0x00fffffffffdf2f9LL:
		//aRGB0: (0            - 0           ) * 0            + Primitive   
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (0            - 0           ) * 0            + Primitive   
		//aA1  : (0            - 0           ) * 0            + Texel0 
			install_texture = true;
			gSetTextureMask = COL32_MASK_RGB;
			gSetTextureColour = COL32_WHITE;
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_A;		// ADD/RGBA modulates alpha at texture fragment
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// AL
		case 0x00fffffffffdf638LL:
		//aRGB0: (0            - 0           ) * 0            + Primitive   
		//aA0  : (0            - 0           ) * 0            + Primitive   
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// Select just the primitive colour
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			break;

		// OoT, fade?
		case 0x00fffffffffdf6fbLL:
		//aRGB0: (0            - 0           ) * 0            + Primitive   
		//aA0  : (0            - 0           ) * 0            + Primitive   
		//aRGB1: (0            - 0           ) * 0            + Primitive   
		//aA1  : (0            - 0           ) * 0            + Primitive   
			// Select just the primitive colour
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;
			break;

		case 0x00fffffffffdfc38LL:
		//aRGB0: (0            - 0           ) * 0            + Primitive   
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_A;		// Alpha is 1.0
			break;

		// XWP
		case 0x00fffffffffe773bLL:
		//aRGB0: (0            - 0           ) * 0            + Shade       
		//aA0  : (0            - 0           ) * 0            + Primitive   
		//aRGB1: (0            - 0           ) * 0            + Shade       
		//aA1  : (0            - 0           ) * 0            + Primitive   
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwPrimitiveColor;
			break;

		// AFA
		case 0x00fffffffffe7838LL:
		//aRGB0: (0            - 0           ) * 0            + Shade       
		//aA0  : (0            - 0           ) * 0            + Shade       
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// Select the diffuse component - this is the default
			break;

		case 0x00fffffffffe793cLL:
		//RGB0: (0            - 0           ) * 0            + Shade       
		//A0  : (0            - 0           ) * 0            + Shade       
		//RGB1: (0            - 0           ) * 0            + Shade       
		//A1  : (0            - 0           ) * 0            + Shade       
			// Select the diffuse component - this is the default
			break;

		case 0x00fffffffffe7b3dLL:
		//aRGB0: (0            - 0           ) * 0            + Shade       
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (0            - 0           ) * 0            + Shade       
		//aA1  : (0            - 0           ) * 0            + Env         
			// Select the diffuse component for RGB, env colour for alpha
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwEnvColor;
			break;

		// SFR2049
		case 0x00fffffffffe7c38LL:
		//aRGB0: (0            - 0           ) * 0            + Shade       
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// Select the diffuse component for RGB, 1.0 for alpha
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = COL32_MASK_A;
			break;

		// BM
		case 0x00fffffffffe7d3eLL:
		//aRGB0: (0            - 0           ) * 0            + Shade       
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (0            - 0           ) * 0            + Shade       
		//aA1  : (0            - 0           ) * 0            + 1    
			// Select the diffuse component for RGB, 1.0 for alpha
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = COL32_MASK_A;
			break;


		// AG
		case 0x00fffffffffefb7dLL:
		//aRGB0: (0            - 0           ) * 0            + Env         
		//aA0  : (0            - 0           ) * 0            + Env         
		//aRGB1: (0            - 0           ) * 0            + Env         
		//aA1  : (0            - 0           ) * 0            + Env     
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwEnvColor;
			break;

		// SpM
		case 0x00fffffffffff238LL:
		//aRGB0: (0            - 0           ) * 0            + 0           
		//aA0  : (0            - 0           ) * 0            + Texel0      
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = COL32_MASK_A;		// RGB = 0, A = 1
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// SpM
		case 0x00fffffffffffe38LL:
		//aRGB0: (0            - 0           ) * 0            + 0           
		//aA0  : (0            - 0           ) * 0            + 0           
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined  
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = MAKE_COLOUR_U32( 0,0,0,0 );
			break;

		// BK
		case 0x00ffffffffffffffLL:
		//aRGB0: (0            - 0           ) * 0            + 0           
		//aA0  : (0            - 0           ) * 0            + 0           
		//aRGB1: (0            - 0           ) * 0            + 0           
		//aA1  : (0            - 0           ) * 0            + 0           
			// Easy :)
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = MAKE_COLOUR_U32( 0,0,0,0 );
			break;

		//
		//	Unsorted, all OoT, just after intro
		//

		// OoT skydome?
		case 0x002527ff1ffc9238LL:
		//aRGB0: (Texel1       - Texel0      ) * Prim_Alpha   + Texel0      
		//aA0  : (Texel1       - Texel0      ) * Primitive    + Texel0      
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// XXXX placeholder implementation - is blending between T0/T1 using primitive colour
			install_texture = true;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = MAKE_COLOUR_U32_REPLICATE_ALPHA( m_dwPrimitiveColor );
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// OoT?
		case 0x00262a60150c937fLL:
		//aRGB0: (Texel1       - Texel0      ) * Env_Alpha    + Texel0      
		//aA0  : (Texel1       - Texel0      ) * Env          + Texel0      
		//aRGB1: (Primitive    - Env         ) * Combined     + Env         
		//aA1  : (Combined     - 0           ) * Primitive    + 0           
			// XXXX placeholder implementation
			install_texture = true;
			sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
			break;

		// OoT ground texture
		case 0x00267e041ffcfdf8LL:
		//aRGB0: (Texel1       - Texel0      ) * Env_Alpha    + Texel0      
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (Combined     - 0           ) * Shade        + 0           
		//aA1  : (0            - 0           ) * 0            + Combined    
			// XXXX placeholder implementation - ok except t1?
			install_texture = true;
			// RGB = Blend(T0, T1, EnvAlpha) * Shade
			// A   = 1
			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = COL32_MASK_A;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
			break;



		case 0x00267e041f0cfdffLL:
		//aRGB0: (Texel1       - Texel0      ) * Env_Alpha    + Texel0      
		//aA0  : (0            - 0           ) * 0            + 1           
		//aRGB1: (Combined     - 0           ) * Shade        + 0           
		//aA1  : (Combined     - 0           ) * Primitive    + 0           
			// XXXX placeholder implementation
			install_texture = true;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
			break;

		// OoT paths
		case 0x00121603ff5bfff8LL:
		//aRGB0: (Texel0       - 0           ) * Shade        + 0           
		//aA0  : (Texel0       - 0           ) * Primitive    + 0           
		//aRGB1: (Combined     - 0           ) * Primitive    + 0           
		//aA1  : (Texel1       - 0           ) * 1            + Combined  
			// XXXX placeholder implementation - think is correct except t1
			install_texture = true;
			// RGB = T0 * Shade * Primitive
			// A   = T0 * Primitive + T1
			if( num_cycles == 1 )
			{
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwPrimitiveColor;
			}
			else
			{
				gModulateDiffuseMask = COL32_MASK_RGB;
				gModulateDiffuseColour = m_dwPrimitiveColor;
				gSetDiffuseMask = COL32_MASK_A;
				gSetDiffuseColour = m_dwPrimitiveColor;
			}
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;


		// Zelda OoT logo flames
		case 0x00272c60350ce37fLL:
		//aRGB0: (Texel1       - Primitive   ) * PrimLODFrac  + Texel0      
		//aA0  : (Texel1       - 1           ) * 1            + Texel0      
		//aRGB1: (Primitive    - Env         ) * Combined     + Env         
		//aA1  : (Combined     - 0           ) * Primitive    + 0           
			// XXXX placeholder implementation
			install_texture = true;

			if( num_cycles == 1 )
			{
				// RGB = T0 + x(T1-Prim)
				// A   = T0+T1-1
				sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);		// XXXX No T1
			}
			else
			{
				// RGB = Blend(Env, Prim, (T0 + x(t1-Prim)))
				// A   = (T0+T1-1)*Prim
				gSetDiffuseMask = COL32_MASK_RGBA;
				gSetDiffuseColour = (m_dwEnvColor & COL32_MASK_RGB) | (m_dwPrimitiveColor & COL32_MASK_A);

				sceGuTexEnvColor( m_dwPrimitiveColor );
				sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);		// XXXX No T1

			}

			break;

		case 0x0030b2045ffefff8LL:
		//aRGB0: (Primitive    - Env         ) * Texel0       + Env         
		//aA0  : (Primitive    - 0           ) * Texel0       + 0           
		//aRGB1: (Combined     - 0           ) * Shade        + 0           
		//aA1  : (0            - 0           ) * 0            + Combined    
			// XXXX placeholder implementation - correct for 1 cycle?
			install_texture = true;

			// Use Env for RGB, Prim for A
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = (m_dwEnvColor & COL32_MASK_RGB) | (m_dwPrimitiveColor & COL32_MASK_A);

			sceGuTexEnvColor( m_dwPrimitiveColor );

			if( num_cycles == 1 )
			{
				// RGB = Blend(Env,Prim,T0)
				// A   = Prim * T0
			}
			else
			{
				// RGB = Blend(Env,Prim,T0) * Shade
				// A   = Prim * T0

				// Can't handle shade for second cycle :(
			}
			sceGuTexFunc(GU_TFX_BLEND,GU_TCC_RGBA);	// _RGBA modulate
			break;

		// OoT, letters
		case 0x00ffadfffffd9238LL:
		//aRGB0: (0            - 0           ) * 0            + Primitive   
		//aA0  : (Texel1       - Texel0      ) * 1            + Texel0      
		//aRGB1: (0            - 0           ) * 0            + Combined    
		//aA1  : (0            - 0           ) * 0            + Combined    
			// XXXX placeholder implementation - needs t1
			// RGB = Prim
			// A   = (T1-T0)+T0 = T1
			install_texture = true;
			gSetTextureMask = COL32_MASK_RGB;
			gSetTextureColour = COL32_WHITE;
			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor | COL32_MASK_A;		// Want to select texture alpha
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);
			break;

		// MKRT ingame - cars, balloons etc
		case 0x0060b2c15565feffLL:
		//aRGB0: (1            - Env         ) * Texel0       + Primitive   
		//aA0  : (Primitive    - 0           ) * Texel0       + 0           
		//aRGB1: (1            - Env         ) * Texel0       + Primitive   
		//aA1  : (Primitive    - 0           ) * Texel0       + 0           
			install_texture = true;
			//XXXX placeholder - seems okish

			gSetDiffuseMask = COL32_MASK_RGBA;
			gSetDiffuseColour = m_dwPrimitiveColor;

			//sceGuTexEnvColor( MAKE_COLOUR_U32_RGB_1MC( m_dwEnvColor ) );	// 1-r, 1-g, 1-b, a
			sceGuTexFunc(GU_TFX_ADD,GU_TCC_RGBA);			// in RGBA mode alpha = texture*diffuse fragment
			break;

		// TSph
		case 0x0026a0041f0c93ffLL:
		//aRGB0: (Texel1       - Texel0      ) * LOD_Frac     + Texel0      
		//aA0  : (Texel1       - Texel0      ) * Combined     + Texel0      
		//aRGB1: (Combined     - 0           ) * Shade        + 0           
		//aA1  : (Combined     - 0           ) * Primitive    + 0           
			install_texture = true;
			//XXXX placeholder

			gSetDiffuseMask = COL32_MASK_A;
			gSetDiffuseColour = m_dwPrimitiveColor;
			sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGBA);			// in RGBA mode alpha = texture*diffuse fragment
			break;

		default:
			found = false;
			break;
	}

	bool	installed_texture( false );

	if(install_texture)
	{
		if(m_lpsTexture[0] != NULL)
		{
			CNativeTexture * p_texture;

			if(gSetTextureMask != 0)
			{
				p_texture = m_lpsTexture[0]->GetRecolouredTexture( gSetTextureMask, gSetTextureColour );
			}
			else
			{
				p_texture = m_lpsTexture[0]->GetTexture();
			}

			if(p_texture != NULL)
			{
				p_texture->InstallTexture();
				installed_texture = true;
			}
		}
	}

	// If no texture was specified, or if we couldn't load it, clear it out
	if( !installed_texture )
	{
		sceGuDisable(GU_TEXTURE_2D);
	}

	if(!found)
	{
		if(mUnhandledCombinderStates.find( gRDPMux._u64 ) == mUnhandledCombinderStates.end())
		{
			char szFilePath[MAX_PATH+1];

			Dump_GetDumpDirectory(szFilePath, g_ROM.settings.szGameName);

			IO::Path::Append(szFilePath, "missing_mux.txt");

			FILE * fh( fopen(szFilePath, mUnhandledCombinderStates.empty() ? "w" : "a") );
			if(fh != NULL)
			{
				PrintMux( fh, gRDPMux._u64 );
				fclose(fh);
			}

			mUnhandledCombinderStates.insert( gRDPMux._u64 );
		}
		

		sceGuEnable(GU_TEXTURE_2D);
		sceGuTexMode(GU_PSM_8888,0,0,GL_TRUE);		// maxmips/a2/swizzle = 0

		// Use the nasty placeholder texture
		sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gPlaceholderTexture);
		sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGBA);
	}
}


//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::PrepareVerticesForBlendMode( DaedalusVtx * p_sw_vertices, u32 num_verts )
{
	DAEDALUS_PROFILE( "PSPRenderer::PrepareVerticesForBlendMode" );

	DAEDALUS_ASSERT( (gSetDiffuseMask & gModulateDiffuseMask) == 0, "Setting and modulating the same component" );

	if( gSetDiffuseMask != 0 )
	{
		if( gSetDiffuseMask == COL32_MASK_RGBA )
		{
			for(u32 v = 0; v < num_verts; v++)
			{
				p_sw_vertices[v].color = gSetDiffuseColour;
			}
		}
		else
		{
			u32		clear_bits( ~gSetDiffuseMask );
			u32		set_bits( gSetDiffuseColour & gSetDiffuseMask );

			for(u32 v = 0; v < num_verts; v++)
			{
				p_sw_vertices[v].color = (p_sw_vertices[v].color & clear_bits) | set_bits;
			}
		}
	}

	if(gModulateDiffuseMask != 0)
	{
		if( gModulateDiffuseMask == COL32_MASK_RGB )
		{
			for(u32 v = 0; v < num_verts; v++)
			{
				p_sw_vertices[v].color = MODULATE_COLOUR_RGB(p_sw_vertices[v].color, gModulateDiffuseColour);
			}
		}
		else if( gModulateDiffuseMask == COL32_MASK_A )
		{
			for(u32 v = 0; v < num_verts; v++)
			{
				p_sw_vertices[v].color = MODULATE_COLOUR_A(p_sw_vertices[v].color, gModulateDiffuseColour);
			}
		}
		else if( gModulateDiffuseMask == COL32_MASK_RGBA )
		{
			for(u32 v = 0; v < num_verts; v++)
			{
				p_sw_vertices[v].color = MODULATE_COLOUR_RGBA(p_sw_vertices[v].color, gModulateDiffuseColour);
			}
		}
		else
		{
			DAEDALUS_ASSERTMSG( "Unhandled gModulateDiffuseMask" );
		}
	}

}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::InitCycle1()
{
	DL_PF("CycleType: 1");
	
	BodgeBlendMode( 1 );
}

//*************************************************************************************
// 
//*************************************************************************************
void	PSPRenderer::InitCycle2()
{
	DL_PF("CycleType: 2");

	BodgeBlendMode( 2 );
}

//*****************************************************************************
// States set/modified by this mode:
// SetTextureStateState COLOROP/ALPHAOP
// SetTexture
// D3DRS_TEXTUREFACTOR
// D3DRS_ZENABLE
// D3DRS_ZWRITEENABLE
// ---D3DRS_TEXTUREPERSPECTIVE 
// D3DRS_ALPHATESTENABLE
// D3DRS_SRCBLEND
// D3DRS_DESTBLEND
// D3DRS_ALPHABLENDENABLE
//*****************************************************************************
RDP_OtherMode	gLastRDPOtherMode;
RDP_Mux			gLastRDPMux;
bool			gTextureChanged = false;
u32				gLastPrimColour = 0x00000000;
u32				gLastEnvColour = 0x00000000;

bool			gLastUseZBuffer = false;
bool			gNeedZBufferUdate = false;


void PSPRenderer::InitBlendMode()
{
	DAEDALUS_PROFILE( "PSPRenderer::InitBlendMode" );

	if ( gNeedZBufferUdate ||
		(gRDPOtherMode._u64 != gLastRDPOtherMode._u64) ||
		 (m_bZBuffer != gLastUseZBuffer) )
	{
		// Only update if ZBuffer is enabled
		if (m_bZBuffer)
		{
			if(gRDPOtherMode.z_cmp)
				sceGuEnable(GU_DEPTH_TEST);
			else
				sceGuDisable(GU_DEPTH_TEST);

			sceGuDepthMask( gRDPOtherMode.z_upd ? GL_FALSE : GL_TRUE );	// GL_TRUE to disable z-writes
		}
		else
		{
			sceGuDisable(GU_DEPTH_TEST);
			sceGuDepthMask( GL_TRUE );	// GL_TRUE to disable z-writes
		}

		gNeedZBufferUdate = false;
		gLastUseZBuffer = m_bZBuffer;
	}


	/*if ( gRDPOtherMode._u64 == gLastRDPOtherMode._u64 &&
		 gRDPMux._u64 == gLastRDPMux._u64 &&
		 (gLastPrimColour == m_dwPrimitiveColor ) &&
		 (gLastEnvColour == m_dwEnvColor ) &&
		 gTextureChanged == false )*/
	{
		//static u32 num_ignored = 0;
		//num_ignored++;
		//if ( (num_ignored%10000) == 0)
		//	DBGConsole_Msg( 0, "Ignoring redundant state initialisation %x", num_ignored );
		//return;
	}

	gLastRDPOtherMode._u64 = gRDPOtherMode._u64;
	gLastRDPMux._u64 = gRDPMux._u64;
	gTextureChanged = false;
	gLastPrimColour = m_dwPrimitiveColor;
	gLastEnvColour = m_dwEnvColor;

	switch ( gRDPOtherMode.cycle_type )
	{
		case CYCLE_COPY:			InitCycleCopy();	break;
		case CYCLE_FILL:			InitCycleFill();	break;
		case CYCLE_1CYCLE:			InitCycle1();		break;
		case CYCLE_2CYCLE:			InitCycle2();		break;
	}

	sceGuShadeModel( mSmooth ? GU_SMOOTH : GU_FLAT );
	
	switch ( gRDPOtherMode.text_filt )
	{
		case 0:			//G_TF_POINT:	// 0
			sceGuTexFilter(GU_NEAREST,GU_NEAREST);
			break;
		case 2:			//G_TF_BILERP:	// 2
			sceGuTexFilter(GU_LINEAR,GU_LINEAR);
			break;
		case 3:			//G_TF_AVERAGE:	// 3?
			sceGuTexFilter(GU_LINEAR,GU_LINEAR);
			break;
	}

#define BLEND_NOOP1				0x00000000
#define BLEND_NOOP2				0x00000000

#define BLEND_FOG_ASHADE1		0xc8000000
#define BLEND_FOG_APRIM1		0xc4000000

#define BLEND_PASS1				0x0c080000
#define BLEND_PASS2				0x03020000
	
#define BLEND_OPA1				0x00440000
#define BLEND_OPA2				0x00110000

#define BLEND_XLU1				0x00400000
#define BLEND_XLU2				0x00100000

#define BLEND_ADD1				0x04400000
#define BLEND_ADD2				0x01100000

#define BLEND_MEM1				0x4c400000		// Mem*0 + Mem*(1-0)?!
#define BLEND_MEM2				0x13100000		// Mem*0 + Mem*(1-0)?!


/*

#define	G_BL_CLR_IN	0
#define	G_BL_CLR_MEM	1
#define	G_BL_CLR_BL	2
#define	G_BL_CLR_FOG	3
#define	G_BL_1MA	0
#define	G_BL_A_MEM	1
#define	G_BL_A_IN	0
#define	G_BL_A_FOG	1
#define	G_BL_A_SHADE	2
#define	G_BL_1		2
#define	G_BL_0		3

#define	GBL_c1(m1a, m1b, m2a, m2b)	\
	(m1a) << 30 | (m1b) << 26 | (m2a) << 22 | (m2b) << 18
#define	GBL_c2(m1a, m1b, m2a, m2b)	\
	(m1a) << 28 | (m1b) << 24 | (m2a) << 20 | (m2b) << 16

#define	RM_AA_ZB_OPA_SURF(clk)					\
	AA_EN | Z_CMP | Z_UPD | IM_RD | CVG_DST_CLAMP |		\
	ZMODE_OPA | ALPHA_CVG_SEL |				\
	GBL_c##clk(G_BL_CLR_IN, G_BL_A_IN, G_BL_CLR_MEM, G_BL_A_MEM)

#define	G_RM_FOG_SHADE_A		GBL_c1(G_BL_CLR_FOG, G_BL_A_SHADE, G_BL_CLR_IN, G_BL_1MA)
#define	G_RM_FOG_PRIM_A			GBL_c1(G_BL_CLR_FOG, G_BL_A_FOG, G_BL_CLR_IN, G_BL_1MA)
#define	G_RM_PASS				GBL_c1(G_BL_CLR_IN, G_BL_0, G_BL_CLR_IN, G_BL_1)
#define	RM_AA_ZB_OPA_SURF(clk)	GBL_c##clk(G_BL_CLR_IN, G_BL_A_IN, G_BL_CLR_MEM, G_BL_A_MEM)
*/
	u32 blendmode_1 = u32( gRDPOtherMode._u32[0] & 0xcccc0000 );
	u32 blendmode_2 = u32( gRDPOtherMode._u32[0] & 0x33330000 );

	if ( gRDPOtherMode.cycle_type == CYCLE_FILL )
	{
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
	}
	else if ( blendmode_1 == 0 && blendmode_2 == 0 )
	{
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: NOOP/NOOP" );
	}
	else if ( blendmode_1 == BLEND_FOG_ASHADE1 && blendmode_2 == BLEND_OPA2 )
	{
		// c811
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: FOG_ASHADE/OPA" );
	}
	else if ( blendmode_1 == BLEND_FOG_ASHADE1 && blendmode_2 == BLEND_XLU2 )
	{
		// c810
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: FOG_ASHADE/XLU" );
	}
	else if ( blendmode_1 == BLEND_XLU1 && blendmode_2 == BLEND_XLU2 )
	{
		// 0x0050
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: XLU/XLU" );
	}
	else if ( blendmode_1 == BLEND_XLU1 && blendmode_2 == BLEND_NOOP2 )
	{
		// 0x0040
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: XLU/NOOP" );
	}
	else if ( blendmode_1 == BLEND_XLU1 && blendmode_2 == BLEND_ADD2 )
	{
		// 0x005n
		// XXXX
		sceGuBlendFunc(GU_ADD, GU_SRC_COLOR, GU_DST_COLOR, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: XLU/ADD" );
	}
	else if ( blendmode_1 == BLEND_OPA1 && blendmode_2 == BLEND_OPA2 )
	{
		// 0x0055
		// Would have thoughts this was SRCALPHA/DSTALPHA, but apparently not
		sceGuDisable(GU_BLEND);		
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: OPA/OPA" );
	}
	else if ( blendmode_1 == BLEND_OPA1 && blendmode_2 == BLEND_NOOP2 )
	{
		// 0x0044
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: OPA/NOOP" );
	}
	else if ( blendmode_1 == BLEND_PASS1 && blendmode_2 == BLEND_PASS2 )
	{
		// 0x0f0a
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: PASS/PASS" );
	}
	else if ( blendmode_1 == BLEND_PASS1 && blendmode_2 == BLEND_XLU2 )
	{
		// 0x0c18
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: PASS/XLU" );
	}
	else if ( blendmode_1 == BLEND_PASS1 && blendmode_2 == BLEND_OPA2 )
	{
		// 0x0c19
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: PASS/OPA" );
	}
	else if ( blendmode_1 == BLEND_MEM1 && blendmode_2 == BLEND_MEM2 )
	{
		// 0x5f50
		sceGuDisable(GU_BLEND);
		sceGuDisable(GU_ALPHA_TEST);
		DL_PF( "      Blend: MEM/MEM" );
	}
	else
	{
		sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
		sceGuEnable(GU_BLEND);
		sceGuEnable(GU_ALPHA_TEST);
		DL_PF( "      Blend: SRCALPHA/INVSRCALPHA (default: 0x%04x)", gRDPOtherMode.blender );
	}
	

	//
	// I can't think why the hand in mario's menu screen is rendered with an opaque rendermode,
	// and no alpha threshold. We set the alpha reference to 1 to ensure that the transparent pixels
	// don't get rendered. I hope this doesn't fuck anything else up.
	//
	if ( gRDPOtherMode.alpha_compare == 0 )
	{
		if ( gRDPOtherMode.cvg_x_alpha )
		{
			sceGuAlphaFunc(GU_GEQUAL, 0x01, 0xff );
			sceGuEnable(GU_ALPHA_TEST);
		}
		else
		{
			sceGuDisable(GU_ALPHA_TEST);
		}
	}
	else
	{
		// G_AC_THRESHOLD || G_AC_DITHER
		sceGuAlphaFunc(GU_GEQUAL, m_dwAlpha, 0xff );
		sceGuEnable(GU_ALPHA_TEST);
	}

}



//*****************************************************************************
//
//*****************************************************************************
inline f32 round( f32 x )
{
	return (f32)(s32)( x + 0.5f );
}

bool PSPRenderer::TexRect( const v2 & xy0, const v2 & xy1, const v2 & uv0, const v2 & uv1 )
{
	v2 screen0;
	v2 screen1;
	v2 tex_uv0( uv0 );
	v2 tex_uv1( uv1 );

	// Scale to Actual texture coords
	// The two cases are to handle the oversized textures hack on voodoos
#ifdef TEXRECT_USE_3D_TRANSFORM
	float width_scale = (m_fTileWidth[0] > m_fTexWidth[0]) ? m_fInvTileWidth[0] : m_fInvTexWidth[0];
	float height_scale = (m_fTileHeight[0] > m_fTexHeight[0]) ? m_fInvTileHeight[0] : m_fInvTexHeight[0];

	tex_uv0.x *= width_scale;
	tex_uv1.x *= width_scale;

	tex_uv0.y *= height_scale;
	tex_uv1.y *= height_scale;
#endif

	//
	// In Fill/Copy mode the coordinates are inclusive (i.e. add 1.0f to the w/h)
	//
	v2 edge( 0, 0 );

	switch ( gRDPOtherMode.cycle_type )
	{
		case CYCLE_COPY:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_FILL:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_1CYCLE:											break;
		case CYCLE_2CYCLE:											break;
	}


	//
	// We round these value here, so that when we scale up the coords to our screen
	// coords we don't get any gaps.
	//
	screen0.x = round( xy0.x * m_fScreenMultX );
	screen0.y = round( xy0.y * m_fScreenMultY );

	screen1.x = round( (xy1.x + edge.x) * m_fScreenMultX );
	screen1.y = round( (xy1.y + edge.x) * m_fScreenMultY );

#ifdef DEBUG_TEXRECT
	DL_PF( "      ScreenMult:  %f,%f", m_fScreenMultX, m_fScreenMultY );
	DL_PF( "      Screen:  %f,%f -> %f,%f", screen0.x, screen0.y, screen1.x, screen1.y );
	DL_PF( "      Texture: %f,%f -> %f,%f", tex_uv0.x, tex_uv0.y, tex_uv1.x, tex_uv1.y );
#endif

#ifdef TEXRECT_USE_3D_TRANSFORM
	DaedalusVtx trv[4];
	static const u16 indices[2*3] = {1,0,2, 2,0,3};

	trv[0].pos = v3( screen0.x, screen0.y, gTexRectDepth );
	trv[0].color = 0xffffffff;
	trv[0].t0 = v2( tex_uv0.x, tex_uv0.y );

	trv[1].pos = v3( screen1.x, screen0.y, gTexRectDepth );
	trv[1].color = 0xffffffff;
	trv[1].t0 = v2( tex_uv1.x, tex_uv0.y );

	trv[2].pos = v3( screen1.x, screen1.y, gTexRectDepth );
	trv[2].color = 0xffffffff;
	trv[2].t0 = v2( tex_uv1.x, tex_uv1.y );

	trv[3].pos = v3( screen0.x, screen1.y, gTexRectDepth );
	trv[3].color = 0xffffffff;
	trv[3].t0 = v2( tex_uv0.x, tex_uv1.y );

	return RenderTriangleList( trv, 4, indices, 6, true, true );
#else
	DaedalusVtx trv[2];

	trv[0].pos = v3( screen0.x, screen0.y, gTexRectDepth );
	trv[0].color = 0xffffffff;
	trv[0].t0 = v2( tex_uv0.x, tex_uv0.y );

	trv[1].pos = v3( screen1.x, screen1.y, gTexRectDepth );
	trv[1].color = 0xffffffff;
	trv[1].t0 = v2( tex_uv1.x, tex_uv1.y );

	return RenderSprites( trv, 2, true, true );
#endif
}

//*****************************************************************************
//
//*****************************************************************************
bool PSPRenderer::TexRectFlip( const v2 & xy0, const v2 & xy1, const v2 & uv0, const v2 & uv1 )
{
	// TODO - In 1/2cycle mode, skip bottom/right edges!?
	v2 screen0;
	v2 screen1;
	v2 tex_uv0( uv0 );
	v2 tex_uv1( uv1 );

	// Scale to Actual texture coords
	// The two cases are to handle the oversized textures hack on voodoos
#ifdef TEXRECT_USE_3D_TRANSFORM
	float width_scale = (m_fTileWidth[0] > m_fTexWidth[0]) ? m_fInvTileWidth[0] : m_fInvTexWidth[0];
	float height_scale = (m_fTileHeight[0] > m_fTexHeight[0]) ? m_fInvTileHeight[0] : m_fInvTexHeight[0];

	tex_uv0.x *= width_scale;
	tex_uv1.x *= width_scale;

	tex_uv0.y *= height_scale;
	tex_uv1.y *= height_scale;
#endif

	//
	// In Fill/Copy mode the coordinates are inclusive (i.e. add 1.0f to the w/h)
	//
	v2 edge( 0, 0 );

	switch ( gRDPOtherMode.cycle_type )
	{
		case CYCLE_COPY:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_FILL:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_1CYCLE:											break;
		case CYCLE_2CYCLE:											break;
	}

	//
	// We round these value here, so that when we scale up the coords to our screen
	// coords we don't get any gaps.
	//
	screen0.x = round( xy0.x * m_fScreenMultX );
	screen0.y = round( xy0.y * m_fScreenMultY );

	screen1.x = round( (xy1.x + edge.x) * m_fScreenMultX );
	screen1.y = round( (xy1.y + edge.x) * m_fScreenMultY );

#ifdef DEBUG_TEXRECT
	DL_PF( "      ScreenMult:  %f,%f", m_fScreenMultX, m_fScreenMultY );
	DL_PF( "      Screen:  %f,%f -> %f,%f", screen0.x, screen0.y, screen1.x, screen1.y );
	DL_PF( "      Texture: %f,%f -> %f,%f", tex_uv0.x, tex_uv0.y, tex_uv1.x, tex_uv1.y );
#endif

#ifdef TEXRECT_USE_3D_TRANSFORM
	DaedalusVtx trv[4];
	static const u16 indices[2*3] = {1,0,2, 2,0,3};

	// Same as TexRect, but with texcoords 0,2 swapped
	trv[0].pos = v3( screen0.x, screen0.y, gTexRectDepth );
	trv[0].color = 0xffffffff;
	trv[0].t0 = v2( tex_uv0.x, tex_uv0.y );
	
	trv[1].pos = v3( screen1.x, screen0.y, gTexRectDepth );
	trv[1].color = 0xffffffff;
	trv[1].t0 = v2( tex_uv0.x, tex_uv1.y );		// In TexRect this is tex_uv1.x, tex_uv0.y

	trv[2].pos = v3( screen1.x, screen1.y, gTexRectDepth );
	trv[2].color = 0xffffffff;
	trv[2].t0 = v2( tex_uv1.x, tex_uv1.y );

	trv[3].pos = v3( screen0.x, screen1.y, gTexRectDepth );
	trv[3].color = 0xffffffff;
	trv[3].t0 = v2( tex_uv1.x, tex_uv0.y );		//tex_uv0.x	tex_uv1.y	

	return RenderTriangleList( trv, 4, indices, 6, true, true );
#else
	DaedalusVtx trv[2];

	trv[0].pos = v3( screen0.x, screen0.y, gTexRectDepth );
	trv[0].color = 0xffffffff;
	trv[0].t0 = v2( tex_uv0.x, tex_uv0.y );

	trv[1].pos = v3( screen1.x, screen1.y, gTexRectDepth );
	trv[1].color = 0xffffffff;
	trv[1].t0 = v2( tex_uv1.x, tex_uv1.y );

	return RenderSprites( trv, 2, true, true );
#endif
}

//*****************************************************************************
//
//*****************************************************************************
bool PSPRenderer::FillRect( const v2 & xy0, const v2 & xy1, u32 color )
{
	v2 screen0;
	v2 screen1;

	if ( (gRDPOtherMode._u32[0] & 0xffff0000) == 0x5f500000 )
	{
		// this blend mode is mem*0 + mem*1, so we don't need to render it... Very odd!
		return true;
	}

	// This if for C&C - It might break other stuff (I'm not sure if we should allow alpha or not..)
//	color |= 0xff000000;

	//
	// We round these value here, so that when we scale up the coords to our screen
	// coords we don't get any gaps.
	//
	screen0.x = round( xy0.x * m_fScreenMultX );
	screen0.y = round( xy0.y * m_fScreenMultY );

	screen1.x = round( xy1.x * m_fScreenMultX );
	screen1.y = round( xy1.y * m_fScreenMultY );

	DL_PF( "      ScreenMult:  %f,%f", m_fScreenMultX, m_fScreenMultY );
	DL_PF( "      Screen:  %f,%f -> %f,%f", screen0.x, screen0.y, screen1.x, screen1.y );

#ifdef TEXRECT_USE_3D_TRANSFORM
	DaedalusVtx frv[4];
	static const u16 indices[2*3] = {1,0,2, 2,0,3};

	frv[0].pos = v3(screen0.x, screen0.y, gFillRectDepth );
	frv[0].color = color;
	frv[0].t0 = v2( 0, 0 );

	frv[1].pos = v3(screen1.x, screen0.y, gFillRectDepth );
	frv[1].color = color;
	frv[1].t0 = v2( 1, 0 );

	frv[2].pos = v3(screen1.x, screen1.y, gFillRectDepth );
	frv[2].color = color;
	frv[2].t0 = v2( 1, 1 );

	frv[3].pos = v3(screen0.x, screen1.y, gFillRectDepth );
	frv[3].color = color;
	frv[3].t0 = v2( 0, 1 );

	return RenderTriangleList( frv, 4, indices, 6, true );
#else
	DaedalusVtx trv[2];

	trv[0].pos = v3( screen0.x, screen0.y, gFillRectDepth );
	trv[0].color = color;
	trv[0].t0 = v2( 0, 0 );

	trv[1].pos = v3( screen1.x, screen1.y, gFillRectDepth );
	trv[1].color = color;
	trv[1].t0 = v2( 1, 1 );

	return RenderSprites( trv, 2, true, false );
#endif
}

//*****************************************************************************
// Returns true if triangle visible and rendered, false otherwise
//*****************************************************************************
bool PSPRenderer::AddTri(u32 v0, u32 v1, u32 v2)
{
	if ( TestTri( v0, v1, v2 ) )
	{
		DL_PF("   Tri: %d,%d,%d", v0, v1, v2);

		m_swIndexBuffer[ m_dwNumIndices + 0 ] = (u16)v0;
		m_swIndexBuffer[ m_dwNumIndices + 1 ] = (u16)v1;
		m_swIndexBuffer[ m_dwNumIndices + 2 ] = (u16)v2;

		m_dwNumIndices += 3;

		m_dwNumTrisRendered++;
		return true;
	}
	else
	{
		DL_PF("   Tri: %d,%d,%d (clipped)", v0, v1, v2);

		m_dwNumTrisClipped++;
		return false;
	}
}

//*****************************************************************************
//
//*****************************************************************************
// Returns true if it thinks the triangle is visible
// Returns false if it is clipped
bool PSPRenderer::TestTri(u32 v0, u32 v1, u32 v2) const
{
#ifdef USE_NATIVE_PROJECTION
	// Don't cull for now!
	return true;
#endif

	//return true;
	//
	// Render all the triangles while exporting
	//
	// Check vertices are valid!
	if ( v0 >= MAX_VERTS || v1 >= MAX_VERTS || v2 >= MAX_VERTS )
		return false;

	// Here we AND all the flags. If any of the bits is set for all
	// 3 vertices, it means that all three x, y or z lie outside of
	// the current viewing volume.
	// Currently disabled - still seems a bit dodgy
#ifdef SOFTWARE_CLIPPING
	if (m_dwVecFlags[v0] & m_dwVecFlags[v1] & m_dwVecFlags[v2])
	{
		//DL_PF("0x%04x 0x%04x 0x%04x", m_dwVecFlags[dwV0], m_dwVecFlags[dwV1], m_dwVecFlags[dwV2]);
		return false;
	}
	else
#endif
	if ((m_bCullFront || m_bCullBack) && !IsZModeDecal())
	{
		const v4 & vert0 = mVtxProjected[v0];
		const v4 & vert1 = mVtxProjected[v1];
		const v4 & vert2 = mVtxProjected[v2];

		// Only try to clip if the tri is onscreen. For some reason, this
		// method doesnt' work well when the z value is outside of screenspace
		//if (vert0.z < 1 && vert1.z < 1 && vert2.z < 1)
		{
			float V1 = vert2.x - vert0.x;
			float V2 = vert2.y - vert0.y;
			
			float W1 = vert2.x - vert1.x;
			float W2 = vert2.y - vert1.y;

			//
			//	Sign flipped to avoid extra float op
			//
			float fDirection = ( (V1 * W2) - (V2 * W1) );

			// Account for being behind the camera
			fDirection *= vert0.w * vert1.w * vert2.w;

			if (fDirection < 0 && m_bCullBack)
				return false;
			else if (fDirection > 0 && m_bCullFront)
				return false;

		}
	}
	return true;
}


//*****************************************************************************
//
//*****************************************************************************
bool PSPRenderer::TestVerts( u32 v0, u32 vn ) const
{
	u32 flags = m_dwVecFlags[v0];

	for ( u32 i = v0+1; i <= vn && i < MAX_VERTS; i++ )
	{
		flags &= m_dwVecFlags[i];
	}

	return flags == 0;
}

//*****************************************************************************
//
//*****************************************************************************
u32 PSPRenderer::LightVert( const v3 & norm )
{
	float r,g,b;
	float fCosT;
	s32 l;

	// Ip = Light intensity
	// Ia = Ambient light intensity
	// Ka = Material ambient reflective property
	// Kd = Material diffuse reflective property
	// L = normalised light direction
	// N = vertex normal

	// I = IaKa + IpKd (N . L)

	// Do ambient
	r = m_AmbientCol.r;
	g = m_AmbientCol.g;
	b = m_AmbientCol.b;
	//a = m_fAmbientA;

	for ( l = 0; l < m_dwNumLights; l++ )
	{
		//DAEDALUS_ASSERT( m_Lights[l].initialised, "Light isn't initialised!" );
		fCosT = norm.Dot( m_Lights[l].dir );
		if (fCosT > 0)
		{
			r += m_Lights[l].col.r * fCosT;
			g += m_Lights[l].col.g * fCosT;
			b += m_Lights[l].col.b * fCosT;
		//	a += m_Lights[l].col.a * fCosT;
		}
	}

	if (r > 255.0f) r = 255.0f;
	if (g > 255.0f) g = 255.0f;
	if (b > 255.0f) b = 255.0f;
	//if (a > 255.0f) a = 255.0f;

	return MAKE_COLOUR_U32((s32)r, (s32)g, (s32)b, (s32)255);

}

//*****************************************************************************
//
//*****************************************************************************
bool PSPRenderer::FlushTris()
{
	DAEDALUS_PROFILE( "PSPRenderer::FlushTris" );

	if ( m_dwNumIndices == 0 )
		return true;

	u16 *			p_sw_indices( m_swIndexBuffer );
	DaedalusVtx *	p_sw_vertices( NULL );

	p_sw_indices = (u16*)sceGuGetMemory(m_dwNumIndices*sizeof(u16));
	memcpy( p_sw_indices, m_swIndexBuffer, m_dwNumIndices*sizeof(u16));

	//
	// Set up the vertices. We could optimise this by re-indexing the vertices
	// so that we only send the used vertices to the card. Here we check 
	// for the greatest index used, and use this for our vertex number
	//
	u32 min_vert_index = MAX_VERTS;
	u32 max_vert_index = 0;
	for ( u32 i = 0; i < m_dwNumIndices; i++ )
	{
		if ( p_sw_indices[ i ] > max_vert_index )
			max_vert_index = p_sw_indices[ i ];
		if ( p_sw_indices[ i ] < min_vert_index )
			min_vert_index = p_sw_indices[ i ];
	}

	if ( min_vert_index > 0 )
	{
		//DBGConsole_Msg( 0, "Min index is %x (max is %x)", min_vert_index, max_vert_index );
		for ( u32 i = 0; i < m_dwNumIndices; i++ )
		{
			p_sw_indices[ i ] -= (u16)min_vert_index;
		}

		max_vert_index -= min_vert_index;
	}

	m_dwNumVertices = max_vert_index + 1;

	if ( m_dwNumVertices == 0 )
	{
		m_dwNumIndices = 0;
		return true;
	}

	if(gSoftwareClipping)
	{
		DAEDALUS_PROFILE( "Software Clipping" );

		static std::vector<SClipVertex>	clip_vertices;

		clip_vertices.reserve( m_dwNumIndices );
		clip_vertices.erase( clip_vertices.begin(), clip_vertices.end() );		// Some implementations of .clear() delete the underlying memory

		for(u32 i = 0; i < m_dwNumIndices; ++i)
		{
			SClipVertex			v;
			u16					idx( p_sw_indices[i]+min_vert_index );
			const DaedalusVtx &	in( m_VtxTransformed[idx] );

			v.x = in.pos.x;
			v.y = in.pos.y;
			v.z = in.pos.z;
			v.rhw = mVtxProjected[idx].w;
			v.r = COL_GETR(in.color);
			v.g = COL_GETG(in.color);
			v.b = COL_GETB(in.color);
			v.a = COL_GETA(in.color);
			v.specular = 0;
			v.tcoord[0].u = in.t0.x;
			v.tcoord[0].v = in.t0.y;

			clip_vertices.push_back( v );
		}

		ClipVertexes( &clip_vertices[0], clip_vertices.size(), gRDPOtherMode.z_cmp, m_bFogEnable );

		m_dwNumVertices = g_clippedVtxCount;
		p_sw_vertices = (DaedalusVtx*)sceGuGetMemory(m_dwNumVertices*sizeof(DaedalusVtx));

		for(u32 i = 0; i < m_dwNumVertices; ++i)
		{
			DaedalusVtx &			out( p_sw_vertices[i] );
			const SClipVertex &		in( g_clippedVtxBuffer[i] );
			
			out.pos = v3( in.x, in.y, in.z );
			out.color = MAKE_COLOUR_U32( in.r, in.g, in.b, in.a );
			out.t0 = v2( in.tcoord[0].u, in.tcoord[0].v );
		}
	}
	else if(gTesselateLargeTriangles)
	{
		DAEDALUS_PROFILE( "Tesselate Large Tris" );

		u32 vertex_count = TesselateVerts( NULL, m_VtxTransformed + min_vert_index, p_sw_indices, m_dwNumIndices, m_mProjection[m_dwProjectionTop] );

		DL_PF( "In: %x - Out: %x", m_dwNumVertices, vertex_count );

		if(vertex_count == 0)
		{
			//
			//	All rejected
			//
			m_dwNumVertices = 0;
			m_dwNumIndices = 0;
			return true;
		}

		p_sw_vertices = (DaedalusVtx*)sceGuGetMemory( vertex_count * sizeof(DaedalusVtx) );

		// Repeat, but pass in array so we can generate the vertices
		u32	new_count = TesselateVerts( p_sw_vertices, m_VtxTransformed + min_vert_index, p_sw_indices, m_dwNumIndices, m_mProjection[m_dwProjectionTop] );

		DL_PF( " - Tesselated: %x\n", new_count );

		DAEDALUS_ASSERT( new_count == vertex_count, "Didn't produce the expected number of triangles!" );

		m_dwNumVertices = vertex_count;

		gTesselatorVerticesIn += m_dwNumIndices;
		gTesselatorVerticesOut += m_dwNumVertices;
	}
	else
	{
		DAEDALUS_PROFILE( "Copy Vertices" );

		p_sw_vertices = (DaedalusVtx*)sceGuGetMemory(m_dwNumVertices*sizeof(DaedalusVtx));
		memcpy( p_sw_vertices, &m_VtxTransformed[ min_vert_index ], m_dwNumVertices * sizeof( DaedalusVtx ) );
	}

	//
	// Process the software vertex buffer to apply a couple of 
	// necessary changed to the texture coords (this is required
	// because some ucodes set the texture after setting the vertices)
	//
	if (m_bTextureEnable)
	{
		// Only scale if not generated automatically
		if ( !( m_bTextureGen && m_bLightingEnable ) )
		{
			for ( u32 v = 0; v < m_dwNumVertices; v++ )
			{
				p_sw_vertices[v].t0.x = (p_sw_vertices[v].t0.x - m_fTexLeft[0]) * m_fInvTexWidth[0];
				p_sw_vertices[v].t0.y = (p_sw_vertices[v].t0.y - m_fTexTop[0]) * m_fInvTexHeight[0];

				//p_sw_vertices[v].t1.x = (p_sw_vertices[v].t1.x - m_fTexLeft[1]) * m_fInvTexWidth[1];
				//p_sw_vertices[v].t1.y = (p_sw_vertices[v].t1.y - m_fTexTop[1]) * m_fInvTexHeight[1];
			}
		}

		// Bias points in decal mode
		if (IsZModeDecal())
		{
			for ( u32 v = 0; v < m_dwNumVertices; v++ )
			{
				p_sw_vertices[v].pos.z += DECAL_Z_OFFSET;
			}
		}
	}

	//
	// If smooth shading is turned off, duplicate the first diffuse color for all three verts
	//
	if ( !mSmoothShade )
	{

	}

	{
		DAEDALUS_PROFILE( "Prepare Blend" );

		InitBlendMode();
		PrepareVerticesForBlendMode( p_sw_vertices, m_dwNumVertices );
	}

#ifdef FORCE_FLAT_SHADING
	//sceGuDisable(GU_ALPHA_TEST);
	//sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
	//sceGuEnable(GU_BLEND);
	sceGuDisable(GU_TEXTURE_2D);
	sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
	//sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);
	sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gWhiteTexture);
	sceGuTexMode(GU_PSM_8888,0,0,GL_TRUE);		// maxmips/a2/swizzle = 0

	//sceGuDisable(GU_BLEND);
	//sceGuDisable(GU_DEPTH_TEST);
#endif

#ifdef USE_NATIVE_PROJECTION

	sceGumMatrixMode(GU_PROJECTION);
	Matrix4x4		proj( m_mProjection[m_dwProjectionTop] );
	Matrix4x4		proj_trans;
	ScePspFMatrix4	psp_proj;

	proj_trans = proj;//.Transpose();
	memcpy( &psp_proj, &proj_trans, sizeof(ScePspFMatrix4));

	sceGumLoadMatrix( &psp_proj );

	if(m_bCullFront)
	{
		sceGuFrontFace(GU_CW);
		sceGuEnable(GU_CULL_FACE);
	}
	else if( m_bCullBack )
	{
		sceGuFrontFace(GU_CCW);
		sceGuEnable(GU_CULL_FACE);
	}
	else
	{
		sceGuDisable(GU_CULL_FACE);
	}

#endif

	{
		DAEDALUS_PROFILE( "sceGumDrawArray" );

		if(gSoftwareClipping)
		{
			sceGumDrawArray(GU_TRIANGLES,DAEDALUS_VERTEX_FLAGS|GU_TRANSFORM_3D,m_dwNumVertices,NULL,p_sw_vertices);
		}
		else if(gTesselateLargeTriangles)
		{
			sceGumDrawArray(GU_TRIANGLES,DAEDALUS_VERTEX_FLAGS|GU_TRANSFORM_3D,m_dwNumVertices,NULL,p_sw_vertices);
		}
		else
		{
			sceGumDrawArray(GU_TRIANGLES,DAEDALUS_VERTEX_FLAGS|GU_TRANSFORM_3D|GU_INDEX_16BIT,m_dwNumIndices,p_sw_indices,p_sw_vertices);
		}
	}

#ifdef USE_NATIVE_PROJECTION
	sceGuDisable(GU_CULL_FACE);
#endif

	// Reset index
	m_dwNumVertices = 0;
	m_dwNumIndices = 0;

	return true;

}

//*****************************************************************************
// Used for TexRect, TexRectFlip, FillRect
//*****************************************************************************
bool	PSPRenderer::RenderTriangleList( const DaedalusVtx * p_verts, u32 num_verts, const u16 * p_indices, u32 num_indices, bool disable_zbuffer, bool force_point_filter )
{
	if ( disable_zbuffer )
	{
		gNeedZBufferUdate = true;
	}

	if ( force_point_filter )
	{
		// Mariokart hack to make bg on main screen work well Can do this or just clamp values
		// Need filter update!
	}
	

	DaedalusVtx*	p_sw_vertices = (DaedalusVtx*)sceGuGetMemory(num_verts*sizeof(DaedalusVtx));
	u16 *			p_sw_indices = (u16*)sceGuGetMemory(num_indices*sizeof(u16));

	memcpy( p_sw_vertices, p_verts, num_verts*sizeof(DaedalusVtx));
	memcpy( p_sw_indices, p_indices, num_indices*sizeof(u16));

	InitBlendMode();
	PrepareVerticesForBlendMode( p_sw_vertices, num_verts );

	if ( disable_zbuffer )
	{
		sceGuDisable(GU_DEPTH_TEST);
		sceGuDepthMask( GL_TRUE );	// GL_TRUE to disable z-writes
	}

	if ( force_point_filter )
	{
		// Mariokart hack to make bg on main screen work well Can do this or just clamp values
		sceGuTexFilter(GU_NEAREST,GU_NEAREST);
	}


	//sceGuEnable(GU_ALPHA_TEST);
	//sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
	//sceGuEnable(GU_BLEND);
	//sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
	//sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);
	//sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gWhiteTexture);

	//sceGuDisable(GU_BLEND);
	//sceGuDisable(GU_DEPTH_TEST);

	sceGumMatrixMode(GU_PROJECTION);
	sceGumLoadIdentity();
	sceGumOrtho(0, 480, 272, 0, 0.0f, -1.0f);

	sceGumDrawArray(GU_TRIANGLES,DAEDALUS_VERTEX_FLAGS|GU_TRANSFORM_3D|GU_INDEX_16BIT,num_indices,p_sw_indices,p_sw_vertices);

	return true;
}

//*****************************************************************************
// Used for TexRect, TexRectFlip, FillRect
//*****************************************************************************
bool	PSPRenderer::RenderSprites( const DaedalusVtx * p_verts, u32 num_verts, bool disable_zbuffer, bool force_point_filter )
{
	if ( disable_zbuffer )
	{
		gNeedZBufferUdate = true;
	}

	if ( force_point_filter )
	{
		// Mariokart hack to make bg on main screen work well Can do this or just clamp values
		// Need filter update!
	}
	
	DaedalusVtx*	p_sw_vertices = (DaedalusVtx*)sceGuGetMemory(num_verts*sizeof(DaedalusVtx));
	memcpy( p_sw_vertices, p_verts, num_verts*sizeof(DaedalusVtx));

	InitBlendMode();
	PrepareVerticesForBlendMode( p_sw_vertices, num_verts );

	if ( disable_zbuffer )
	{
		sceGuDisable(GU_DEPTH_TEST);
		sceGuDepthMask( GL_TRUE );	// GL_TRUE to disable z-writes
	}

	if ( force_point_filter )
	{
		// Mariokart hack to make bg on main screen work well Can do this or just clamp values
		sceGuTexFilter(GU_NEAREST,GU_NEAREST);
	}



	//sceGuEnable(GU_ALPHA_TEST);
	//sceGuBlendFunc(GU_ADD, GU_SRC_ALPHA, GU_ONE_MINUS_SRC_ALPHA, 0, 0);
	//sceGuEnable(GU_BLEND);
	//sceGuTexFunc(GU_TFX_MODULATE,GU_TCC_RGB);
	//sceGuTexFunc(GU_TFX_REPLACE,GU_TCC_RGB);
	//sceGuTexImage(0,gPlaceholderTextureWidth,gPlaceholderTextureHeight,gPlaceholderTextureWidth,gWhiteTexture);

	//sceGuDisable(GU_BLEND);
	//sceGuDisable(GU_DEPTH_TEST);

	sceGumMatrixMode(GU_PROJECTION);
	sceGumLoadIdentity();
	//sceGumOrtho(0, 480 +1, 272 +1, 0, 0.0f, -1.0f);

	//sceGumDrawArray(GU_TRIANGLES,DAEDALUS_VERTEX_FLAGS|GU_TRANSFORM_3D|GU_INDEX_16BIT,num_indices,p_sw_indices,p_sw_vertices);
	sceGumDrawArray(GU_SPRITES,DAEDALUS_VERTEX_FLAGS|GU_TRANSFORM_2D,num_verts,NULL,p_sw_vertices);

	return true;
}



//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetNewVertexInfo(u32 dwAddress, u32 dwV0, u32 dwNum)
{
	//DBGConsole_Msg(0, "In SetNewVertexInfo");
	FiddledVtx * pVtxBase = (FiddledVtx*)(g_pu8RamBase + dwAddress);

#ifdef USE_NATIVE_PROJECTION
	Matrix4x4 matWorldProject = m_mModelView[m_dwModelViewTop];
#else
	Matrix4x4 matWorldProject = m_mModelView[m_dwModelViewTop] * m_mProjection[m_dwProjectionTop];
#endif

/*
		DL_PF( "MatModel:\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n",
			m_mModelView[m_dwModelViewTop].m[0][0], m_mModelView[m_dwModelViewTop].m[0][1], m_mModelView[m_dwModelViewTop].m[0][2], m_mModelView[m_dwModelViewTop].m[0][3],
			m_mModelView[m_dwModelViewTop].m[1][0], m_mModelView[m_dwModelViewTop].m[1][1], m_mModelView[m_dwModelViewTop].m[1][2], m_mModelView[m_dwModelViewTop].m[1][3],
			m_mModelView[m_dwModelViewTop].m[2][0], m_mModelView[m_dwModelViewTop].m[2][1], m_mModelView[m_dwModelViewTop].m[2][2], m_mModelView[m_dwModelViewTop].m[2][3],
			m_mModelView[m_dwModelViewTop].m[3][0], m_mModelView[m_dwModelViewTop].m[3][1], m_mModelView[m_dwModelViewTop].m[3][2], m_mModelView[m_dwModelViewTop].m[3][3]);

		DL_PF( "MatProject:\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n",
			m_mProjection[m_dwProjectionTop].m[0][0], m_mProjection[m_dwProjectionTop].m[0][1], m_mProjection[m_dwProjectionTop].m[0][2], m_mProjection[m_dwProjectionTop].m[0][3],
			m_mProjection[m_dwProjectionTop].m[1][0], m_mProjection[m_dwProjectionTop].m[1][1], m_mProjection[m_dwProjectionTop].m[1][2], m_mProjection[m_dwProjectionTop].m[1][3],
			m_mProjection[m_dwProjectionTop].m[2][0], m_mProjection[m_dwProjectionTop].m[2][1], m_mProjection[m_dwProjectionTop].m[2][2], m_mProjection[m_dwProjectionTop].m[2][3],
			m_mProjection[m_dwProjectionTop].m[3][0], m_mProjection[m_dwProjectionTop].m[3][1], m_mProjection[m_dwProjectionTop].m[3][2], m_mProjection[m_dwProjectionTop].m[3][3]);

		DL_PF( "MatWorldProject:\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n",
			matWorldProject.m[0][0], matWorldProject.m[0][1], matWorldProject.m[0][2], matWorldProject.m[0][3],
			matWorldProject.m[1][0], matWorldProject.m[1][1], matWorldProject.m[1][2], matWorldProject.m[1][3],
			matWorldProject.m[2][0], matWorldProject.m[2][1], matWorldProject.m[2][2], matWorldProject.m[2][3],
			matWorldProject.m[3][0], matWorldProject.m[3][1], matWorldProject.m[3][2], matWorldProject.m[3][3]);


	DL_PF( "scale: %f,%f,%f trans: %f,%f,%f", mVpScale.x, mVpScale.y, mVpScale.z, mVpTrans.x, mVpTrans.y, mVpTrans.z );
*/
	u32 i;
	for (i = dwV0; i < dwV0 + dwNum; i++)
	{
		v3 vecTransformedNormal;		// Used only when m_bLightingEnable
		v4 w;
		u32 clip_flags;

		FiddledVtx & vert = pVtxBase[i - dwV0];

		w.x = (float)vert.x;
		w.y = (float)vert.y;
		w.z = (float)vert.z;
		w.w = 1.0f;

		v4 & projected( mVtxProjected[i] );
		projected = matWorldProject.Transform( w );

		float	rhw;
		if(fabsf(projected.w) > 0.0f)
		{
			rhw = 1.0f / ( projected.w );
		}
		else
		{
			rhw = 0.0f;
		}

		projected.x *= rhw;
		projected.y *= rhw;
		projected.z *= rhw;
		projected.w = rhw;		// Store the reciprocal here (*= rhw would just give us 1, which is a bit useless)

#ifdef USE_NATIVE_PROJECTION
		m_VtxTransformed[i].pos.x = projected.x;
		m_VtxTransformed[i].pos.y = projected.y;
		m_VtxTransformed[i].pos.z = projected.z;
#else
		// Convert to screen coords here
		m_VtxTransformed[i].pos.x = ( projected.x * mVpScale.x ) + mVpTrans.x;
		m_VtxTransformed[i].pos.y = ( projected.y * mVpScale.y ) + mVpTrans.y;
		m_VtxTransformed[i].pos.z = ( projected.z * mVpScale.z ) + mVpTrans.z;
		// PSP doesn't use this...
		//m_VtxTransformed[i].rhw = rhw;
#endif

		clip_flags = 0;
#ifdef SOFTWARE_CLIPPING
		v3 vecW1( projected.x, projected.y, projected.z );

		vecW1 /= ( projected.w );

		if (vecW1.x > 1.0f) clip_flags |= X_POS;
		else if (vecW1.x < -1.0f) clip_flags |= X_NEG;

		if (vecW1.y > 1.0f) clip_flags |= Y_POS;
		else if (vecW1.y < -1.0f) clip_flags |= Y_NEG;

		if (vecW1.z > 1.0f) clip_flags |= Z_POS;
//		else if (vecW1.z < -1.0f) clip_flags |= Z_NEG;		// NoN
#endif

		m_dwVecFlags[i] = clip_flags;

		if (m_bLightingEnable)
		{
			v3 mn;		// modelnorm

			mn.x = (float)vert.norm_x;
			mn.y = (float)vert.norm_y;
			mn.z = (float)vert.norm_z;

			vecTransformedNormal = m_mModelView[m_dwModelViewTop].TransformNormal( mn );
			vecTransformedNormal.Normalise();

			m_VtxTransformed[i].color = LightVert(vecTransformedNormal);
		}
		else
		{
			m_VtxTransformed[i].color = MAKE_COLOUR_U32( vert.rgba_r, vert.rgba_g, vert.rgba_b, vert.rgba_a );
		}

		// I decided to take this out for v0.06b release - it's not quite
		// stable enough yet
		if (m_bFogEnable)
		{
			float eyespace_z = projected.z / fabsf( projected.w );
			float fFog = (eyespace_z * m_fFogMult) + m_fFogOffset;

			s32 fog_coeff = (s32)(fFog * 255.0f);

			s32 alpha = std::max( std::min( fog_coeff, 255 ), 0 );

			// Set the alpha
			m_VtxTransformed[i].color &= 0x00ffffff;
			m_VtxTransformed[i].color |= alpha << 24;
		}

		if (m_bTextureEnable)
		{
			// Update texture coords n.b. need to divide tu/tv by bogus scale on addition to buffer

			// If the vert is already lit, then there is no normal (and hence we
			// can't generate tex coord)
			if (m_bTextureGen && m_bLightingEnable)
			{
				Matrix4x4 & matWV = m_mModelView[m_dwModelViewTop];
				v3 & norm = vecTransformedNormal;

				// Assign the spheremap's texture coordinates
				m_VtxTransformed[i].t0.x = (0.5f * ( 1.0f + ( norm.x*matWV.m11 +
						                 norm.y*matWV.m21 +
										 norm.z*matWV.m31 ) ))/* * m_fTexScaleX*/;

				m_VtxTransformed[i].t0.y = (0.5f * ( 1.0f - ( norm.x*matWV.m12 +
						                 norm.y*matWV.m22 +
										 norm.z*matWV.m32 ) ))/* * m_fTexScaleY*/;
			}
			else
			{
				m_VtxTransformed[i].t0.x = (float)vert.tu * m_fTexScaleX;
				m_VtxTransformed[i].t0.y = (float)vert.tv * m_fTexScaleY;
			}

			//m_VtxTransformed[i].t1 = m_VtxTransformed[i].t0;
		}
/*
		DL_PF( "   %x: in(%f,%f,%f) -> out(%f,%f,%f,%f) trans(%f,%f,%f) tu/v(%f,%f)",
			i,
			w.x, w.y, w.z,
			projected.x, projected.y, projected.z, projected.w,
			m_VtxTransformed[i].pos.x, m_VtxTransformed[i].pos.y, m_VtxTransformed[i].pos.z,
			m_VtxTransformed[i].t0.x, m_VtxTransformed[i].t0.y );
*/
	}
}


//*****************************************************************************
// Assumes dwAddress has already been checked!	
// Don't inline - it's too big with the transform macros
// DKR seems to use longer vert info
//*****************************************************************************
void PSPRenderer::SetNewVertexInfoDKR(u32 dwAddress, u32 dwV0, u32 dwNum)
{
	s32 nFogR, nFogG, nFogB, nFogA;
	s32 nR, nG, nB, nA;

	if (m_bFogEnable)
	{
		nFogR = COL_GETR(m_dwFogColor);
		nFogG = COL_GETG(m_dwFogColor);
		nFogB = COL_GETB(m_dwFogColor);
		nFogA = COL_GETA(m_dwFogColor);
	}
	s16 * pVtxBase = (s16*)(g_pu8RamBase + dwAddress);

#ifdef USE_NATIVE_PROJECTION
	Matrix4x4 matWorldProject = m_mModelView[m_dwModelViewTop];
#else
	Matrix4x4 matWorldProject = m_mModelView[m_dwModelViewTop] * m_mProjection[m_dwProjectionTop];
#endif

	u32 i;
	s32 nOff;

	nOff = 0;
	for (i = dwV0; i < dwV0 + dwNum; i++)
	{
		v4 w;
		u32 dwFlags;

		w.x = (float)pVtxBase[(nOff + 0) ^ 1];
		w.y = (float)pVtxBase[(nOff + 1) ^ 1];
		w.z = (float)pVtxBase[(nOff + 2) ^ 1];
		w.w = 1.0f;

		v4 & projected( mVtxProjected[i] );
		projected = matWorldProject.Transform( w );	// Convert to w=1

		float	rhw;
		if(fabsf(projected.w) > 0.0f)
		{
			rhw = 1.0f / ( projected.w );
		}
		else
		{
			rhw = 1.0f;
		}

		projected.x *= rhw;
		projected.y *= rhw;
		projected.z *= rhw;
		projected.w = rhw;		// Store the reciprocal here (*= rhw would just give us 1, which is a bit useless)

#ifdef USE_NATIVE_PROJECTION
		m_VtxTransformed[i].pos.x = projected.x;
		m_VtxTransformed[i].pos.y = projected.y;
		m_VtxTransformed[i].pos.z = projected.z;
#else
		// Convert to screen coords here
		m_VtxTransformed[i].pos.x = ( projected.x * mVpScale.x ) + mVpTrans.x;
		m_VtxTransformed[i].pos.y = ( projected.y * mVpScale.y ) + mVpTrans.y;
		m_VtxTransformed[i].pos.z = ( projected.z * mVpScale.z ) + mVpTrans.z;
		//m_VtxTransformed[i].rhw = rhw;
#endif

		dwFlags = 0;

		// Clip
		m_dwVecFlags[i] = dwFlags;

		s16 wA = pVtxBase[(nOff + 3) ^ 1];
		s16 wB = pVtxBase[(nOff + 4) ^ 1];

		s8 r = (s8)(wA >> 8);
		s8 g = (s8)(wA);
		s8 b = (s8)(wB >> 8);
		s8 a = (s8)(wB);

		v3 vecTransformedNormal;

		if (m_bLightingEnable)
		{
			v3 mn;		// modelnorm
			u32 colour;
			Matrix4x4 matWorld = m_mModelView[m_dwModelViewTop];

			mn.x = (float)r; //norma_x;
			mn.y = (float)g; //norma_y;
			mn.z = (float)b; //norma_z;

			vecTransformedNormal = matWorld.TransformNormal( mn );
			vecTransformedNormal.Normalise();

			colour = LightVert(vecTransformedNormal);

			nR = COL_GETR(colour); 
			nG = COL_GETG(colour);
			nB = COL_GETB(colour);
			nA = a;//COL_GETA(colour);
		}
		else
		{
			nR = r;
			nG = g;
			nB = b;
			nA = a;
		}

		// Assign true vert colour after lighting/fogging
		m_VtxTransformed[i].color = MAKE_COLOUR_U32(nR, nG, nB, nA);

		/*if (m_bTextureEnable)
		{
			// Update texture coords n.b. need to divide tu/tv by bogus scale on addition to buffer
			v2 & t = m_vecTexture[i];

			// If the vert is already lit, then there is no normal (and hence we
			// can't generate tex coord)
			if (m_bTextureGen && m_bLightingEnable)
			{
				Matrix4x4 & matWV = m_mModelView[m_dwModelViewTop];
				v3 & norm = vecTransformedNormal;

				// Assign the spheremap's texture coordinates
				t.x = (0.5f * ( 1.0f + ( norm.x*matWV.m00 +
										 norm.y*matWV.m10 +
										 norm.z*matWV.m20 ) ))/ * * m_fTexScaleX* /;

				t.y = (0.5f * ( 1.0f - ( norm.x*matWV.m01 +
										 norm.y*matWV.m11 +
										 norm.z*matWV.m21 ) ))/ * * m_fTexScaleY * /;
			}
			else
			{
				t.x = (float)vert.t.x * m_fTexScaleX;
				t.y = (float)vert.t.y * m_fTexScaleY;
			}
		}*/

		nOff += 5;
	}

}


//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetVtxColor( u32 vert, u32 color )
{
	if ( vert < MAX_VERTS )
	{
		m_VtxTransformed[vert].color = color;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetVtxTextureCoord( u32 vert, short tu, short tv )
{
	if ( vert < MAX_VERTS )
	{
		m_VtxTransformed[vert].t0.x = (float)tu / 32.0f;
		m_VtxTransformed[vert].t0.y = (float)tv / 32.0f;
		// XXXX
//		m_VtxTransformed[vert].t1 = m_VtxTransformed[vert].t0;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetVtxXY( u32 vert, float x, float y )
{
	if ( vert < MAX_VERTS )
	{
		// Convert to screen coords here
		m_VtxTransformed[vert].pos.x = /*round*/( x * m_fScreenMultX + m_VP.X );
		m_VtxTransformed[vert].pos.y = /*round*/( y * m_fScreenMultY + m_VP.Y );

		//m_VtxTransformed[vert].pos.z = (vecProjected.z + 1.0f) * 0.5f;

		// XXXX necessary/correct?
		mVtxProjected[vert].x = m_VtxTransformed[vert].pos.x;
		mVtxProjected[vert].y = m_VtxTransformed[vert].pos.y;
	}
}

//*****************************************************************************
//
//*****************************************************************************

static const int gAddressModeLookup[PSPRenderer::TAM_NUM_ADDRESS_MODES] =
{
	GU_CLAMP,	// TAM_CLAMP = 0,
	GU_CLAMP,	// TAM_MIRROR,
	GU_REPEAT,	// TAM_WRAP,
};

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetAddressMode( u32 index, ETextureAddressMode mode_u, ETextureAddressMode mode_v )
{
	sceGuTexWrap(gAddressModeLookup[mode_u],gAddressModeLookup[mode_v]);
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetAmbientLight( float r, float g, float b)
{
	m_AmbientCol.r = r * 255.0f;
	m_AmbientCol.g = g * 255.0f;
	m_AmbientCol.b = b * 255.0f;
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetLightCol(u32 dwLight, u32 colour)
{
	m_Lights[dwLight].col.r = (float)((colour >> 24)&0xFF);
	m_Lights[dwLight].col.g = (float)((colour >> 16)&0xFF);
	m_Lights[dwLight].col.b = (float)((colour >>  8)&0xFF);
	m_Lights[dwLight].col.a = 255.0f;	// Ignore light alpha
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetLightDirection(u32 l, float x, float y, float z)
{
	// Invert the x/y/z
	m_Lights[l].dir.x = x;
	m_Lights[l].dir.y = y;
	m_Lights[l].dir.z = z;
	m_Lights[l].dir.Normalise();

	m_Lights[l].initialised = true;
}

//*****************************************************************************
// Init matrix stack to identity matrices
//*****************************************************************************
void PSPRenderer::ResetMatrices()
{
	Matrix4x4 mat;

	mat.SetIdentity();

	m_dwProjectionTop = 0;
	m_dwModelViewTop = 0;
	m_mProjection[0] = mat;
	m_mModelView[0] = mat;
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetTextureFactor(u32 colour)
{
	if (m_dwTextureFactor != colour)
	{
		m_dwTextureFactor = colour;
		sceGuTexEnvColor(colour);
	}
}



//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetTexture( u32 index,
							CTexture* lpsTexture,
							s32  nTileLeft, s32 nTileTop,
							u32 dwTileWidth, u32 dwTileHeight )
{
	//lpsTexture = NULL; - For wireframe
	if (m_lpsTexture[index] != lpsTexture)
	{
		// Dereference the previous texture
		if ( m_lpsTexture[index] )
		{
			m_lpsTexture[index]->Release();
			m_lpsTexture[index] = NULL;
		}

		m_lpsTexture[index] = lpsTexture;

		if ( m_lpsTexture[index] )
		{
			m_lpsTexture[index]->AddRef();

			m_fTileWidth[index] = (float)dwTileWidth;
			m_fTileHeight[index] = (float)dwTileHeight;

			// Update the bogus scale parameters
			m_fTexLeft[index] = (float)nTileLeft;
			m_fTexTop[index]  = (float)nTileTop;

			m_fTexWidth[index] = (float)m_lpsTexture[index]->GetWidth();
			m_fTexHeight[index] = (float)m_lpsTexture[index]->GetHeight();
			
			m_fInvTileWidth[index] = 1.0f / m_fTileWidth[index];
			m_fInvTileHeight[index] = 1.0f / m_fTileHeight[index];
			m_fInvTexWidth[index] = 1.0f / m_fTexWidth[index];
			m_fInvTexHeight[index] = 1.0f / m_fTexHeight[index];

		}


		gTextureChanged = true;
	}
}

//*****************************************************************************
//
//*****************************************************************************
// Generate texture coords?
void PSPRenderer::SetTextureGen(bool bTextureGen)
{
	m_bTextureGen = bTextureGen;
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetTextureScale(float fScaleX, float fScaleY)
{
	m_fTexScaleX = fScaleX;
	m_fTexScaleY = fScaleY;
}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::SetCullMode( bool bCullFront, bool bCullBack )
{
	m_bCullFront = bCullFront;
	m_bCullBack = bCullBack;
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetAlphaRef(u32 dwAlpha)
{
	if (m_dwAlpha != dwAlpha)
	{
		m_dwAlpha = dwAlpha;
		sceGuAlphaFunc(GU_GEQUAL, dwAlpha, 0xff );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	PSPRenderer::SetScissor( s32 x0, s32 y0, s32 x1, s32 y1 )
{
	s32	x = s32(x0 * m_fScreenMultX);
	s32 y = s32(y0 * m_fScreenMultY);
	s32 w = s32((x1-x0) * m_fScreenMultX);
	s32 h = s32((y1-y0) * m_fScreenMultY);

	sceGuScissor(x,y, x+w,y+h);
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetProjection(const Matrix4x4 & mat, bool bPush, EMatrixLoadStyle load_style) 
{
	if (bPush)
	{
		if (m_dwProjectionTop >= (MATRIX_STACK_SIZE-1))
			DBGConsole_Msg(0, "Pushing past proj stack limits!");
		else
			m_dwProjectionTop++;

		if (load_style == MATRIX_LOAD)
			// Load projection matrix
			m_mProjection[m_dwProjectionTop] = mat;
		else
			m_mProjection[m_dwProjectionTop] = mat * m_mProjection[m_dwProjectionTop-1];
		
	}
	else
	{
		if (load_style == MATRIX_LOAD)
			// Load projection matrix
			m_mProjection[m_dwProjectionTop] = mat;
		else
			m_mProjection[m_dwProjectionTop] = mat * m_mProjection[m_dwProjectionTop];

	}

}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::SetWorldView(const Matrix4x4 & mat, bool bPush, EMatrixLoadStyle load_style)
{

	// ModelView
	if (bPush)
	{
		if (m_dwModelViewTop >= (MATRIX_STACK_SIZE-1))
			DBGConsole_Msg(0, "Pushing past modelview stack limits!");
		else
			m_dwModelViewTop++;

		// We should store the current projection matrix...
		if (load_style == MATRIX_LOAD)
		{
			// Load projection matrix
			m_mModelView[m_dwModelViewTop] = mat;
		}
		else			// Multiply projection matrix
		{
			m_mModelView[m_dwModelViewTop] = mat * m_mModelView[m_dwModelViewTop-1];
		}
	}
	else	// NoPush
	{
		if (load_style == MATRIX_LOAD)
		{
			// Load projection matrix
			m_mModelView[m_dwModelViewTop] = mat;
		}
		else
		{
			// Multiply projection matrix
			m_mModelView[m_dwModelViewTop] = mat * m_mModelView[m_dwModelViewTop];
		}

	}
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::PopProjection()
{
	if (m_dwProjectionTop > 0)
		m_dwProjectionTop--;
	//else
	//	DBGConsole_Msg(0, "Popping past projection stack limits");
}

//*****************************************************************************
//
//*****************************************************************************
void PSPRenderer::PopWorldView()
{
	if (m_dwModelViewTop > 0)
		m_dwModelViewTop--;
	//else
	//	DBGConsole_Msg(0, "Popping past worldview stack limits");
}

//*****************************************************************************
//
//*****************************************************************************

void PSPRenderer::PrintActive()
{
#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	if (gDisplayListFile != NULL)
	{
		Matrix4x4 mat = m_mModelView[m_dwModelViewTop] * m_mProjection[m_dwProjectionTop];
		DL_PF(
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n"
			" %#+12.5f %#+12.5f %#+12.5f %#+12.5f\n",
			mat.m[0][0], mat.m[0][1], mat.m[0][2], mat.m[0][3],
			mat.m[1][0], mat.m[1][1], mat.m[1][2], mat.m[1][3],
			mat.m[2][0], mat.m[2][1], mat.m[2][2], mat.m[2][3],
			mat.m[3][0], mat.m[3][1], mat.m[3][2], mat.m[3][3]);
	}
#endif
}




// Draw nodes on vertices
/*v3 a, b, c;

a.x = (v0.x * 160) + 160; a.y = (-v0.y * 120) + 120;
b.x = (v1.x * 160) + 160; b.y = (-v1.y * 120) + 120;
c.x = (v2.x * 160) + 160; c.y = (-v2.y * 120) + 120;

g_Renderer->FillRect(a.x, a.y, a.x+2, a.y+2, RGBA_MAKE(255,0,255,255));
g_Renderer->FillRect(b.x, b.y, b.x+2, b.y+2, RGBA_MAKE(255,0,255,255));
g_Renderer->FillRect(c.x, c.y, c.x+2, c.y+2, RGBA_MAKE(255,0,255,255));* /
*/



